/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.aspect;

import jasco.runtime.RuntimeContext;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

public class PCutpointConstructorRegular extends PCutpointConstructorApplicationDesignator  
{	
	private State current;
	private State startState;
	private Vector states = new Vector();
	private int labelToExecute = -1;
	private TreeMap statementMap = null;
	private State prevState = null;
	
	public PCutpointConstructorRegular()  {
		super(null);
	}
	
	public void setStatementMap(TreeMap map)  {
	
		statementMap=map;
	}
	
	public TreeMap getStatementMap()  {
	
		return statementMap;
	}
	
	protected String generateClassCreateCode(String argname) {
		return argname+" = new PCutpointConstructorRegular();"+NEWLINE;
	}
	
	public String getPreviousStateID() {
		
		return prevState.getLabel();
	}
	
	public String generateCreateCode(String argname)  {
	
		String code="";
		
		String newArgName = this.getNextArgName();
		
		code+="PCutpointConstructorRegular "+newArgName+" = null;"+NEWLINE;
		code+=generateClassCreateCode(newArgName);
		code+=argname+" = "+newArgName+";"+NEWLINE;
		
		Iterator i = states.iterator();
		while (i.hasNext())  {
			State state = (State) i.next();
			code+="State "+state.getLabel()+" = new State(\""+state.getLabel()+"\");"+NEWLINE;
			if(state==current)
				code+=newArgName+".setStartState("+state.getLabel()+");"+NEWLINE;
			else code+=newArgName+".addState("+state.getLabel()+");"+NEWLINE;
		}
		TreeMap transitionMap = new TreeMap();
		
		Iterator j = allTransitions().iterator();
		while (j.hasNext())  {
			Transition trans = (Transition) j.next();
			String param =  (String) transitionMap.get(trans.getLabel()+"");
			if (param==null)  {
				param = getNextArgName();
				transitionMap.put(trans.getLabel()+"",param);
			
				code+="PCutpointConstructorApplicationDesignator "+param+" = null;"+NEWLINE;
				code+=trans.getParam().generateCreateCode(param);
			}
			State start = trans.getStartState();
			State dest = trans.getDestState();
			code+=newArgName+".addTransition("+start.getLabel()+", "+trans.getLabel()+", "+param+", "+dest.getLabel()+");"+NEWLINE;
			
		}
		
		return code;
	}
	
	public boolean eval(RuntimeContext context,Hashtable mappings,boolean forJutta)  {
		
		Iterator transitions = current.getTransitions();
		
		if(forJutta)
			transitions =allTransitions().iterator();
		
		while (transitions.hasNext())  {
			Transition trans = (Transition) transitions.next();
			PCutpointConstructorApplicationDesignator next = trans.getParam();
			if (next.eval(context,mappings,forJutta))  {
				if (!forJutta)  {
					prevState=current;
					current=current.getDestState(trans);
					labelToExecute=trans.getLabel();
				}
				return true;
			}
		}
		return false;
	}
	
	public Iterator getNextStateIDs() {
		
		Vector v = new Vector();
		Iterator transitions = current.getTransitions();
		while (transitions.hasNext())  {
			Transition trans = (Transition) transitions.next();
			v.add(trans.getDestState().getLabel());
		}
		return v.iterator();
	}

	/*
	public PCutpointConstructorApplicationDesignator getParameter()  {
	
		current.
	}
	*/
	public boolean doNotCache()  {
		return true;
	}
	
	
	public void addState(State state)  {
		
	//	if(!states.contains(state))
			states.add(state);
	}
	
	public void addTransition(State start, int label, PCutpointConstructorApplicationDesignator des, State dest)  {
		
		Transition trans = new Transition(start,label,des);
		start.addTransition(trans,dest);
	}
	
	public void setStartState(State state)  {
		
		if(!states.contains(state))
			states.add(state);
		
		current=state;
		startState=state;
	}
	
	public State getStartState() {
		
		return startState;
	}
	
	
	public Iterator getStates()  {
		return states.iterator();
	}
	
	public String toDescString()  {
	
		return "Stateful pointcut";
	}
	
	public boolean definesName(String name) {
		
		return getStatementMap().get(name)!=null;
	}
	
	public Vector allPointcuts()  {
	
		Vector trans = new Vector();
		Iterator statesI = getStates();
		while (statesI.hasNext())  {
			State s = (State) statesI.next();
			Iterator transI = s.getTransitions();
			while (transI.hasNext())  {
				Transition t = (Transition) transI.next();
				trans.add(t.getParam());
			}
		}
		return trans;
	}
	public Vector allTransitions()  {
	
		Vector trans = new Vector();
		Iterator statesI = getStates();
		while (statesI.hasNext())  {
			State s = (State) statesI.next();
			Iterator transI = s.getTransitions();
			while (transI.hasNext())  {
				Transition t = (Transition) transI.next();
				trans.add(t);
			}
		}
		return trans;
	}
	
	public String getCurrentStateID()  {
		
		return current.getLabel();
	}
		
		
}
	
	