/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.cache;

import jasco.options.Options;
import jasco.runtime.CombinedHooks;
import jasco.runtime.CutpointAfterComparator;
import jasco.runtime.CutpointBeforeComparator;
import jasco.runtime.CutpointReplaceComparator;
import jasco.runtime.DoNotCache;
import jasco.runtime.JascoMethod;
import jasco.runtime.VectorSorter;
import jasco.runtime.aspect.Cutpoint;
import jasco.runtime.aspect.HookContextProperty;
import jasco.runtime.aspect.HookHelper;
import jasco.runtime.aspect.StatefulAspect;
import jasco.runtime.connector.CutpointElement;
import jasco.runtime.hotswap1.RuntimeBeanTransformerJavassist;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;

import java.io.File;
import java.io.FileWriter;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;


public class CombinedHooksFactory  {

	private static final RuntimeBeanTransformerJavassist beantransformer = new RuntimeBeanTransformerJavassist();
	
	private Hashtable codeCache = new Hashtable();
	private static long classId = 0;
	//private static final File outputDir = new File(Options.getTempDir().getPath() + Options.FILE_SEPARATOR + "jasco"+Options.FILE_SEPARATOR+"runtime"+Options.FILE_SEPARATOR+"cache"+Options.FILE_SEPARATOR);
	protected static final ClassPool classPool = ClassPool.getDefault();
	private static CombinedHooksFactory factory=new CombinedHooksFactory();
	
	private boolean clearLinkedHooks=false;
	private boolean clearProceedCheck=false;
	
	public static final int NO_CONTEXT = 10;
	public static final int STATIC_CONTEXT = 11;
	public static final int DYNAMIC_CONTEXT = 12;
	
	
	
	static  { 
		classPool.appendSystemPath();
	}
	protected static long getNextClassId()  {
	
		return ++classId;
	}
	
	protected void setClearLinkedHooks() {
		
		clearLinkedHooks=true;
	}
	
	protected void setClearProceedCheck() {
		
		clearProceedCheck=true;
	}
	
	public static CombinedHooksFactory getDefault()  {
	
		return factory;
	}
	
	protected String getHookName(int i)  {
	
		return "hook"+i;
	}
	
	protected String accessField(String field)  {
	
		return field;
	}
	
	protected String accessFieldHook(int hook)  {
	
		return accessField(getHookName(hook));
	}
	
	protected CombinedHooks checkForPredef(Vector hooks)  {
		if(hooks.size()!=1)
			return null;
		
		CutpointElement elm = (CutpointElement) hooks.elementAt(0);
		
		if(elm.getCutpoint() instanceof StatefulAspect || elm.getCutpoint() instanceof DoNotCache)
			return null;
		
		boolean hasAfterThrowing = HookHelper.definesAdvice(elm.getCutpoint(),Cutpoint.AFTER_THROWING);
		boolean hasAfterReturning = HookHelper.definesAdvice(elm.getCutpoint(),Cutpoint.AFTER_RETURNING);
		if(hasAfterThrowing||hasAfterReturning) //this is not ever predefined -> return null
			return null;
		
		boolean hasAroundThrowing = HookHelper.definesAdvice(elm.getCutpoint(),Cutpoint.AROUND_THROWING);
		boolean hasAroundReturning = HookHelper.definesAdvice(elm.getCutpoint(),Cutpoint.AROUND_RETURNING);
		if(hasAroundThrowing||hasAroundReturning) //this is not ever predefined -> return null
			return null;
		
		boolean hasBefore = elm.getBefore().getExecution();	
		boolean hasReplace = elm.getReplace().getExecution();
		boolean hasAfter = elm.getAfter().getExecution();
		
		if(hasBefore && hasReplace && hasAfter)
			return new OneReplaceBeforeAfter(hooks);
		else if(hasBefore && !hasReplace && hasAfter)
			return new OneBeforeAfter(hooks);
		else if(hasBefore && !hasReplace && !hasAfter)
			return new OneBefore(hooks);
		else if(!hasBefore && !hasReplace && hasAfter)
			return new OneAfter(hooks);
		else if(!hasBefore && hasReplace && !hasAfter)
			return new OneReplace(hooks);		
		else return null;	
	}
	
	protected void addFieldImpl(String f, CtClass clz) throws Exception {
		
		CtField field = CtField.make(f,clz);
		clz.addField(field);
	}
	
	protected void addField(String f, CtClass clz) throws Exception {
		
		
		addFieldImpl("private final "+f,clz); 
	}
	
	private static int logid;
	
	protected void generateConstructorCode(String cname, java.util.Vector hooks, JascoMethod originalMethod, CtClass fieldClass, CachedAspectRep cachedAspectRep) throws java.lang.Exception {
				
		
		CtClass cutp = classPool.get("jasco.runtime.aspect.Cutpoint");
		CtClass hash = classPool.get("java.util.Hashtable");
		CtClass cutparray = classPool.get("jasco.runtime.aspect.Cutpoint[]");

		
		MethodGenerator constr = new MethodGenerator(cname,"");
		constr.addVariable("java.util.Vector","hooks");
		constr.addModifier(Modifier.PUBLIC);
		
		String constrCode = "";
		boolean processCombStrats = false;
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpointElement = (CutpointElement)hooks.elementAt(i);
			cutpointElement.setId(i);
			
			Cutpoint cp = cutpointElement.getCutpoint();
			
			addField(cp.getClass().getName()+" "+getHookName(i)+";", fieldClass);
			constrCode+="jasco.runtime.aspect.Cutpoint cp = "+"((jasco.runtime.connector.CutpointElement)hooks.elementAt("+i+")).getCutpoint(); //"+cp.getClass().getName()+JavaGenerator.NEWLINE;
			constrCode+=accessFieldHook(i)+"= ("+cp.getClass().getName()+") cp;"+JavaGenerator.NEWLINE;
			
			
			if (cutpointElement.getCutpoint()._Jasco_getConstructorRep().doNotCache())  {
				addField("java.util.Hashtable "+getHookName(i)+"hashtable;", fieldClass);
				
				int index = cutpointElement.getCutpoint()._Jasco_getConnector().indexOf(cutpointElement);
				constrCode+=accessFieldHook(i)+"hashtable = "+" "+accessFieldHook(i)+"._Jasco_getConnector().getMethodHashtable("+index+");"+JavaGenerator.NEWLINE;
				processCombStrats=true;
			}
		}
		
		constr.setImplementation(constrCode);
		
		cachedAspectRep.setConstructorCode(constr);
	}
	
	protected HookContextProperty computeContext(Vector hooks)  {
		
		HookContextProperty[] v = new HookContextProperty[hooks.size()];
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpointElement = (CutpointElement)hooks.elementAt(i);
			Cutpoint cp = cutpointElement.getCutpoint();
			v[i]=cp._Jasco_getHookContextPropertyForAdvice(HookContextProperty.GLOBAL);
		}
			 
		return HookContextProperty.intersection(v);
	}
	
	protected void generateInvokeCode(java.util.Vector hooks, JascoMethod originalMethod, CtClass fieldClass, CachedAspectRep cachedAspectRep) throws java.lang.Exception {
	
		clearLinkedHooks=false;
		clearProceedCheck=false;
		MethodGenerator invokeMethod = new MethodGenerator("invoke","java.lang.Object");
		invokeMethod.addVariable("jasco.runtime.JascoMethod","method");
		invokeMethod.addException("java.lang.Exception");
		invokeMethod.addModifier(Modifier.PUBLIC);
		
		String code = "";
		String startCode= "";
		
		//kan beter
		boolean noReplace = false;
		boolean noAfter = true;
		HookContextProperty requiredHooksContext = computeContext(hooks);

		boolean contextDone=false;
		Vector testHooks = new Vector();
		VectorSorter.sort(hooks,new CutpointReplaceComparator());
		Vector replaceVector = new Vector();
		String firstReplaceHook = "";
		
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpointElement = (CutpointElement)hooks.elementAt(i);
			Cutpoint cp = cutpointElement.getCutpoint();
			int id = cutpointElement.getId();
			
			if(cutpointElement.getAfter().getExecution())
				noAfter=false;
			
			if (cutpointElement.getReplace().getExecution())  
				replaceVector.add(cutpointElement);
			
			if (cp._Jasco_getConstructorRep().doNotCache())  {
				testHooks.add(cutpointElement);
				if (!contextDone)  {
					code+="jasco.runtime.RuntimeContext context = new jasco.runtime.RuntimeContext(method.getFullName(),method);"+JavaGenerator.NEWLINE;
					contextDone=true;
				}
				code+="boolean "+"hook"+id+"condition = "+accessFieldHook(id)+"._Jasco_getConstructorRep().eval(context,"+accessFieldHook(id)+"hashtable,false);"+JavaGenerator.NEWLINE;
			}
			
			if(HookHelper.definesAdvice(cp,Cutpoint.IS_APPLICABLE)&&!testHooks.contains(cutpointElement)) {
				code+="boolean "+"hook"+id+"condition = "+accessFieldHook(id)+".isApplicable("+generateAdviceArgs(Cutpoint.IS_APPLICABLE, cutpointElement)+");"+JavaGenerator.NEWLINE;
				testHooks.add(cutpointElement);
			} else if(HookHelper.definesAdvice(cp,Cutpoint.IS_APPLICABLE))
					code+="hook"+id+"condition = "+accessFieldHook(id)+".isApplicable("+generateAdviceArgs(Cutpoint.IS_APPLICABLE,cutpointElement)+") && hook"+id+"condition;"+JavaGenerator.NEWLINE;
			
		}
		if(replaceVector.size()==0)
			noReplace=true;
		else  {
			firstReplaceHook=accessFieldHook(((CutpointElement)replaceVector.elementAt(0)).getId());
		}
		
		code = startCode+code;
		
		if (replaceVector.size()>0)  {
			//constrCode+="linkedHooks=new jasco.runtime.aspect.Cutpoint["+(replaceVector.size()-1)+"];"+JavaGenerator.NEWLINE;
			addField("jasco.runtime.aspect.Cutpoint[] linkedHooks=new jasco.runtime.aspect.Cutpoint["+(replaceVector.size()-1)+"];",fieldClass);//=new jasco.runtime.aspect.Cutpoint["+(replaceVector.size()-1)+"];",cachedAspect);
		}
		
		String replaceTestCode = "java.util.Vector linkedHooksVector = new java.util.Vector();"+JavaGenerator.NEWLINE;
		boolean testReplaces = false;
		//watch out to never invoke setLinkedHook on original method, otherwise in de soup!
		for (int i = 0; i < replaceVector.size() ; ++i) {
			CutpointElement cutpointElement = (CutpointElement)replaceVector.elementAt(i);
			int id = cutpointElement.getId();
			
			if(i>0)
				cachedAspectRep.appendToConstructorCode(accessField("linkedHooks")+"["+(i-1)+"]="+accessFieldHook(id)+";"+JavaGenerator.NEWLINE);
			
			if (testHooks.contains(cutpointElement))  {
				testReplaces=true;
				replaceTestCode+="if("+"hook"+id+"condition) "+JavaGenerator.NEWLINE;
				replaceTestCode+="linkedHooksVector.add("+accessFieldHook(id)+");"+JavaGenerator.NEWLINE;
			}
			else replaceTestCode+="linkedHooksVector.add("+accessFieldHook(id)+");"+JavaGenerator.NEWLINE;
		}
		
		this.computeInit(replaceVector,requiredHooksContext,testReplaces);
		
		String aroundExType = computeAroundThrowingExceptionType(hooks);
		String aroundThrowingCatchCode="} catch("+aroundExType+" ex) {"+JavaGenerator.NEWLINE;
		boolean noAroundThrowing = true;
		
		String tempCatchCode = "";
		//	do around throwing
		for (int i = 0; i < replaceVector.size() ; ++i) {
			
			CutpointElement cutpointElement = (CutpointElement)replaceVector.elementAt(i);
			int id = cutpointElement.getId();
			if(!HookHelper.definesAdvice(cutpointElement.getCutpoint(),Cutpoint.AROUND_THROWING))
				continue;
			
			if(!noAroundThrowing) {
				tempCatchCode="try{"+tempCatchCode+"continueBoolean=false;}catch("+aroundExType+" ex2) {ex=ex2;}";
				aroundThrowingCatchCode+=tempCatchCode;
				tempCatchCode="";
			}
			else {
				aroundThrowingCatchCode+="boolean continueBoolean=true;"+JavaGenerator.NEWLINE;
				noAroundThrowing=false;
			}
			
			String args= generateAdviceArgs(Cutpoint.AROUND_THROWING,cutpointElement);
			if(!args.equals(""))
				args=", "+args;
			if (testHooks.contains(cutpointElement))  {
				tempCatchCode+="if(hook"+id+"condition) ";
			}
			tempCatchCode+="if(continueBoolean)";
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null) {
				tempCatchCode+="output=";
			}
			tempCatchCode+=accessFieldHook(id)+"."+Cutpoint.AROUND_THROWING+"(ex "+args+");"+JavaGenerator.NEWLINE;
		}
		{
		aroundThrowingCatchCode+=tempCatchCode;
		
		}
		
		String aroundReturningCode = "";
		//		do around returning
		for (int i = replaceVector.size()-1; i >=0 ; --i) {
			
			CutpointElement cutpointElement = (CutpointElement)replaceVector.elementAt(i);
			int id = cutpointElement.getId();
			if(!HookHelper.definesAdvice(cutpointElement.getCutpoint(),Cutpoint.AROUND_RETURNING))
				continue;
			
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null) { //no output define with inline -> no around returning possible...
				if (testHooks.contains(cutpointElement))  {
					aroundReturningCode+="if(hook"+id+"condition) ";
				}
				String args= generateAdviceArgs(Cutpoint.AROUND_RETURNING,cutpointElement);
				if(!args.equals(""))
					args=", "+args;
				aroundReturningCode+="output="+accessFieldHook(id)+"."+Cutpoint.AROUND_RETURNING+"("+this.castAdviceArgs(Cutpoint.AROUND_RETURNING)+"output"+args+");"+JavaGenerator.NEWLINE;
			}
		}
		
		
		cachedAspectRep.addInvokeInitCode(generateInvokeInitCode(replaceVector,requiredHooksContext,noReplace,testReplaces));
		
		
		VectorSorter.sort(hooks,new CutpointBeforeComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpoint = (CutpointElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if (cutpoint.getBefore().getExecution())  {
				if (testHooks.contains(cutpoint))  {
					code+="if(hook"+id+"condition) ";
				}
				code+=accessFieldHook(id)+".before("+generateAdviceArgs(Cutpoint.BEFORE,cutpoint)+");"+JavaGenerator.NEWLINE;
			}
		}
		
		String afterThrowingCatchCode="} catch("+computeAfterThrowingExceptionType(hooks)+" ex) {"+JavaGenerator.NEWLINE;
		boolean noAfterThrowing = true;
		
		VectorSorter.sort(hooks,new CutpointAfterComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpoint = (CutpointElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if(cutpoint.getAfter().getExecution())  {
	
				if(!HookHelper.definesAdvice(cutpoint.getCutpoint(),Cutpoint.AFTER_THROWING))
					continue;
				
				noAfterThrowing=false;
				if (testHooks.contains(cutpoint))  {
					afterThrowingCatchCode+="if(hook"+id+"condition) ";
				}
				String args= generateAdviceArgs(Cutpoint.AFTER_THROWING,cutpoint);
				if(!args.equals(""))
					args=", "+args;
				afterThrowingCatchCode+=accessFieldHook(id)+".after_throwing(ex "+args+");"+JavaGenerator.NEWLINE;
			}	
		}
		
		boolean outputDefined =false;
		String tryCode="";
		if(!noAroundThrowing) {
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null&&!outputDefined)  {
				tryCode+=definedOutput+JavaGenerator.NEWLINE; 
				outputDefined=true;
			}
			tryCode+="try{"+JavaGenerator.NEWLINE;
			aroundThrowingCatchCode+="}";
		}
		else aroundThrowingCatchCode="";
		
		if(!noAfterThrowing) {
			afterThrowingCatchCode+="throw(ex);}"+JavaGenerator.NEWLINE; 
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null&&!outputDefined)  {
				tryCode+=definedOutput+JavaGenerator.NEWLINE; 
				outputDefined=true;
			}
			tryCode+="try{"+JavaGenerator.NEWLINE;
		}
		else afterThrowingCatchCode="";
		
		code+=tryCode;
		
		if (noReplace)  {
			code+=generateOriginalMethodInvokeNoReplace(originalMethod,requiredHooksContext,outputDefined)+";"+JavaGenerator.NEWLINE;
		}
		else if (testReplaces)  {
			if(replaceVector.size()==1) {
				if(!outputDefined) {
					code+="Object output = null;"+JavaGenerator.NEWLINE;
				}
				
				code+="if("+firstReplaceHook+"condition) "+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"output = "+firstReplaceHook+"."+Cutpoint.REPLACE+"("+generateAdviceArgs(Cutpoint.REPLACE,replaceVector)+");"+JavaGenerator.NEWLINE;
				code+="else ";
				code+="output = "+generateOriginalMethodInvoke(originalMethod)+";"+JavaGenerator.NEWLINE;
			}
			else  {
				code+=replaceTestCode;
				if(!outputDefined) {
					code+="Object output = null;"+JavaGenerator.NEWLINE;
				}
				code+="if(linkedHooksVector.size()>0) {"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"jasco.runtime.aspect.Cutpoint firstReplaceHook = linkedHooksVector.elementAt(0);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"linkedHooksVector.removeElementAt(0);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"jasco.runtime.aspect.Cutpoint[] linkedHooksTested = new jasco.runtime.aspect.Cutpoint[linkedHooksVector.size()];"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"linkedHooksVector.copyInto(linkedHooksTested);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"method.setLinkedHooks(linkedHooksTested);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"output = "+firstReplaceHook+"."+Cutpoint.REPLACE+"("+generateAdviceArgs(Cutpoint.REPLACE,replaceVector)+");"+JavaGenerator.NEWLINE;
				clearLinkedHooks=true;
				code+="}"+JavaGenerator.NEWLINE;
				code+="else output = method.invokeOriginalJAsCoMethod();"+JavaGenerator.NEWLINE;	
			}
		}
		else code+=generateReplaceSequence(requiredHooksContext,replaceVector,testReplaces,fieldClass,outputDefined);
		
		code+=aroundThrowingCatchCode+afterThrowingCatchCode;
		code+=aroundReturningCode;
		
		if(clearLinkedHooks)
			code+="method.clearLinkedHooks();"+JavaGenerator.NEWLINE;
		else if(clearProceedCheck)
			code+="method.clearProceedCheck();"+JavaGenerator.NEWLINE;
		
		VectorSorter.sort(hooks,new CutpointAfterComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpoint = (CutpointElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if(cutpoint.getAfter().getExecution())  {
				if (testHooks.contains(cutpoint))  {
					code+="if(hook"+id+"condition) { ";
				}
				else code+="{";
				
				
				String definedOutput = defineOutput(originalMethod,requiredHooksContext);
				if(definedOutput!=null) { //no output define with inline -> no after returning possible...
					
					if(HookHelper.definesAdvice(cutpoint.getCutpoint(), Cutpoint.AFTER_RETURNING)) {
						String args= generateAdviceArgs(Cutpoint.AFTER_RETURNING,cutpoint);
						if(!args.equals(""))
							args=", "+args;
						code+=accessFieldHook(id)+".after_returning("+this.castAdviceArgs(Cutpoint.AROUND_RETURNING)+"output"+args+");"+JavaGenerator.NEWLINE;
					}
					
				}
				code+=accessFieldHook(id)+".after("+generateAdviceArgs(Cutpoint.AFTER,cutpoint)+");"+JavaGenerator.NEWLINE;
				code+="}"+JavaGenerator.NEWLINE;
			}	
		}
		
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpoint = (CutpointElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			code+=moreInvokeCode(cutpoint.getCutpoint(), accessFieldHook(id), "hook"+id+"condition");
	
		}
	
		code+=doReturn("output",noReplace,requiredHooksContext)+JavaGenerator.NEWLINE;
		
		invokeMethod.setImplementation(code);
		
		
		cachedAspectRep.setInvokeCode(invokeMethod);
	}
	
	protected String moreInvokeCode(Cutpoint hook, String accessHook, String conditionboolean) {
		
		return "";
	}
	
	protected final String computeAfterThrowingExceptionType(Vector hooks) {
		
		return computeThrowingExceptionType(hooks,Cutpoint.AFTER_THROWING);
	}
	
	protected final String computeAroundThrowingExceptionType(Vector hooks) {
		
		return computeThrowingExceptionType(hooks,Cutpoint.AROUND_THROWING);
	}
	
	protected String computeThrowingExceptionType(Vector hooks, String adviceType) {
		
		return "Exception";
	}
	
	public String generateAdviceArgs(String adviceType, CutpointElement applicableHook) {
		
		Vector v = new Vector();
		v.add(applicableHook);
		return generateAdviceArgs(adviceType,v);
	}
	
	public String generateAdviceArgs(String adviceType, Vector applicableHooks) {
		
		return "method, method.getCalledObject(), method.getArgumentsArray()";
	}
	
	public String defineOutput(JascoMethod newMethod, HookContextProperty requiredHooksContext) throws Exception {
		
		return "Object output = null;";
	}
	
	protected String castAdviceArgs(String adviceType) {
		
		
		return "";
			
	}
	
	protected String generateReplaceSequence(HookContextProperty requiredHooksContext, Vector replaceVector, boolean testReplaces, CtClass fieldClass, boolean outputDefined) throws Exception {
	
		String code="";
		String firstReplaceHook = accessFieldHook(((CutpointElement)replaceVector.elementAt(0)).getId());
		
		if(replaceVector.size()==1) {
			if(!outputDefined)
				code+="Object ";
			code+="output = "+firstReplaceHook+"."+Cutpoint.REPLACE+"("+generateAdviceArgs(Cutpoint.REPLACE,replaceVector)+");"+JavaGenerator.NEWLINE;
			setClearProceedCheck();
		}
		else  {		
			code+="method.setLinkedHooks("+accessField("linkedHooks")+");"+JavaGenerator.NEWLINE;
			if(!outputDefined)
				code+="Object ";
			code+="output = "+firstReplaceHook+"."+Cutpoint.REPLACE+"("+generateAdviceArgs(Cutpoint.REPLACE,replaceVector)+");"+JavaGenerator.NEWLINE;
			setClearLinkedHooks();
		}
		
		return code;
	}
	
	protected String generateInvokeInitCode(Vector hooks, HookContextProperty requiredHooksContext, boolean noReplace, boolean testReplaces) throws Exception {
		
		return "";
	}
	 
	protected String generateOriginalMethodInvokeNoReplace(JascoMethod method, HookContextProperty requiredHooksContext, boolean outputDefined) throws Exception {
		
		String code="";
		if(!outputDefined)
			code+="Object ";
		code+="output = "+generateOriginalMethodInvoke(method);
			
		
		return code;
	}
	
	protected String generateOriginalMethodInvoke(JascoMethod method) throws Exception {
		
		return "method.invokeOriginalJAsCoMethod()";
	}
	
	protected String doReturn(String var, boolean noReplace, HookContextProperty requiredHooksContext) throws Exception   {
	
		return "return "+var+";";
	}
	
	//TODO change field accessess...
		
	public CombinedHooks generateCachedAspectJavassist(java.util.Vector thehooks, JascoMethod originalMethod) throws java.lang.Exception {
		
		
		if (Options.hasJuttaPredefinedCache())  {
			CombinedHooks res = checkForPredef(thehooks);
			if (res != null)  {
				if(Options.showDebugOutput())
					jasco.util.logging.Logger.getInstance().showOutput("JASCO DEBUG: using predefined cached behavior "+res.getClass().getName());
				return res;
			}
		}
		
		Vector hooks=(Vector)thehooks.clone();
		String cname = "CachedAspect"+getNextClassId();	
		CachedAspectRep rep = new CachedAspectRep(cname);
		CtClass cachedAspect = classPool.makeClass("jasco.runtime.cache."+cname);
		 
		CtClass combHooks = classPool.get("jasco.runtime.CombinedHooks");
		cachedAspect.addInterface(combHooks);
		

		
		generateConstructorCode(cname,hooks,originalMethod,cachedAspect,rep);	
		generateInvokeCode(hooks,originalMethod,cachedAspect,rep);
		
		if(Options.showDebugOutput()) {
			//System.out.println(Options.getTempDir());
			FileWriter boe = new FileWriter(Options.getLogDir()+File.separator+"jutta"+(logid++)+".txt");
			boe.write(rep.getInvokeCode());boe.write(rep.getConstructorCode());
			boe.close();
		}
		
		cachedAspect.stopPruning(true);
		
		Class cachedClass = null;
		CombinedHooks cached = getCachedResult(rep.getInvokeCode(),rep.getConstructorCode());
		if (cached!=null)  {
			cachedClass=cached.getClass();	
		}
		else  {
			cachedAspect.addConstructor(CtNewConstructor.make(rep.getConstructorCode(),cachedAspect));
			CtMethod invokeMethodBytes = CtNewMethod.make(rep.getInvokeCode(),cachedAspect);
			cachedAspect.addMethod(invokeMethodBytes);
			cachedClass = cachedAspect.toClass();
		}
	
		cached = (CombinedHooks) cachedClass.getConstructor(new Class[]  {Vector.class}).newInstance(new Object[] {thehooks});
		addCachedResult(rep,cached);
		
		cachedAspect.defrost();
		return cached;
		
	} 
	
	
	protected void addCachedResult(CachedAspectRep rep, CombinedHooks hooks)  {
	
		codeCache.put(rep.getInvokeCode()+rep.getConstructorCode(),hooks); 
	}
	
	protected CombinedHooks getCachedResult(String icode, String ccode)  {
	
		CombinedHooks res = (CombinedHooks) codeCache.get(icode+ccode);
		return res;
	}	
	
	protected void computeInit(Vector replaceVector, HookContextProperty requiredHooksContext, boolean testReplaces) {
		
	}
}