/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.connector;


import jasco.runtime.JascoMethod;
import jasco.runtime.RuntimeContext;
import jasco.runtime.aspect.Cutpoint;
import jasco.runtime.aspect.PCutpointConstructorApplicationDesignator;

import java.util.Hashtable;

public class CutpointElement implements Cloneable
{
	private Cutpoint thecutpoint = null;
	private CutpointExecutionElement before = null;
	private CutpointExecutionElement after = null;
	private CutpointExecutionElement replace = null;
	private PCutpointConstructorApplicationDesignator context;
	private String name;
	private Hashtable properties = new Hashtable();
	
	public CutpointElement(Cutpoint aCutpoint, CutpointExecutionElement aBefore, CutpointExecutionElement aAfter, CutpointExecutionElement aReplace, String aName) 
	{
		thecutpoint = aCutpoint;
		before = aBefore;
		after = aAfter;
		replace = aReplace;
		name=aName;
	}
	
	public CutpointElement transformOrder(int start)  {
	
		getBefore().setExecutionNumber(start+getBefore().getOriginalExecutionNumber());
		getAfter().setExecutionNumber(start+getAfter().getOriginalExecutionNumber());
		getReplace().setExecutionNumber(start+getReplace().getOriginalExecutionNumber());
		return this;
	}
	
	public String getName()  {
	
		return name;
	}
	
	public void setCutpointContext(PCutpointConstructorApplicationDesignator context)  {
		
		this.context= context;
	}	
	
	public Cutpoint getCutpoint() 
	{
		return thecutpoint;
	}
	
	public CutpointExecutionElement getBefore() 
	{
		return before;
	}
	
	public CutpointExecutionElement getAfter() 
	{
		return after;
	}
	
	public CutpointExecutionElement getReplace() 
	{
		return replace;
	}
	
	public void setBefore(CutpointExecutionElement aBefore) 
	{
		before = aBefore;
	}
	
	public void setAfter(CutpointExecutionElement aAfter) 
	{
		after = aAfter;
	}
	
	public void setReplace(CutpointExecutionElement aReplace) 
	{
		replace = aReplace;
	}
	
	public boolean isApplicable(RuntimeContext runtimeContext, java.util.Hashtable mappings,JascoMethod aJascoMethod,boolean forJutta)  {
		
		if (aJascoMethod.getCalledObject()==getCutpoint())  {
			return false;
		}
		
		if (context != null) 
		{
			if (context.eval(runtimeContext,mappings,forJutta))
			{	
				if(forJutta)
					return true;
					
				Cutpoint temp = getCutpoint();
				boolean res = temp.isApplicable(aJascoMethod,aJascoMethod.getCalledObject(),aJascoMethod.getArgumentsArray()); 
				return res;
			}
			else 
			{
				return false;
			}
		}
		else 
		{
			return false;
		}
	}
	
	public String toString()  {
	
		return getName();
	}
	
	public CutpointElement copy()  {
		try  {
		return (CutpointElement) clone();
		}catch(CloneNotSupportedException ex)  {
			jasco.util.logging.Logger.getInstance().showError("unexpected exception in cutpointElement.copy() "+ex.getMessage());	
			return null; 	
		}
	}
	
	public boolean equals(Object o)  {
		
		
//		System.out.println("equals "+o.toString()+":"+toString());
		if(o instanceof CutpointElement)
			return o == this;
		else return o.toString().equals(this.toString());
	}	
	
	public void addProperty(Object key, Object value)  {
	
		properties.put(key,value);
	}
	
	protected Hashtable getProperties()  {
	
		return properties;
	}
	private int id = 0;
	public void setId(int id)  {
	
		this.id=id;
	}
	
	public int getId()  {
	
		return id;
	}
		
}