/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.connector;

//import jasco.aspectparser.*;
import jasco.runtime.DoNotCache;
import jasco.runtime.JascoMethod;
import jasco.runtime.RuntimeContext;
import jasco.runtime.aspect.Cutpoint;

import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Hashtable;

public class DynamicCutpointElement extends CutpointElement implements DoNotCache  {
	
	private Hashtable objectMap = new Hashtable();
	private Object currentObject = null;
	private JascoMethod currentMethod = null;
	private CutpointInstanceCreator creator = null;
	
	public static final char INITIALIZER = '$';
	
	public DynamicCutpointElement(Cutpoint aCutpoint, CutpointExecutionElement aBefore, CutpointExecutionElement aAfter, CutpointExecutionElement aReplace, String aName) 
	{
		super(aCutpoint,aBefore,aAfter,aReplace,aName);
	}
	
	public void setInstanceCreator(CutpointInstanceCreator c)  {
		creator = c;
	}
	
	public Cutpoint getCutpoint() 
	{
		if(currentObject==null)
			return super.getCutpoint();
		
		Cutpoint currentCutpoint = (Cutpoint) objectMap.get(currentObject);
		if (currentCutpoint==null)  {
		
			currentCutpoint = createNewInstance(currentMethod);
			objectMap.put(currentObject,currentCutpoint);
		}
		return currentCutpoint;
	}
	
	protected Method findMethod(String name, Class c)  {
	
		Method[] methods = c.getMethods();
		for(int i=0;i<methods.length;++i)
			if(methods[i].getName().equals(name))
				return methods[i];
		
		return null;
	}	
	
	protected Cutpoint createNewInstance(JascoMethod jmethod)  {
		
		Cutpoint newInstance = super.getCutpoint();
		Class c = newInstance.getClass();
		
		try {
		newInstance = (Cutpoint) c.newInstance();
		} catch(java.lang.Exception e)  {
			System.err.println("!!!!!!!!!JASCO SYSTEM ERROR!!!!!!!! @DynamicCutpointElement createNewInstance: "+e.getMessage());
			e.printStackTrace();
		}
		
		
		
		Enumeration propI = getProperties().keys();
		while (propI.hasMoreElements())  {
		
			Object key = (Object) propI.nextElement();
			Object value = getProperties().get(key);
			
			
			String f = (String) key;
			
			if (f.charAt(0)==INITIALIZER)  {
				HookInitializer h = (HookInitializer) value;
				//System.out.println("initialiser: "+h.getClass().getName());
				h.init(newInstance);
				continue;
			}
			
			
			
			Object [] params = new Object[]  {value
			};
			Method m = findMethod(f,c);
			if(m==null)
				throw new IllegalArgumentException("Method "+value+" not found in class "+c.getName());
			
			try {
			m.invoke(newInstance,params);
			}
			catch(java.lang.Exception e)  {
				System.err.println("@DynamicCutpointElement createNewInstance: "+e.getMessage());
				e.printStackTrace();
			}
		} 
			
		
		return newInstance;
	}
	
	//possiblle thread unsafe, keeps object and expects next to be getCutpoint...
	public boolean isApplicable(RuntimeContext runtimeContext, java.util.Hashtable mappings,JascoMethod aJascoMethod, boolean forJutta)  {
	
		currentObject=aJascoMethod.getCalledObject();
		currentMethod=aJascoMethod;
		return super.isApplicable(runtimeContext,mappings,aJascoMethod,forJutta);
	}
	
	
}