/*
 * Created on 24-feb-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jasco.runtime.connector;

import jasco.runtime.MethodJoinpoint;
import jasco.util.RegexpMatcher;

import java.lang.annotation.Annotation;
import java.util.Iterator;
import java.util.Vector;

import javassist.CtClass;

/* 
 Copyright (C) 2002-2005 JAsCo Development Team
 http://ssel.vub.ac.be/jasco
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA 
 */

/**
 * @author Wim
 */

public class MethodSignature {

	private String type;
	private String returnType;
	private String methodname;
	private boolean matchSubTypes=false;
	private Vector annotations = new Vector();
	private Vector arguments = new Vector();
	private String argumentsString = "";
	
	public MethodSignature(String returnType, String type, String methodname) {
		super();
		
		this.type=type;
		this.returnType=returnType;
		this.methodname=methodname;
	}
	
	public boolean matchesJoinpoint(MethodJoinpoint jp) {
		
		if(!RegexpMatcher.match(this.getReturnType(), jp.getReturnType().getName()))
			return false;
		
		if(!RegexpMatcher.match(this.getMethodname(), jp.getName()))
			return false;
		
		if(!matchesClass(jp.getClassName(),jp.getClassLoader()))
			return false;
		
		if(!RegexpMatcher.match(this.getVariableString(), this.getVariableString(jp)))
			return false;
		
		Iterator i = annotations.iterator();
		while(i.hasNext()) {
			String annotation  = i.next().toString();
			Annotation[] realAnnotations = jp.getAnnotations();
			boolean match = false;
			for(int j=0;j<realAnnotations.length;++j) {
				if(RegexpMatcher.match(annotation, realAnnotations[j].annotationType().getName())) {
					match=true;
					break;
				}
			}
			if(!match)
				return false;
		}
		
		
		
		return true;
	}
	
	public boolean matchesClass(String clzName, ClassLoader classLoader) {
		
		if(!RegexpMatcher.match(this.getType(), clzName)) {
			if(this.isMatchSubTypes()) {
				
				try {
					Class superClass = classLoader.loadClass(this.getType());
					
					Class subClass = classLoader.loadClass(clzName);
					
					if(superClass.isAssignableFrom(subClass))
						return true;
					else return false;
					
				}catch(Exception ex) {return false;}
			}
			else return false;
		}
		
		
		return true;
	}
	
	public boolean matchesClass(String clzName) {
		
		return matchesClass(clzName,ClassLoader.getSystemClassLoader());
	}
	
	public boolean matchesClass(String clzName, javassist.ClassPool pool) {
		
		if(!RegexpMatcher.match(this.getType(), clzName)) {
			if(this.isMatchSubTypes()) {
				
				try {
					CtClass superClass = pool.get(this.getType());
					
					CtClass subClass = pool.get(clzName);
					
					if(subClass.subtypeOf(superClass))
						return true;
					else return false;
					
				}catch(Exception ex) {return false;}
			}
			else return false;
		}
		
		
		return true;
	}
	
	public String toString() {
		
		String result = "";
		
		Iterator i = annotations.iterator();
		while(i.hasNext()) {
			String annotation  = i.next().toString();
			result+="@"+annotation;
			result+=" ";
		}
		
		result+=this.getReturnType().toString()+" ";
		result+=this.getType()+".";
		result+=this.getMethodname();
		result+="(";
		result+=getVariableString();
		result+=")";
		
		return result;
	}
	
	public String getVariableString() {
		
		String result ="";
		
		Iterator i = arguments.iterator();
		while(i.hasNext()) {
			String s = i.next().toString();
			result+=s;
			if(i.hasNext())
				result+=", ";
		}
		
		return result;
	}
	
	public String getVariableString(MethodJoinpoint jp) {
		
		String result ="";
		
		Class[] clz = jp.getFormalArgumentTypes();
		for(int i=0;i<clz.length;++i) {
			result+=clz[i].getName();
			if(i<(clz.length-1))
				result+=", ";
		}

		return result;
	}
	
	public Iterator getAnnotations() {
		
		return annotations.iterator();
	}
	
	public Iterator getArguments() {
		
		return arguments.iterator();
	}
	
	protected void addAnnotation(String annotation) {
		
		annotations.add(annotation);
	}
	
	protected void addArgument(String argument) {
		
		arguments.add(argument);
	}
	

	/**
	 * @return Returns the matchSubTypes.
	 */
	public boolean isMatchSubTypes() {
		return matchSubTypes;
	}
	/**
	 * @param matchSubTypes The matchSubTypes to set.
	 */
	protected void setMatchSubTypes(boolean matchSubTypes) {
		this.matchSubTypes = matchSubTypes;
	}
	/**
	 * @return Returns the methodname.
	 */
	public String getMethodname() {
		return methodname;
	}
	/**
	 * @param methodname The methodname to set.
	 */
	protected void setMethodname(String methodname) {
		this.methodname = methodname;
	}
	/**
	 * @return Returns the returnType.
	 */
	public String getReturnType() {
		return returnType;
	}
	/**
	 * @param returnType The returnType to set.
	 */
	protected void setReturnType(String returnType) {
		this.returnType = returnType;
	}
	/**
	 * @return Returns the type.
	 */
	public String getType() {
		return type;
	}
	/**
	 * @param type The type to set.
	 */
	protected void setType(String type) {
		this.type = type;
	}
	
	
	public static MethodSignature build(String returnType, String type, String methodName, String[] arguments, String[] annotations, boolean matchSubtypes) {
		
		MethodSignature result = new MethodSignature(returnType,type,methodName);
		
		for(int i=0;i<arguments.length;++i)
			result.addArgument(arguments[i]);
		
		for(int i=0;i<annotations.length;++i)
			result.addAnnotation(annotations[i]);
		
		result.setMatchSubTypes(matchSubtypes);
		
		return result;
		
	}

}
