/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.connector.traversal;

import gnu.regexp.RE;
import jasco.runtime.JascoMethod;
import jasco.runtime.aspect.Cutpoint;

public class TraversalHookWrapper  implements Comparable{
	
	private int priority = Integer.MAX_VALUE;
	private Cutpoint wrappedHook;
	private String varname;
	private String contextclass;

	public TraversalHookWrapper(String contextclass, Cutpoint hook, String varname, int beforePriority)  {
		this.wrappedHook=hook;
		this.priority=beforePriority;
		this.varname=varname;
		this.contextclass=contextclass;
	}
	
	public Cutpoint getHook()  {
	
		return wrappedHook;
	}
	
	public String getVarName()  {
	
		return varname;
	}
	
	public int compareTo(Object o2)  {
	
		TraversalHookWrapper wrapper = (TraversalHookWrapper) o2;
		
		return this.priority - wrapper.priority;
	}
	
	public String getContextClass()  {
	
		return contextclass;
	}
	
	protected static String makeGNUReg(String s)  {
		
		return jasco.util.RegexpMatcher.makeGNUReg(s);
	}
	
	public boolean matchesContext(String context)  {
		
	
		try {
		RE re2 = new RE(makeGNUReg(getContextClass()));
		
		boolean b =  re2.isMatch(makeGNUReg(context));
		//System.out.println(b);
		
		return b;
		}catch(Exception ex)  {
		
			System.err.println(ex.getMessage());
			ex.printStackTrace();
		}
		return false;
	}
	
	public boolean isApplicable(JascoMethod method)  {
		return wrappedHook.isApplicable(method,method.getCalledObject(),method.getArgumentsArray());
	}
	 
}
	
	
	