/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 8-nov-2004
 */
package jasco.runtime.hotswap;

import jasco.runtime.CRChangeEvent;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.aspect.Cutpoint;
import jasco.runtime.aspect.PCutpointConstructorApplicationDesignator;
import jasco.runtime.aspect.PCutpointConstructorCall;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.CutpointElement;
import jasco.runtime.connector.MethodSignature;

import jasco.runtime.transform.TransformerConstants;
import jasco.util.logging.Logger;

import java.lang.instrument.ClassDefinition;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import javassist.CodeConverter;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.Modifier;
import javassist.expr.ExprEditor;
import javassist.expr.MethodCall;

/**
 * @author Wim
 */
public class CallJPManager implements jasco.runtime.ConnectorRegistryListener,ClassLoadingListener {

	private static Hashtable registeredtasks = new Hashtable();
	private static CallJPManager instance = new CallJPManager();
	private static TreeSet loadedClasses = new TreeSet();
	
	public static CallJPManager getInstance() {
		
		return instance;
	}
	
	public static void load() {} //make sure to load before connector registry...
	
	private CallJPManager() {
		
		//ConnectorRegistry.addConnectorRegistryListener(this);
		//JAsCoHotSwap.getTransformer().addClassLoadingListener(this);
	}
	
	public void classLoaded(CtClass clz, String classname) {
		
		loadedClasses.add(clz.getName());
	
		try {
		Vector v = (Vector) registeredtasks.get(clz.getName());
		if(v==null) return;
		Iterator tasks = v.iterator();
		while(tasks.hasNext()) {
			TransformTask task = (TransformTask) tasks.next();
			task.process();
		}
		}catch(Exception ex) {Logger.getInstance().showError(ex);}
	}
	
	public Vector getCallJPs() {
		
		Vector result = new Vector();
		Iterator enum1 = ConnectorRegistry.getConnectors();
		while(enum1.hasNext())
			getCallJPs((Connector)enum1.next(),result);
		
		return result;
	}
	
	public void getCallJPs(Connector theconnector, Vector result) {
		
		Enumeration enumer = theconnector.getMethods();
		Enumeration hooks = theconnector.getHooks();
		while (enumer.hasMoreElements())  {
			MethodSignature sig = (MethodSignature)enumer.nextElement();
			String classnameString = sig.getType();
			String methodnameString = sig.getMethodname();
			Cutpoint cp = ((CutpointElement) hooks.nextElement()).getCutpoint();
			
			System.out.println(cp._Jasco_getCutpointName());
			
			if(!isCall(cp._Jasco_getConstructorRep()))
				continue;
			
			CallJoinpoint jp = new CallJoinpoint(methodnameString,classnameString,cp);
			result.add(jp);
		}
	}
	
	public void registryChanged(CRChangeEvent event)  {
			
			System.out.println(event);
			switch (event.getType())  {
			
				case CRChangeEvent.CONNECTOR_ADDED:
					patchCallJoinpoints((Connector)event.getParam());
					break;
				case CRChangeEvent.CONNECTOR_REMOVED:
					unpatchCallJoinpoints((Connector)event.getParam());
					break;
				case CRChangeEvent.PROPERTY_CHANGE:
				case CRChangeEvent.COMBSTRAT_ADDED:
				case CRChangeEvent.COMBSTRAT_REMOVED:
					break;
				
			}
	}
	
	
	public void registryUpdated() {}
	
	protected synchronized void patchCallJoinpoints(Connector theconnector) {
		
		Enumeration enumer = theconnector.getMethods();
		Enumeration hooks = theconnector.getHooks();
		while (enumer.hasMoreElements())  {
			MethodSignature sig = (MethodSignature)enumer.nextElement();
			String classnameString = sig.getType();
			String methodnameString = sig.getMethodname();
			Cutpoint cp = ((CutpointElement) hooks.nextElement()).getCutpoint();
			
			System.out.println(cp._Jasco_getCutpointName());
			
			if(!isCall(cp._Jasco_getConstructorRep()))
				continue;
			
			CallJoinpoint jp = new CallJoinpoint(methodnameString,classnameString,cp);
			
			hotswapTargetClasses(jp);
		}
	}
	
	protected void hotswapTargetClasses(CallJoinpoint jp) {
		
		try {
			Vector swaptable = new Vector();
			Class[] classes = JAsCoHotSwap.getSingleton().getAllClasses();
			for(int i=0;i<classes.length;++i)  {
				Class nextClass = classes[i];
		
				CtClass clazz = TransformClassChecker.getSingleton().checkValidType(nextClass.getName());
				if(clazz==null) //not a valid type
					continue;
				
				byte[] result = patchCallJoinpoints(clazz,clazz.getName());
				if(result!=null)
					swaptable.add(new ClassDefinition(nextClass,result));
			}
			
			ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
			swaptable.copyInto(clzs);
			
			JAsCoHotSwap.getSingleton().redefineClasses(clzs);
		}
		catch(Exception ex) {
			Logger.getInstance().showError(ex);
		}
	}
	
	protected boolean isCall(PCutpointConstructorApplicationDesignator des) {
		if(des==null)
			return false;
		else if(des instanceof PCutpointConstructorCall)
			return true;
		else return isCall(des.getParameter())||isCall(des.getParameter2());
	}
	
	
	//TODO: unpacth call joinpoints
	protected synchronized void unpatchCallJoinpoints(Connector theconnector) {
		
	}
	
	protected synchronized byte[] patchCallJoinpoints(CtClass cls, String classname)  {
		
		try  {
		Iterator callJoinpoints = getCallJPs().iterator();
		
		boolean changed=false;
		while (callJoinpoints.hasNext())  {
			CallJoinpoint call = (CallJoinpoint) callJoinpoints.next();
			changed=changed||patchCallJP(cls, call);
		}
		
		byte[] code = null;
		if (changed)  {
			code=cls.toBytecode();
			cls.defrost();
		}
		
		return code;
		}
		
		catch(Throwable x)  {
			Logger.getInstance().showError("Transforming call jp failed :"+classname);
			Logger.getInstance().showError(x.getMessage());
			x.printStackTrace();
		}
		
		return null;
	}
	
	protected boolean isLoaded(String clz) {
		
		return loadedClasses.contains(clz);
	}
	
	class TransformTask {
		
		private CtClass clz;
		private CallJoinpoint jp;
		
		TransformTask(CtClass clz, CallJoinpoint jp) {
			this.clz=clz;
			this.jp=jp;
		}
		
		public void process() throws Exception {
			boolean res = patchCallJP(clz,jp);
			if(res) {
				System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!hotswapping"+clz.getName());
				JAsCoHotSwap.getSingleton().redefineClasses(new CtClass[] {clz});
			}
			else System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!NOT"+clz.getName());
				
		}
	}
	
	protected synchronized boolean patchCallJP(CtClass cls, CallJoinpoint jp) throws Exception {
		
		boolean changed = false;
		if(!isLoaded(jp.getClassName())) { //otherwise not loaded anymore through transformer...
			Vector v = (Vector) registeredtasks.get(jp.getClassName());
			if(v==null) {
				v=new Vector();
				registeredtasks.put(jp.getClassName(),v);
			}
			v.add(new TransformTask(cls,jp));
			return false;
		}
		
		Vector vec  = new Vector();
		CtMethod methods[] = cls.getDeclaredMethods(); 
		
		for(int i=0;i<methods.length;++i) {
			
			if(Modifier.isAbstract(methods[i].getModifiers())) //do not adapt abstract methods
				continue;
			else if(Modifier.isNative(methods[i].getModifiers())) //do not adapt native methods
				continue;
			// only adapted methods, other methods contain JAsCo related stuff now, do not adapt!!1
			else if(TransformerConstants.addedByJAsCo(methods[i])&&!TransformerConstants.isForCall(methods[i].getName()))
				vec.add(methods[i]);
		}
		
		
		Iterator vecI = vec.iterator();
		while(vecI.hasNext()) {
			CtMethod next = (CtMethod) vecI.next();
			changed=changed||patchCallJP(next,jp);
		}
		
		return changed;
	}
	
	protected CtMethod findNewCallMethod(CtClass clz, CtMethod old, String name) throws Exception {
		
		CtMethod[] methods = clz.getDeclaredMethods();
		for(int i=0;i<methods.length;++i) 
			if(methods[i].getName().equals(name)) //&&methods[i].getParameterTypes().equals(old.getParameterTypes())
					return methods[i];
			
		Logger.getInstance().showFullError("Error: method "+name+" not found");
		throw new javassist.NotFoundException(name);
	}
	
	
	protected synchronized boolean patchCallJP(CtMethod method, CallJoinpoint jp)  {
		
		try {
		System.out.println("patching call jps"+method.getName());
		boolean changed = false;
		CallFetcher fetcher = new CallFetcher(jp);
		method.instrument(fetcher);
		
		Iterator calls = fetcher.getCalls();
		while(calls.hasNext()) {
			changed=true;
			CtMethod call = (CtMethod) calls.next();
			if(TransformerConstants.addedByJAsCo(call))
				continue;
			else if(Modifier.isAbstract(call.getModifiers())) //do not adapt abstract methods
				continue;
			else if(Modifier.isNative(call.getModifiers())) //do not adapt native methods
				continue;
			else if(call.getName().indexOf("$")!=-1)
				continue;

			String newName = TransformerConstants.getNewCallMethodName(call);
			 
			System.out.println(call.getName()+" redirecting to "+newName);
			
			CtClass clz = call.getDeclaringClass();
			
			
			try {
			CtMethod newCall = findNewCallMethod(clz,call,newName);
			CodeConverter convertor = new CodeConverter();
			convertor.redirectMethodCall(call,newCall);
			changed= true;
			}
			catch(javassist.NotFoundException ex) {
				
				Logger.getInstance().showFullError(ex);
			}
		}
		
		return changed;
		
		}catch(Exception ex) {
			
			Logger.getInstance().showError(ex);
		}
		return false;
	}
	
	class CallFetcher extends ExprEditor {
		
		private Vector calls = new Vector();
		private CallJoinpoint jp;
		
		public CallFetcher(CallJoinpoint jp) {
			this.jp=jp;
		}
		
		public void edit(MethodCall methodCall) {
		try{
			if(jp.isMatch(methodCall.getMethod()))
				calls.add(methodCall.getMethod());
		}catch(javassist.NotFoundException ex){}
		}
		
		public Iterator getCalls() {
			
			return calls.iterator();
		}
	}
	
	

}
