/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;

public class JAsCoClassLoader {

	private final ClassPool classPool = ClassPool.getDefault();
	private static JAsCoClassLoader instance = new JAsCoClassLoader();
	private Vector encounteredLoaders = new Vector();
	
	public static JAsCoClassLoader getSingleton()  {
		return instance;
	}
	
	public CtClass loadClass(String name) throws NotFoundException {
		
		CtClass clz =  classPool.get(name);
		clz.stopPruning(true);
		return clz;
	}
	
	public void appendPathsToClassPath(ClassLoader loader) {
		
		if(encounteredLoaders.contains(loader))
			return;
		else {
			classPool.appendClassPath(new javassist.LoaderClassPath(loader));
			encounteredLoaders.add(loader);
		}
	}
	
	public void appendListToClassPath(String list) throws NotFoundException {
		
		getClassPool().appendPathList(list);
	}
	
	public CtClass makeClass(String name)  {
	
		CtClass clz =  classPool.makeClass(name);
		clz.stopPruning(true);
		return clz;
	}
	
	public CtClass loadClass(String className, byte[] bytes) throws IOException {
		
		try  {
			return loadClass(className);
		}
		catch(Exception ex)  {
		}
		
		CtClass clz =  classPool.makeClass(new ByteArrayInputStream(bytes));
		clz.stopPruning(true);
		return clz;
	}
	
	public ClassPool getClassPool()  {
	
		return classPool;
	}
}