/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap;

import jasco.options.Options;
import jasco.runtime.transform.TransformerConstants;
import jasco.util.logging.Logger;

import java.io.File;
import java.lang.instrument.ClassDefinition;
import java.lang.instrument.Instrumentation;
import java.util.StringTokenizer;
import java.util.Vector;


public class JAsCoHotSwap  {
	
	private static JAsCoHotSwap instance = new JAsCoHotSwap();
	private Instrumentation inst;
	private static JAsCoTransformer transformer;
	
	private JAsCoHotSwap()  {
	
	}
	
	public static JAsCoTransformer getTransformer()  {
	
		return transformer;
	}
	
	
	
	protected static void emptyTempFolder()  {
		
		if(!Options.deleteTempFiles())
			TransformerConstants.moveClassFiles(Options.FINAL_TEMP_DIR);
		else TransformerConstants.deleteClassFiles();
	}
	
	public static JAsCoHotSwap getSingleton()  {
	
		return instance;
	}
	
	public Instrumentation getInstrumentation()  {
	
		return inst;
	}
	
	public void redefineClasses(javassist.CtClass[] classes) throws Exception {
	
		ClassDefinition[] defs = new ClassDefinition[classes.length];
		for(int i=0;i<classes.length;++i)  {
			classes[i].stopPruning(true);
			defs[i]=new ClassDefinition(Class.forName(classes[i].getName()),classes[i].toBytecode());
			
			classes[i].defrost();
		}
		
		redefineClasses(defs);
	}
	
	public Class[] getAllClasses()  {
	
		return getInstrumentation().getAllLoadedClasses();
	}
	
	public void redefineClasses(ClassDefinition[] classes)  {
		
		/*debug code
		System.out.println("b");
		for(int i=0;i<classes.length;++i)  {
			System.out.println("c");
			try {
			FileOutputStream w = new FileOutputStream(new File(Options.getTempDir(),classes[i].getDefinitionClass().getName()));
			w.write(classes[i].getDefinitionClassFile());
			w.close();
			}catch(Exception ex) {Logger.getInstance().showError(ex);}
		}*/
		redefineClassesImpl(classes);
		
		
	}
	
	public void redefineClassesImpl(ClassDefinition[] classes)  {
		//long temp = System.nanoTime();
		
		try {
		getInstrumentation().redefineClasses(classes);
		}catch(Throwable ex)  {
			Logger.getInstance().showError("error occured while redifining classes...."+ex.getClass().getName()+" "+ex.getMessage());
			ex.printStackTrace();
			//Logger.getInstance().showError(ex);
		}
		
		//long result = System.nanoTime()-temp;
		
		//System.out.println("Boe"+result+";"+classes.length);
	}
	
	protected static class ShutdownHookThread extends Thread {
	
		ShutdownHookThread()  {
		}
		
		public void run()  {
			
			emptyTempFolder();
		}
	}	
			
	
	protected static void addShutdownHook()  {
	
		Runtime.getRuntime().addShutdownHook(new ShutdownHookThread());
	}
	
	public static boolean hotswap1_5Enabled()  {
	
		return getSingleton().inst!=null;
	}
	
	public static void premain(String agentArgs, Instrumentation inst) throws Exception {
		//System.out.println(agentArgs);
		getSingleton().inst=inst;
		Options.setTempDir(new File("."));
		loadHotSwapOptions();
		
		addShutdownHook();
		
		transformer = new JAsCoTransformer();
		
		getSingleton().inst.addTransformer(transformer);
		
		Logger.getInstance().showOutput("JAsCo HotSwap 2: Loading and trapping application, please wait...");
		
		/*
		Class clz = Class.forName("java.net.URLClassLoader");
		clz.newInstance();
		
		jasco.runtime.hotswap.ClassPathChanger.alterClassPath((java.net.URLClassLoader)ClassLoader.getSystemClassLoader(),Options.getTempDir().toURL());
		*/
	}
	
	public static void loadHotSwapOptions()  {
	
		String[] excludedtypes = new String[]  {
				"jasco.","javassist.", "Connector.","gnu.regexp."
		};
		
		String[] nativetypes = new String[]  {
			"short","boolean","int","float","long","char","void","double"
		};
		
		String[] userexcludedtypes = new String[]  {
			"sun.","com.sun.","java.","javax."
		};
	
		boolean overwrite = loadBooleanOption("jasco.hotswap.replaceexcludedtypes");
		
		String tmp = System.getProperty("jasco.hotswap.excludedtypes");
		if(tmp!=null&&!tmp.equals(""))
			userexcludedtypes=initUserExcludedTypes(userexcludedtypes,tmp,overwrite);
		 
		Options.setUseReflectionForProceed(loadBooleanOption("jasco.hotswap.applicationserver"));
		Options.setTransformAllMethods(loadBooleanOption("jasco.hotswap.trapall"));
		Options.setTransformHooks(loadBooleanOption("jasco.hotswap.trapaspects"));
		
		TransformClassChecker.install(new TransformClassChecker(userexcludedtypes,excludedtypes,nativetypes));		
	}
	
	public static boolean loadBooleanOption(String option) {
		
		String tmp = System.getProperty(option);
		if(tmp!=null&&!tmp.equals(""))
			return new Boolean(tmp).booleanValue();
		else return false;
	}
	
	protected static String[] initUserExcludedTypes(String[] userexcludedtypes, String s, boolean overwrite)  {
		
		Vector res = new Vector();
		if (!overwrite)  {
			for(int i=0;i<userexcludedtypes.length;++i)
				res.add(userexcludedtypes[i]);
		}	
		StringTokenizer tokenizer = new StringTokenizer(s,Options.PATH_SEPARATOR);
		while (tokenizer.hasMoreElements())  {
			String o = tokenizer.nextToken();
			if(o.endsWith("*"))
				o=o.substring(0,o.length()-1);
			
			res.add(o); 
		}
		userexcludedtypes = new String[res.size()];
		res.toArray(userexcludedtypes);
		return userexcludedtypes;
	}
}