/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap;

import jasco.options.Options;
import jasco.runtime.CRChangeEvent;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.ConnectorRegistryListener;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.MethodSignature;
import jasco.runtime.inline.InlineCompiler;
import jasco.runtime.transform.JAsCoClassTransformer;
import jasco.util.RegexpMatcher;
import jasco.util.logging.ErrorOutputLogger;
import jasco.util.logging.Logger;

import java.io.File;
import java.lang.instrument.ClassDefinition;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import javassist.ByteArrayClassPath;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtMethod;

public class JAsCoTransformer implements ClassFileTransformer,ConnectorRegistryListener {
	
	//private JAsCoHotSwap hotswap;
	private Hashtable adaptedMethods = new Hashtable();
	private Logger logger = new ErrorOutputLogger();
	private Set loadedClasses = new TreeSet();
	private Vector listeners = new Vector();
	
	
	public JAsCoTransformer()  {
	
		
		ConnectorRegistry.addConnectorRegistryListener(this);
	}
	
	public JAsCoHotSwap getHotSwap()  {
	
		return JAsCoHotSwap.getSingleton();
	}
	
	
	public TransformClassChecker getTransformChecker()  {
	
		return TransformClassChecker.getSingleton();
	}
	

	public Hashtable getAdaptedMethods()  {
	
		return adaptedMethods;
	}
	
	
	public Hashtable getAdaptedMethodsForRemoval()  {
	
		return join(getAdaptedMethods(),jasco.runtime.inline.InlineCompiler.getDefault().getInlineCompiledMethods());
	}
	
	protected Hashtable join(Hashtable table1, Hashtable table2)  {
		
		Hashtable res= new Hashtable();
		addAll(res,table1);
		addAll(res,table2);
		return res;
	}
	
	protected void addAll(Hashtable res, Hashtable table)  {
			
		Iterator i = table.entrySet().iterator();
		while (i.hasNext())  {
			
			Map.Entry entry = (Map.Entry) i.next();
			String s = (String) entry.getKey();
			Vector v = (Vector) entry.getValue();
			Vector resV = (Vector) res.get(s);
			if (resV==null)  {
				resV=new Vector();
				res.put(s,resV);
			}
			resV.addAll(v);
		}
	}
		
	public boolean trapRemoved(String classname, String methodname)  {
	
		Vector v = (Vector)getAdaptedMethods().get(classname);
		if(v==null)
			return false;
		
		return v.remove(methodname);
	}
			
	
	public void registryChanged(CRChangeEvent event)  {
	
		switch (event.getType())  {
		
			case CRChangeEvent.CONNECTOR_ADDED:
				hotswaptargetClasses((Connector)event.getParam());
				break;
			case CRChangeEvent.CONNECTOR_REMOVED:
				unhotswaptargetClasses((Connector)event.getParam());
				break;
			case CRChangeEvent.PROPERTY_CHANGE:
				cpropertyChange((Connector) event.getParam());
				break;
			case CRChangeEvent.ALLCHANGE:
			case CRChangeEvent.COMBSTRAT_ADDED:
			case CRChangeEvent.COMBSTRAT_REMOVED:
				connectorCombStratChange();
				break;
			
		}
	}
	
	public void registryUpdated()  {
	}
	
	protected void connectorCombStratChange()  {
	
		Iterator i = ConnectorRegistry.getConnectors();
		while (i.hasNext())  {
			removeInlineCompiledMethods((Connector)i.next());
		}
	}
	
	protected void cpropertyChange(Connector c)  {
		
		removeInlineCompiledMethods(c);
	}
	
		
	public void doFullClassUnhotswap(Vector classes)  {
	
		Vector swaptable = new Vector();
		Iterator toremoveIt = classes.iterator();
		while (toremoveIt.hasNext()) {
			String classname = (String) toremoveIt.next();
			getAdaptedMethods().remove(classname);
			InlineCompiler.getDefault().removeInlineCompiledMethods(classname);
			
			Logger.getInstance().showDebug("Unswapping "+classname);
			
			try  {
				String classFileName = classname.replace('.','/');
				File f = new File(Options.getTempDir().getCanonicalPath()+File.separator+classFileName+".class");
				if(f.exists())
					f.delete();
				
				ClassPool cp = new ClassPool(null);
				cp.appendSystemPath();
				CtClass clz = cp.get(classname);
				clz.stopPruning(true);
				byte[] bytes = clz.toBytecode();
			 	swaptable.add(new ClassDefinition(Class.forName(classname),bytes));
			 
			}
			catch(Exception ex)  {
				Logger.getInstance().showError("Unswapping failed: "+classname);
				Logger.getInstance().showError(ex);
			}
		}
		ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
		swaptable.copyInto(clzs);
		
		getHotSwap().redefineClasses(clzs);
	}
	
	//todo finish thios mkethod....
	protected void doSingleMethodUnhotswap(Hashtable removalMethods) throws Exception {
		
		Vector swaptable = new Vector();
		Iterator i = removalMethods.entrySet().iterator();
		while (i.hasNext())  {
			Map.Entry entry = (Map.Entry) i.next();
			String classname = (String) entry.getKey();
			Vector methods = (Vector) entry.getValue();
			
			byte[] code = removeTraps(classname,methods);
			
			if(code!=null)
				swaptable.add(new ClassDefinition(Class.forName(classname),code));
		}
		
		ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
		swaptable.copyInto(clzs);
		
		getHotSwap().redefineClasses(clzs);
	}
	
	
	protected byte[] removeTraps(String classname, Vector methods) throws Exception {
		
		CtClass clazz = JAsCoClassLoader.getSingleton().loadClass(classname);
		JAsCoClassTransformer transformer = new JAsCoClassTransformer(clazz);
			
		byte[] code = transformer.removeTrapsAtRuntimeByName(methods);
		if(code==null)
			return null;
		
		Vector v = (Vector) getAdaptedMethods().get(classname);
		if(v!=null)
			v.removeAll(methods);
		
		InlineCompiler.getDefault().removeInlineCompiledMethods(classname,methods);
		
		return transformer.getResultingByteCode();
	}
		
	public void unhotswaptargetClasses(Connector unload)  {
		
		removeInlineCompiledMethods(unload);
	try {
		
		Hashtable toremoveMethods = new Hashtable();
		Hashtable removalMethods = getAdaptedMethods();
		Enumeration adaptedMethodsEnum = removalMethods.keys();
		while (adaptedMethodsEnum.hasMoreElements())  {
			String matchingclassname = (String)adaptedMethodsEnum.nextElement();
			Vector matchingmethods = (Vector) removalMethods.get(matchingclassname);
			Set matchedmethods = new TreeSet();
			
			Iterator matchingmethodsI = matchingmethods.iterator();
			while (matchingmethodsI.hasNext())  {
				String matchingmethod = (String) matchingmethodsI.next();
				Iterator i = ConnectorRegistry.getConnectors();
				while (i.hasNext())  {
					Connector theconnector = (Connector) i.next();
					Enumeration enumer = theconnector.getMethods();
					while (enumer.hasMoreElements())  {
						MethodSignature sig = (MethodSignature)enumer.nextElement();
						String methodname = sig.getMethodname();
						if (sig.matchesClass(matchingclassname) &&RegexpMatcher.match(methodname,matchingmethod))  {
							matchedmethods.add(matchingmethod);
						}
					}	
				}
			}
			
			Vector toremove = new Vector();
			matchingmethodsI = matchingmethods.iterator();
			while (matchingmethodsI.hasNext())  {
				
				String matchingmethod = (String) matchingmethodsI.next();
				if (!matchedmethods.contains(matchingmethod))  {
					toremove.add(matchingmethod);
				}
			}	
			toremoveMethods.put(matchingclassname,toremove);
		} 
		
		doSingleMethodUnhotswap(toremoveMethods);
	}
	catch(Exception ex)  {
			Logger.getInstance().showError("Unloading failed: ");
			Logger.getInstance().showError(ex);		
	}
	}
	
	public void removeInlineCompiledMethods(Connector theconnector)  {
		
		if(!InlineCompiler.getDefault().isEnabled())
			return;
		
		try {
		Hashtable toremoveMethods = new Hashtable();
		Hashtable inlineCompiled = InlineCompiler.getDefault().getInlineCompiledMethods();
		Enumeration adaptedMethodsEnum = inlineCompiled.keys();
		while (adaptedMethodsEnum.hasMoreElements())  {
			String matchingclassname = (String)adaptedMethodsEnum.nextElement();
			Vector matchingmethods = (Vector) inlineCompiled.get(matchingclassname);
			Vector matchedmethods = new Vector();
			
			Iterator matchingmethodsI = matchingmethods.iterator();
			while (matchingmethodsI.hasNext())  {
				String matchingmethod = (String) matchingmethodsI.next();
				Enumeration enumer = theconnector.getMethods();
				while (enumer.hasMoreElements())  {
					MethodSignature sig = (MethodSignature)enumer.nextElement();
					
					String methodnameString = sig.getMethodname();
					if (sig.matchesClass(matchingclassname)&&RegexpMatcher.match(methodnameString,matchingmethod))   {
						//System.out.println("removing inlne compiled method: "+matchingmethod+"/"+matchingclassname);
						matchedmethods.add(matchingmethod);
					}
				}
			}
			
			toremoveMethods.put(matchingclassname,matchedmethods);
		}
		doSingleMethodUnhotswap(toremoveMethods);
		checkWhetherConnectorsAppy(toremoveMethods);
		}
		catch(Exception ex)  {
			Logger.getInstance().showError("Unloading inline compiled methods failed: ");
			Logger.getInstance().showError(ex);		
		}
	}
	
	//TODO possible to optimize...
	public void checkWhetherConnectorsAppy(Hashtable removalMethods)  {
		
		Iterator i = ConnectorRegistry.getConnectors();
		while (i.hasNext())  {
			Connector theconnector = (Connector) i.next();
			hotswaptargetClasses(theconnector);
		}
		
		/*
		Iterator i = removalMethods.entrySet().iterator();
		while (i.hasNext())  {
			Map.Entry entry = (Map.Entry) i.next();
			String matchingclassname = (String) entry.getKey();
			Vector methods = (Vector) entry.getValue();
			
			Iterator matchingmethodsI = matchingmethods.iterator();
			while (matchingmethodsI.hasNext())  {
				String matchingmethod = (String) matchingmethodsI.next();
				Iterator i = ConnectorRegistry.getConnectors();
				while (i.hasNext())  {
					Connector theconnector = (Connector) i.next();
					Enumeration enumer = theconnector.getMethods();
					while (enumer.hasMoreElements())  {
						String em = (String)enumer.nextElement();
						String classname = getClassname(em);
						String methodname = getMethodname(em);
						if (RegexpMatcher.match(classname,matchingclassname)&&RegexpMatcher.match(methodname,matchingmethod))  {
							matchedmethods.add(matchingmethod);
						}
					}	
				}
			}*/
	}
		
		
	
	public void hotswaptargetClasses(Connector c)  {
		
		
	//System.out.println("JASCO DEBUG: TESTING TESTING TESTING TESTING");
		
		try  {
		//Logger.setInstance(new OutputLogger());
		Vector swaptable = new Vector();
		Enumeration enummer = c.getMethods();
		while (enummer.hasMoreElements())  {
			MethodSignature sig = (MethodSignature)enummer.nextElement();
			String methodname = sig.getMethodname();
			
			try {
				Class[] classes = getHotSwap().getAllClasses();
				for(int i=0;i<classes.length;++i)  {
					Class nextClass = classes[i];
				
					try {
					if(getTransformChecker().checkValidType(nextClass.getName())==null)
						continue;
					
					if (sig.matchesClass(nextClass.getName()))  {
						
						byte[] code = transformClassAndMethod(nextClass,methodname);
					
						if(code!=null)
							swaptable.add(new ClassDefinition(nextClass,code));
					}
					}catch(Exception ex)  {
					Logger.getInstance().showError("Transforming failed: "+nextClass.getName());
					Logger.getInstance().showError(ex);
					}
				}
			}catch(Exception ex)  {
				Logger.getInstance().showError("Transforming failed: "+sig);
				Logger.getInstance().showError(ex);
			}
				
		}
		
		ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
		swaptable.copyInto(clzs);
		
		getHotSwap().redefineClasses(clzs);
		}
		catch(Exception ex)  {
				Logger.getInstance().showError("Hotswapping failed for connector: "+c.getName());
				Logger.getInstance().showError(ex);
			}
	}
	
	protected byte[] transformClassAndMethod(Class cls, String methodname) throws Exception {
			
		
			Logger.getInstance().showDebug("Hotswapping: "+cls.getName());	
			//origcode.put(type.name(),getClassPool().get(type.name()).toBytecode());
			byte[] code = transfromClass(cls,cls.getName(),methodname);
			return code;
	}
	
	protected ClassPool getClassPool()  {
	
		return JAsCoClassLoader.getSingleton().getClassPool();
	}
	
	protected  synchronized byte[] transfromClass(Class theclass, String classname,String methodname) throws Exception  {
		
		
		Vector v = (Vector) getAdaptedMethods().get(classname);
		if(v==null)
			v=new Vector();	
		
		CtClass clazz = JAsCoClassLoader.getSingleton().loadClass(classname);
		JAsCoClassTransformer transformer = new JAsCoClassTransformer(clazz);
			
		byte[] code = transformer.insertTrapsAtRuntime(v,methodname,theclass.getClassLoader());
		if(code==null)
			return null;
			
		addAdaptedMethods(classname,transformer.getAdaptedMethods());
		
		return transformer.getResultingByteCode();
	}
	
	protected void addAdaptedMethods(String classname, Vector methods)  {
	
		Vector adapted = (Vector) getAdaptedMethods().get(classname);
		if (adapted==null)  {
			adapted=new Vector();
			getAdaptedMethods().put(classname,adapted);
		}
			
		adapted.addAll(methods);
		
		InlineCompiler.getDefault().removeInlineCompiledMethods(classname,methods);
	}
	
	protected static String getMethodname(String method)  {
		
		method=removeAnnotation(method);
		
		int index = method.indexOf("(");
		if(index==-1)
			return null;
		method=method.substring(0,index);	
		index = method.lastIndexOf(".");
		if(index==-1)
			return null;
		method=method.substring(index+1);
		return method;
	}
	
	public static String removeAnnotation(String signature) {
		
		if(!signature.startsWith("@"))
			return signature;
		
		int index = signature.indexOf(" ");
		if(index!=-1) {
			return signature.substring(index+1);
		}
		else return signature;
	}
				
	
	protected static String getClassname(String method)  {
		
		method=removeAnnotation(method);
		
		int index = method.indexOf(" ");
		if(index==-1)
			return null;
		
		method=method.substring(index+1);
		index = method.indexOf("(");
		if(index==-1)
			return null;
		method=method.substring(0,index);	
		index = method.lastIndexOf(".");
		if(index==-1)
			return null;
		method=method.substring(0,index);
		return method;
	}
	
	protected synchronized byte[] transformLoadedClass(CtClass cls, String classname, byte[] bytes, ClassLoader loader)  {
		
		
		try  {
			
		ClassPool pool = ClassPool.getDefault();
		pool.insertClassPath(new ByteArrayClassPath(classname, bytes));
		
		JAsCoClassTransformer transformer = new JAsCoClassTransformer(cls); 
		
		Vector adaptedMethodsV = new Vector();
		Iterator connectors = ConnectorRegistry.getConnectors();
		
		
		
		boolean changed = false;
		
		while (connectors.hasNext())  {
			Connector theconnector = (Connector) connectors.next();
			//System.out.println("matching "+theconnector.getName());
			Enumeration enumer = theconnector.getMethods();
			while (enumer.hasMoreElements())  {
				MethodSignature sig = (MethodSignature)enumer.nextElement();
				
				String methodnameString = sig.getMethodname();
				
				//System.out.println("matching "+classnameString+";"+classname+"..."+RegexpMatcher.match(classnameString,classname));
				
				if (sig.matchesClass(cls.getName(),JAsCoClassLoader.getSingleton().getClassPool()))  {
					Logger.getInstance().showDebug("hotswapping "+classname);
					//boolean res = beantransformer.transformBean(classname,adaptedMethods,true,methodnameString);
					byte[] result = transformer.insertTrapsAtRuntime(adaptedMethodsV,methodnameString,loader);
					boolean res = result!=null;
					changed=changed||res;
					
					if(res)	
					 	adaptedMethodsV.addAll(transformer.getAdaptedMethods());	
				}
			}
		}
		
		byte[] code = bytes;
		if (changed)  {
			code=transformer.getResultingByteCode();
			getAdaptedMethods().put(classname,adaptedMethodsV);
		}
		
		return code;
		}
		
		catch(Throwable x)  {
			Logger.getInstance().showError("Transforming failed :"+classname);
			Logger.getInstance().showError(x.getMessage());
			x.printStackTrace();
		}
		
		return null;
	}
	
	
	
	public Instrumentation getIntrumentation()  {
	
		return JAsCoHotSwap.getSingleton().getInstrumentation();
	}
	
	protected byte[] alterURLClassLoader(CtClass clz) throws Exception {
	
	
		
		CtMethod[] methods = clz.getMethods();
		for(int i=0;i<methods.length;++i)  {
			if (methods[i].getName().equals("addURL"))  {
				int m = methods[i].getModifiers();
				m=javassist.Modifier.setPublic(m);
			    methods[i].setModifiers(m);
			}
		}
		clz.stopPruning(true);
		return clz.toBytecode();
	}	

	public void addClassLoadingListener(ClassLoadingListener l) {
		
		listeners.add(l);
	}
	public void removeClassLoadingListener(ClassLoadingListener l) {
		
		listeners.remove(l);
	}	
	
	protected void addPathsToClassPath(ClassLoader loader) {
		
		JAsCoClassLoader.getSingleton().appendPathsToClassPath(loader);
		//Options.appendPathsToRuntimeClassPath(loader); uses default as well
	}
	
	public byte[] transform(ClassLoader loader, String className, Class classBeingRedefined, ProtectionDomain protectionDomain, byte[] classfileBuffer)  {
		
		
		if(classBeingRedefined!=null)
			return null;
		
		try  {
		
		String clazzname = className.replace('/','.') ;
		if(loader!=null)
			addPathsToClassPath(loader);

		CtClass clz = getTransformChecker().checkValidType(clazzname);
		if(clz==null)
			return null;
		
		JAsCoClassTransformer preTransformer = new JAsCoClassTransformer(clz);
		boolean result =  preTransformer.preTransform();
		
		if(!result)
			return null;
		
		
		byte[] code =  transformLoadedClass(clz,clz.getName(),preTransformer.getResultingByteCode(),loader);
		//byte[] code2 = CallJPManager.getInstance().patchCallJoinpoints(clz,clz.getName());
		//if(code2!=null)
		//	code=code2;
		
		fireClassLoadingEvent(clz,className);
		
		return code;
		}
		catch(Exception ex)  {
			Logger.getInstance().showFullError(ex);
		}
		return null;
	}
	
	protected void fireClassLoadingEvent(CtClass clz, String classname) {
		
		Iterator i = listeners.iterator();
		while(i.hasNext()) {
			ClassLoadingListener l = (ClassLoadingListener) i.next();
			l.classLoaded(clz,classname);
		}
	}
}
	