/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap;

import jasco.options.Options;
import jasco.util.logging.Logger;
import javassist.CtClass;
//import jasco.util.RegexpMatcher;
//import jasco.util.logging.Logger;

public class TransformClassChecker {
	
	private static TransformClassChecker checker = new TransformClassChecker();
	
	private String[] userexcludedtypes = new String[0];
	private String[] excludedtypes = new String[0];
	private String[] nativetypes = new String[]  {
			"short","boolean","int","float","long","char","void","double"
	};
	
	private TransformClassChecker() {}
	
	public static TransformClassChecker getSingleton()  {
		return checker;
	}
	
	public static void install(TransformClassChecker checker2)  {
	
		checker=checker2;
	}
	
	public TransformClassChecker(String[] userexcludedtypes, String[] excludedtypes, String[] nativetypes)  {
	
		this.nativetypes=nativetypes;
		this.excludedtypes=excludedtypes;
		this.userexcludedtypes=userexcludedtypes;
	}
	
	public boolean transformAspects()  {
	
		return jasco.options.Options.transformHooks();
	}
	
	protected boolean checkTypes(String name, String[] types) {
		
		for(int i=0;i<types.length;++i)  {
			
			if(name.startsWith(types[i])) {
				return false;
			}
		}
		
		return true;
	}
	public CtClass checkValidType(String name)  {
			
		
			if(!checkTypes(name,excludedtypes))
				return null;
			
			
			if(!checkTypes(name,userexcludedtypes))
				return null;
			
			try {
			
			CtClass clz = JAsCoClassLoader.getSingleton().loadClass(name);
			
			if (Options.definesNoAspects(clz))  {
				//Logger.getInstance().showDebug("Checking valid type failed (is jasco bean): "+name);
				return null;			
			}
			if(!transformAspects()&&Options.isAspect(clz)) {
				//Logger.getInstance().showDebug("!!!!!!!!Checking valid type failed (is hook): "+name);
				return null;
			}
			
			if (!transformAspects()&&Options.isAspectBean(clz))  {
				//Logger.getInstance().showDebug("!!!!!!!!Checking valid type failed (is aspect bean): "+name);
				return null;
			}
			if (Options.isTraversalConnector(clz))  {
				//Logger.getInstance().showDebug("Checking valid type failed (is traversal): "+name);
				return null;
			}
			if (Options.isCombinationStrategy(clz))  {
					//Logger.getInstance().showDebug("Checking valid type failed (is combination strategy): "+name);
				return null;
			}
			if (Options.isJAsCoBean(clz))  {
				//Logger.getInstance().showDebug("Checking valid type failed (is jasco bean): "+name);
				return null;			
			}
			if (clz.isInterface())  {
				//Logger.getInstance().showDebug("Checking valid type failed (is jasco bean): "+name);
				return null;			
			}
			
			return clz;	
			}catch(Exception ex)  {
				Logger.getInstance().showDebug(ex);
			}
			return null;
		}
	
	
}

	
	