/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.inline;

import jasco.options.Options;
import jasco.runtime.transform.ClassProcessor;
import jasco.runtime.transform.TransformerConstants;
import jasco.runtime.transform.TrapInserter;
import jasco.util.generators.JavaGenerator;

import java.io.File;
import java.io.FileWriter;
import java.util.HashMap;

import javassist.CtMethod;

public class InlineInserter extends TrapInserter  {
	
	private String invokeCode;
	private HashMap generatedCode = new HashMap();
	private boolean addJPInit = true;
	
	public InlineInserter(ClassProcessor prev, String invokeCode)  {
	
		this(prev,invokeCode,true);
	}
	
	public InlineInserter(ClassProcessor prev, String invokeCode, boolean addJPInit)  {
		
			super(prev);
			this.invokeCode=invokeCode;
			this.addJPInit = addJPInit;
		}
	
	protected String getInvokeCode()  {
	
		return invokeCode;
	}
	
	protected String createNewBody(CtMethod method, int methodid) throws Exception  {
	
		String code = "";//createInitCode(method,methodid);
		
		if(addJPInit)
			code+="jasco.runtime.JascoMethod method = "+accessField(TransformerConstants.getMethodField(methodid))+";"+JavaGenerator.NEWLINE;
		code+=invokeCode;
		if(addJPInit)
			code="{"+code+"}";
		
		
		if (Options.showDebugOutput())  {
			File out = new File(Options.getLogDir()+File.separator+"juttainline"+getUniqueClassID()+".txt");
			FileWriter w = new FileWriter(out);
			w.write(code);
			w.close();
		}
		
		generatedCode.put(method,code);
		
		return code;
	}
	
	public String getGeneratedCode(CtMethod method) {
		
		return (String) generatedCode.get(method);
	}
	
	
	public boolean processMethod(CtMethod themethod, int methodid) throws Exception {


		themethod.setBody(createNewBody(themethod,methodid));
		
		return true;
	}
	
}