/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.inline;

import jasco.options.Options;
import jasco.runtime.JascoMethod;
import jasco.runtime.aspect.Cutpoint;
import jasco.runtime.aspect.HookContextProperty;
import jasco.runtime.aspect.HookHelper;
import jasco.runtime.aspect.PCutpointConstructorRegular;
import jasco.runtime.cache.CachedAspectRep;
import jasco.runtime.cache.CombinedHooksFactory;
import jasco.runtime.connector.CutpointElement;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.runtime.transform.CallbackPerMethodGenerator;
import jasco.runtime.transform.ClassProcessor;
import jasco.runtime.transform.FieldGeneratorForInlineJutta;
import jasco.runtime.transform.MethodProcessor;
import jasco.runtime.transform.TransformerConstants;
import jasco.runtime.transform.TrapInserter;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;
import jasco.util.logging.Logger;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.bytecode.CodeAttribute;

public class InlineJutta extends CombinedHooksFactory  {
	
	private long id = getNextClassId();
	private CtClass targetClass;
	private CtMethod theMethod;
	private JascoMethod jascoMethod;
	
	private boolean inlineReplaceCompilingPossible = false;
	
	public InlineJutta(CtClass clazz)  {
	
		targetClass=clazz;
	}
	
	protected String accessField(String field)  {
	
		return getFieldClassName()+"#"+field;
	}
	
	protected String getFieldClassName()  {
	
		return targetClass.getName()+"$"+"JuttaFieldClass"+id;
	}
	
	protected String getSimpleFieldClassName()  {
	
		return "JuttaFieldClass"+id;
	}
	
	protected CtMethod getOrigMethod()  {
	
		return theMethod;
	}
	
	protected JascoMethod getJascoMethod()  {
	
		return jascoMethod;
	}
	
	public byte[] process(JascoMethod originalMethod, Vector orighooks, ClassLoader loader) throws Exception {
		
		Vector hooks=(Vector)orighooks.clone();
		CachedAspectRep rep = new CachedAspectRep(getFieldClassName());
		CtClass fieldClass = JAsCoClassLoader.getSingleton().makeClass(getFieldClassName());
		fieldClass.setModifiers(Modifier.setPublic(fieldClass.getModifiers()));
		
		CtMethod origMethod = initCtMethod(originalMethod);
		this.theMethod=origMethod;
		this.jascoMethod=originalMethod;
		
		generateConstructorCode(getFieldClassName(),hooks,originalMethod,fieldClass,rep);	
		generateInvokeCode(hooks,originalMethod,fieldClass,rep);
				
		MethodGenerator constr = rep.getConstructorMethod();
		constr.addModifier(java.lang.reflect.Modifier.STATIC);
		constr.addModifier(java.lang.reflect.Modifier.PUBLIC);
		constr.setName("init");
		constr.setReturnType("void");
						
		CtMethod init = CtNewMethod.make(constr.toString(),fieldClass);
		fieldClass.addMethod(init);
		
		String invokeCode = rep.getInvokeMethod().getImplementation();
		invokeCode=rep.getInvokeInitCode()+invokeCode;
		
		Vector methods = new Vector();
		methods.add(origMethod);
		
		MethodProcessor processor = new TrapInserter(classPool);
		processor.setFieldClass(fieldClass);
		processor.setProcessMethods(methods);
		
		
		//disabled this for now, still testing
		//FieldGenerator fgenerator = new FieldGeneratorForCallbackPerMethod(processor);
		FieldGeneratorForInlineJutta fgenerator = new FieldGeneratorForInlineJutta(processor);
		InlineInserter inserter = new InlineInserter(processor,invokeCode);
		CallbackPerMethodGenerator cgenerator = new CallbackPerMethodGenerator(processor);
		//CallbackGenerator cgenerator = new CallbackGenerator(processor);
		
		cgenerator.process(targetClass,originalMethod.getClassName());
		cgenerator.generateClass(Options.getTempDir());
		
		fgenerator.process(targetClass,originalMethod.getClassName());
		try {
		inserter.process(targetClass,originalMethod.getClassName());
		saveInlineCode(orighooks,origMethod);
		}catch(Exception ex)  {
			Logger.getInstance().showError(ex);
		}
		fgenerator.generateClass(Options.getTempDir());
		
		fieldClass.defrost();
		Class javaClass = Class.forName(fieldClass.getName());
		
		java.lang.reflect.Method m = javaClass.getMethod("init",new Class[]  {Vector.class});
		m.invoke(null,new Object[] {orighooks});
		
		return inserter.getTargetByteCode();
	}
	
	protected void computeInit(Vector replaceVector, HookContextProperty requiredHooksContext, boolean testReplaces) {
		
		inlineReplaceCompilingPossible = inlineReplaceCompilingPossible(replaceVector,requiredHooksContext,testReplaces);
	}
	
	protected void saveInlineCode(Vector hooks, CtMethod method) {
		
		if(!InlineCompiler.getDefault().isJumpingAspectsEnabled())
			return;
		
		
		//TODO: to make work with several hooks, Vector of hooks and Vector of desginators (state Ids) to
		// retrieve code
		if(hooks.size()!=1) //jumping aspects only work for 1 hook per jp for the moment...
			return;
		
		CodeAttribute code = method.getMethodInfo().getCodeAttribute();
		
		CutpointElement cpElement = (CutpointElement) hooks.elementAt(0);
		Cutpoint cp = cpElement.getCutpoint();
		if(cp._Jasco_getConstructorRep() instanceof PCutpointConstructorRegular) {//jumping aspect
			PCutpointConstructorRegular reg = (PCutpointConstructorRegular) cp._Jasco_getConstructorRep();
				
			String stateID = InlineCodeCache.NOSTATE;//reg.getCurrentStateID();
			InlineCodeCache.getInstance().addInlineCode(method,code);
			
		}
	}
	
	protected void addFieldClassInstance(CtClass fieldClass) throws Exception {
	
		addField("public static final "+getFieldClassName()+" INSTANCE =new "+getFieldClassName()+"();",fieldClass);
	}
	
	private boolean inlineCompileGenerateReplaceSucceeded=false;
	
	protected String generateReplaceSequence(HookContextProperty requiredHooksContext, Vector replaceVector, boolean testReplaces, CtClass fieldClass, boolean outputDefined) throws Exception {
		
		inlineCompileGenerateReplaceSucceeded=false;
		
		if(!inlineReplaceCompilingPossible(replaceVector,requiredHooksContext,testReplaces))
			return super.generateReplaceSequence(requiredHooksContext,replaceVector,testReplaces,fieldClass,outputDefined);
		
		String code=null;
		Vector accessHooks = new Vector();
		addFieldClassInstance(fieldClass);
		
		for(int i=0;i<replaceVector.size();++i)  {
			CutpointElement cpElem = (CutpointElement) replaceVector.elementAt(i);
			String accessHook =accessFieldHook(cpElem.getId());
			accessHooks.add(accessHook);
		}
		
		CtMethod themethod = getOrigMethod();
		boolean isStatic = Modifier.isStatic(themethod.getModifiers());
		
		int startForArgs = 1;
		if(requiredHooksContext.isDirectCalledObject())
			startForArgs=2;
		String args = generateAdviceArgs(Cutpoint.REPLACE,replaceVector);
		
		InlineReplaceGenerator generator = new InlineReplaceGenerator(fieldClass, accessHooks, replaceVector, classPool, themethod, id, args, startForArgs);
		generator.start();
		
		CutpointElement cpElem = (CutpointElement) replaceVector.elementAt(0);
		String accessHook =accessFieldHook(cpElem.getId());
		
		code=getFieldClassName()+".INSTANCE."+InlineReplaceGenerator.getNewMethodName(Cutpoint.REPLACE,cpElem.getCutpoint(),id)+"("+args+");"+JavaGenerator.NEWLINE;
			
		
		if(!themethod.getReturnType().equals(CtClass.voidType)) {
			String tmp = "";
			if(!outputDefined)
				tmp=themethod.getReturnType().getName()+" ";
			tmp+="output";
			code=tmp+code;
		}
		
		inlineCompileGenerateReplaceSucceeded=true;	
		
		return code;
	}
	
	protected boolean castReturn()  {
	
		return !inlineCompileGenerateReplaceSucceeded;
	}
	
	
	
	//TODO: update this test...
	protected boolean inlineReplaceCompilingPossible(Vector hooks, HookContextProperty requiredHooksContext, boolean testReplaces) {
		
		if(hooks.size()==0)
			return true;
		
		//if(true) return true; //FIXME: implement this method properly!
		if(testReplaces)
			return false;
		else if(requiredHooksContext.isProceedWithDefinedArgs())
			return false;
		else if(!allHooksHaveSameArgs(hooks,Cutpoint.REPLACE))
			return false;
		else if(!hooksExactlyDefineArgs(hooks,Cutpoint.REPLACE))
			return false;
		else return true;
	}
	
	protected String getThisInstance() {
		
		String thisInstance="$0";
		if(Modifier.isStatic(getOrigMethod().getModifiers()))
			thisInstance="jasco.runtime.NULL#instance";
		return thisInstance;
	}
	
	protected String generateInvokeInitCode(Vector hooks, HookContextProperty requiredHooksContext, boolean noReplace, boolean testReplaces) throws Exception {
		
		String code = "";
		
		boolean inlineReplacePossible = noReplace||inlineReplaceCompilingPossible(hooks,requiredHooksContext,testReplaces);
		
		if(requiredHooksContext.isDirectArgsArray()||requiredHooksContext.isReflArgs()||!inlineReplacePossible)
			code+="Object[] args = $args;"+JavaGenerator.NEWLINE;
		if(requiredHooksContext.isReflArgs()||!inlineReplacePossible)
			code+="method.setArgumentsArray(args);"+JavaGenerator.NEWLINE;
		if(requiredHooksContext.isReflCalledObject()||requiredHooksContext.isProceedWithDefinedArgs()||!inlineReplacePossible) {
			code+="method.setCalledObject("+getThisInstance()+");"+JavaGenerator.NEWLINE;
		}
		
		/*
		if(Modifier.isStatic(getOrigMethod().getModifiers()))
			code=TransformerConstants.getClallBackPerMethodObject(getOrigMethod())+".initialise($$);"+JavaGenerator.NEWLINE;
		else code=TransformerConstants.getClallBackPerMethodObject(getOrigMethod())+".initialise("+thisInstance+",$$);"+JavaGenerator.NEWLINE;
		*/
		
		return code;
	}
	
	protected HookContextProperty computeAdviceContext(String adviceType, Vector advices) {
		
		HookContextProperty[] combinedHooksForAdvice = new HookContextProperty[advices.size()];
		Iterator i = advices.iterator();
		int j = 0;
		while(i.hasNext()) {
			CutpointElement hook = (CutpointElement) i.next();
			HookContextProperty prop = hook.getCutpoint()._Jasco_getHookContextPropertyForAdvice(adviceType);
			combinedHooksForAdvice[j]=prop;
		}
		
		return HookContextProperty.intersection(combinedHooksForAdvice);
	}
	
	public String generateAdviceArgs(String adviceType, Vector advices) {
		
		HookContextProperty requiredContext = computeAdviceContext(adviceType,advices);
		CutpointElement firstHook = (CutpointElement) advices.elementAt(0);
		
		if(adviceType.startsWith(Cutpoint.REPLACE))
			return generateReplaceAdviceArgs(requiredContext,adviceType,advices); //TODO: implement replace
		else return generateNonReplaceAdviceArgs(requiredContext,adviceType,firstHook.getCutpoint());
	}
	
	/**
	 * prereq: all hooks have the same args and return type!!!
	 * 
	 * checks whether hooks have the same return type and expected jp arguments as the jp method itself
	 * 
	 * @param hooks
	 * @param adviceType
	 * @return
	 */
	// 
	protected boolean hooksExactlyDefineArgs(Vector hooks, String adviceType) {
		
		try {
		//		ceck whether all hooks have the same arguments...:
		Iterator hooksI = hooks.iterator();
		Class[] args = null;
		boolean allHooksTheSame = true;
		CutpointElement hook = (CutpointElement) hooksI.next();
			
		Method advicemethod = findReflectMethod(hook.getCutpoint(),adviceType);
		if(advicemethod==null) //args used, no reason for inline replace...
			return false;
			
		Class[] paramtypes = advicemethod.getParameterTypes();
		CtClass[] paramsJP = getOrigMethod().getParameterTypes();
		
		int startForArgs = 2;
		
		if(paramtypes.length-startForArgs!=paramsJP.length)
			return false;
		
		for(int i=startForArgs,j=0;j<paramsJP.length&&i<paramtypes.length;++i,++j)
			if(!paramtypes[i].getName().equals(paramsJP[j].getName()))
				return false;
		
		if(advicemethod.getReturnType().getName().equals( getOrigMethod().getReturnType().getName()))
			return true;
		else return false;
		} catch(Exception ex) {Logger.getInstance().showError(ex);}
		
		return false;
	}
	
	/**
	 * Checks whether all hooks have the same expected jp args and jp return type
	 * @param hooks
	 * @param adviceType
	 * @return
	 */
	protected boolean allHooksHaveSameArgs(Vector hooks, String adviceType) {
		
		//		ceck whether all hooks have the same arguments...:
		Iterator hooksI = hooks.iterator();
		Class[] args = null;
		Class rtType = null;
		boolean allHooksTheSame = true;
		while(hooksI.hasNext()) {
			CutpointElement hook = (CutpointElement) hooksI.next();
			
			Method advicemethod = findReflectMethod(hook.getCutpoint(),adviceType);
			
			if(advicemethod==null) //args used, no use for inline generation...
				return false;
			
			if(args!=null) {
				allHooksTheSame=allHooksTheSame&&Arrays.equals(args,advicemethod.getParameterTypes());
				allHooksTheSame=allHooksTheSame&&advicemethod.getReturnType().equals(rtType);
			}
			else {
				args=advicemethod.getParameterTypes();
				rtType=advicemethod.getReturnType();
			}
				
		} 
		
		
		return allHooksTheSame;
		
	}
	
	
	//watch out: around throwing and around returning also here, proceed must work there as well! (idea BART V)
	protected String generateReplaceAdviceArgs(HookContextProperty requiredContext, String adviceType, Vector hooks) {
		
		CutpointElement firstHook = (CutpointElement) hooks.elementAt(0);
		
		HookContextProperty prop = new HookContextProperty(requiredContext.isReflArgs(),requiredContext.isDirectArgsArray(),
				true, true, requiredContext.isReflCalledObject(), requiredContext.isProceedWithDefinedArgs());
		
		String s= generateNonReplaceAdviceArgs(prop,adviceType,firstHook.getCutpoint());
		/* when args at the end for optimzied around
		try {
		CtClass[] fargs = theMethod.getParameterTypes();
		s+=", "+getThisInstance();
		for(int i=1;i<=fargs.length;++i)
			s+=", $"+i;
		}catch (javassist.NotFoundException ex) {jasco.util.logging.Logger.getInstance().showError(ex);}
		System.out.println(s);
		*/
		return s;
	}
	
	protected String castAdviceArgs(String adviceType) {
		
		
		if(adviceType.equals(Cutpoint.AFTER_RETURNING)||adviceType.equals(Cutpoint.AROUND_RETURNING)) {
			Class expectedType = this.getJascoMethod().getReturnType();
			if(expectedType.equals(Void.TYPE))
				return super.castAdviceArgs(adviceType);
			else return "("+expectedType.getName()+")";
		}
		else return super.castAdviceArgs(adviceType);
			
	}
	
	
	protected String generateNonReplaceAdviceArgs(HookContextProperty requiredContext, String adviceType, Cutpoint hook) {
		
		Method adviceMethod  = findReflectMethod(hook,adviceType);
	
		if(adviceMethod==null) //default method
			return "method, "+getThisInstance()+", args";
		
		String argCode="";
		Class[] args = adviceMethod.getParameterTypes();
		int start = 0;
		if(requiredContext.isDirectCalledMethod()) {
			argCode+="method";
			start++;
		}
		if(requiredContext.isDirectCalledObject()) {
			if(start>0)
				argCode+=", ";			
			argCode+=getThisInstance();
			start++;
		}
		if(adviceType.equals(Cutpoint.AFTER_RETURNING)||adviceType.equals(Cutpoint.AFTER_THROWING)||adviceType.equals(Cutpoint.AROUND_THROWING)||adviceType.equals(Cutpoint.AROUND_RETURNING))
			start++; //additional argument to ignore..
		for(int i=start,j=1;i<args.length;++i,++j) {
			
			if(i>0)
				argCode+=", ";	
			//if this var is the array with arguments, this is always last
			if(i==args.length-1&&args[i].getName().equals("[Ljava.lang.Object;")&&requiredContext.isDirectArgsArray()) {
				
				argCode+="args";
			}
			else {
				
				argCode+="$"+j; //access argument of original method...
			}
		}
		
		return argCode;
	}
	
	/**
	 * finds the most specific advice method, if only default method is available, null is returned
	 * @param hook
	 * @param adviceType
	 * @return
	 */
	protected Method findReflectMethod(Cutpoint hook, String adviceType) {
		
		Method[] methods = hook.getClass().getMethods();
		for(int i=0;i<methods.length;++i)
			if(methods[i].getName().equals(adviceType)) { //method found, check whether not default
				Class[] args = methods[i].getParameterTypes();
				if(adviceType.equals(Cutpoint.AFTER_THROWING)||adviceType.equals(Cutpoint.AFTER_RETURNING)||adviceType.equals(Cutpoint.AROUND_RETURNING)||adviceType.equals(Cutpoint.AROUND_THROWING)) {
					if(args.length!=4) {
						if(adviceType.equals(Cutpoint.AFTER_RETURNING)||adviceType.equals(Cutpoint.AROUND_RETURNING)) {
							Class expectedType = this.getJascoMethod().getReturnType();
							if((methods[i].getParameterTypes()[0]).isAssignableFrom(expectedType))//if type of return is assignable to the first arg of this method
								return methods[i];
						}
						else return methods[i];
					}
				}
				else if(args.length!=3)
					return methods[i];
				else {
					int	start=0;
					if(adviceType.equals(Cutpoint.AFTER_THROWING)||adviceType.equals(Cutpoint.AFTER_RETURNING)||adviceType.equals(Cutpoint.AROUND_RETURNING)||adviceType.equals(Cutpoint.AROUND_THROWING))
						start=1;
					if(!args[start].getName().equals("jasco.runtime.JascoMethod"))
						return methods[i];
					if(!args[start+1].getName().equals("java.lang.Object"))
						return methods[i];
					if(!args[start+2].getName().equals("[Ljava.lang.Object;"))
						return methods[i];
				}
			}
		
			return null;
	}
	
	//TODO: find commomn ancestor!!!!!s
	protected String computeThrowingExceptionType(Vector hooks, String adviceType) {
		
		Class exceptionType = null;
		for (int i = 0; i < hooks.size() ; ++i) {
			CutpointElement cutpoint = (CutpointElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if(cutpoint.getAfter().getExecution())  {
				if(!HookHelper.definesAdvice(cutpoint.getCutpoint(),adviceType))
					continue;
				
				Method method = findReflectMethod(cutpoint.getCutpoint(),adviceType);
				if(method==null) {
					return super.computeThrowingExceptionType(hooks,adviceType); 
				}
				Class clz = method.getParameterTypes()[0];
				if(exceptionType==null)
					exceptionType=clz;
				else return super.computeThrowingExceptionType(hooks,adviceType); 
			}	
		}
		if(exceptionType==null)
			return super.computeThrowingExceptionType(hooks,adviceType);
		else return exceptionType.getName();
	}
	
	
	
	public String defineOutput(JascoMethod newMethod, HookContextProperty requiredHooksContext) throws Exception {
		
		if(!inlineReplaceCompilingPossible)
			return super.defineOutput(newMethod,requiredHooksContext);
		
		CtMethod themethod = getOrigMethod();
		if (!themethod.getReturnType().equals(CtClass.voidType))  {
			return themethod.getReturnType().getName()+" output;";
		} 
		else return null;
	}
	
	protected String generateOriginalMethodInvokeNoReplace(JascoMethod method, HookContextProperty requiredHooksContext, boolean outputDefined) throws Exception {
		
		CtMethod themethod = getOrigMethod();
		
		String code="";
		String newname = TransformerConstants.getNewMethodName(themethod);
		
		
		if (!themethod.getReturnType().equals(CtClass.voidType))  {
			if(!outputDefined)
				code+=themethod.getReturnType().getName()+ " ";
			code+="output = "+newname+"($$)";
		} else code+=newname+"($$)";
		
		return code;
	}
	
	
	protected String doReturn(String var, boolean noReplace, HookContextProperty requiredHooksContext) throws Exception {
		
		CtMethod themethod = getOrigMethod();
			
		if (noReplace||!castReturn())  {
			if(themethod.getReturnType()!=CtClass.voidType) 
				return "return output;";
			else return "";
		}	
		
		String startCode="";
		String returnCode="";
		String code="";
		if (themethod.getReturnType()!=CtClass.voidType)    {
			returnCode+="return ";
			
			if (themethod.getReturnType().isPrimitive())  {
				String type = ClassProcessor.makePrimitiveObject(themethod.getReturnType());
				startCode+=type+" res = "+"("+type+")";
				returnCode+="res."+themethod.getReturnType().getName()+"Value();";
			}
			else  {
				String type = themethod.getReturnType().getName();
				startCode+=type+" res = "+" ("+type+") ";
				returnCode+=" res;";
			}
		
			code+=startCode;
			code+=var+";"+JavaGenerator.NEWLINE;
			code+=returnCode;
		}
		else code="return;"+JavaGenerator.NEWLINE;
		
	
		
		return code;
	}
	
	
	
	//TODO UNSAFE WITH METHOD OVERLOADING!!!!!
	protected CtMethod initCtMethod(JascoMethod method) throws Exception {
		
		CtMethod res = targetClass.getDeclaredMethod(method.getName());
		return res;
	}
	
	protected void makeFieldsStaticPublic(CtClass clazz) throws Exception {
	
		CtField[] fields = clazz.getDeclaredFields();
		
		for(int i=0;i<fields.length;++i)  {
			CtField field = fields[i];
			int modifiers = field.getModifiers();
			modifiers=Modifier.setPublic(modifiers);
			modifiers|= Modifier.STATIC;
			field.setModifiers(modifiers);
		}
	}
	
	protected void addField(String f, CtClass clz) throws Exception {
	
		addFieldImpl("public static "+f,clz); 
	}
	
	protected String moreInvokeCode(Cutpoint hook, String accessHook, String conditionboolean) {
		
		if(!(hook._Jasco_getConstructorRep() instanceof PCutpointConstructorRegular))
			return super.moreInvokeCode(hook,accessHook, conditionboolean);
		
		if(!InlineCompiler.getDefault().isJumpingAspectsEnabled())
			return super.moreInvokeCode(hook,accessHook,conditionboolean); 
		
	
		String code = super.moreInvokeCode(hook,accessHook,conditionboolean);
		code+="if("+conditionboolean+")"+"jasco.runtime.inline.JumpingAspectsWeaver#getInstance().reweave("+accessHook+", method);";

		return code;
	}
}
