/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.options.Options;
import jasco.util.generators.JavaGenerator;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.util.Iterator;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;

public class CallbackPerMethodGenerator extends MethodProcessor  {
	
	private Vector clzs = new Vector();
	
	public CallbackPerMethodGenerator(ClassProcessor g)  {
	
		super(g);
	}
	
	public void generateClass(File dir) throws Exception {
		
		
		Iterator clzsI = clzs.iterator();
		while (clzsI.hasNext())  {
			CtClass clz = (CtClass) clzsI.next();
			generateDaClass(clz,dir);
		}
		
	}
	
	public void generateDaClass(CtClass clz, File dir) throws Exception {
	
		String classFileName = dir.getPath()+File.separator+clz.getName().replace('.','/');
		
		addNoAspects(clz); 
		makeDir(dir,clz);
		
		File classFile = new File(classFileName+".class");
		FileOutputStream writer = new FileOutputStream(classFile);
				clz.stopPruning(true);
		writer.write(clz.toBytecode());
		writer.close();
		clz.defrost();
		
		TransformerConstants.registerClassFile(classFile);
	}
		
	public boolean startProcessing(CtClass clz, String classname) throws Exception {
		
		clzs.clear();
		
		super.startProcessing(clz,classname);
	
		return true; 
	}
	
	private static int logid = 0;
	
	
	//TODO: add invoke(args)

	public boolean processMethod(CtMethod method, int methodid) throws Exception {
		
		String fullClassName = TransformerConstants.getCallbackPerMethodName(method);
		
		CtClass callbackInterface = getClassPool().get("jasco.runtime.JAsCoBeanCallbackMethod");
		
		CtClass callback = getClassPool().makeClass(fullClassName);
		callback.addInterface(callbackInterface);
		
		CtMethod oldMethod=getNewMethod(method);
		
		String clzname=method.getDeclaringClass().getName();
		String invokeParams="";
		String initMethodBody = "";
		CtClass params[] = method.getParameterTypes();
		String initMethodHeader = "public final void initialise(";
		if (!Modifier.isStatic(method.getModifiers()))  {
			initMethodHeader+=	clzname+" pobject";	
			if(params.length>0)
				initMethodHeader+=", ";
			addField("private static "+clzname+" object;",callback);
			initMethodBody+="object=pobject;";
		}	
		for(int i=0;i<params.length;++i)  {
			String var = "var"+i;
			String pvar = "pvar"+i;
			invokeParams+="var"+i;
			if(i<params.length-1)
				invokeParams+=", ";
			addField("private static "+params[i].getName()+" "+var+";",callback);
			initMethodBody+=var+"="+pvar+";"+JavaGenerator.NEWLINE;
			initMethodHeader+=params[i].getName()+" "+pvar;
			if(i<params.length-1)
				initMethodHeader+=", ";
		}
		String initCode = initMethodHeader+") {"+JavaGenerator.NEWLINE+initMethodBody+JavaGenerator.NEWLINE+"}";
		
		
		
		String runCode="";
		
		boolean isStatic = Modifier.isStatic(oldMethod.getModifiers());
		if(isStatic)
			runCode+=getTargetClass().getName()+"."+oldMethod.getName()+"(";
		else runCode+="object."+oldMethod.getName()+"(";

		runCode+=invokeParams+")"+JavaGenerator.NEWLINE;
		
		if(oldMethod.getReturnType().equals(CtClass.voidType))
			runCode+=";return jasco.runtime.NULL#instance;"+JavaGenerator.NEWLINE;
		else if (oldMethod.getReturnType().isPrimitive())  
			runCode="return "+"new "+makePrimitiveObject(oldMethod.getReturnType())+"("+runCode+");"+JavaGenerator.NEWLINE;
		else runCode="return "+runCode+";"+JavaGenerator.NEWLINE;
		
		String runMethod = "public final java.lang.Object dispatch() throws Exception {"+JavaGenerator.NEWLINE+runCode+"}";
		
		if(Options.showDebugOutput()) {
			//System.out.println(Options.getTempDir());
			FileWriter boe = new FileWriter(Options.getLogDir()+File.separator+"callback"+(logid++)+".txt");
			boe.write(runMethod);boe.write(JavaGenerator.NEWLINE+"===================="+JavaGenerator.NEWLINE);
			boe.write(initCode);
			boe.close();
		}
		
		CtMethod init = CtNewMethod.make(initCode,callback);
		callback.addMethod(init);
		
		CtMethod dispatch = CtNewMethod.make(runMethod,callback);
		callback.addMethod(dispatch);
		
		String dispatch2Code=clzname +" object"+" = ("+clzname+") obj;"+JavaGenerator.NEWLINE;
		CtClass[] params2 = oldMethod.getParameterTypes();
		for(int i=0;i<params2.length;++i) {
			String src = getSourceName(params[i]);
			if(params[i].isPrimitive()) 
				dispatch2Code+="var"+i+"= (("+makePrimitiveObject(params[i])+") v["+i+"])."+params[i].getName()+"Value();"+JavaGenerator.NEWLINE;
			else dispatch2Code+="var"+i+"= ("+src+") v["+i+"];"+JavaGenerator.NEWLINE;
		}
		dispatch2Code+=runCode;
		String dispatch2="public final java.lang.Object dispatch(int i, java.lang.Object obj, java.lang.Object[] v) throws Exception {"+JavaGenerator.NEWLINE;
		dispatch2+=dispatch2Code+="}"; 
		
		CtMethod dispatch2m = CtNewMethod.make(dispatch2,callback);
		callback.addMethod(dispatch2m);
		
		addField("public static final "+fullClassName+" INSTANCE = new "+fullClassName+"();",callback);
		
		addGetArgsMethod(callback,oldMethod);
		addGetObjectMethod(callback,oldMethod);
		
		clzs.add(callback);
		
		return true;
	}
	
	protected void addGetObjectMethod(CtClass callback, CtMethod oldMethod)  throws Exception  {
	
		String code = "";
		if (!Modifier.isStatic(oldMethod.getModifiers()))  
			code+="return object;";
		else code+="return jasco.runtime.NULL#instance;";	
		
		code="public java.lang.Object getCalledObject() {"+JavaGenerator.NEWLINE+code+JavaGenerator.NEWLINE+"}";
		
		CtMethod addGetArgsMethod = CtNewMethod.make(code,callback);
		callback.addMethod(addGetArgsMethod);
		
		
	}
	
	protected void addGetArgsMethod(CtClass callback, CtMethod oldMethod)  throws Exception  {
	
		CtClass[] params2 = oldMethod.getParameterTypes();
		String code = "";
		//code+="if(args==jasco.runtime.NULL#instance) {"+JavaGenerator.NEWLINE;
		//code+="   args=new java.lang.Object["+params2.length+"];"+JavaGenerator.NEWLINE;
		
		for(int i=0;i<params2.length;++i) {
			String var = "var"+i;
			if(params2[i].isPrimitive())
				code+=	"args["+i+"] = "+"new "+makePrimitiveObject(params2[i])+"("+var+");"+JavaGenerator.NEWLINE;
			else code+="args["+i+"] = "+var+";"+JavaGenerator.NEWLINE;
		}
		code+="return args;"+JavaGenerator.NEWLINE;
		
		code="public java.lang.Object[] getArgumentsArray() {"+code+"}";
		
		addField("public static java.lang.Object[] args=new java.lang.Object["+params2.length+"];",callback);
		
		CtMethod addGetArgsMethod = CtNewMethod.make(code,callback);
		callback.addMethod(addGetArgsMethod);
	}
	
}
		