/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 22-okt-2004
 */
package jasco.runtime.transform;

import java.io.File;
import java.io.FileOutputStream;
import java.util.Vector;

import javassist.ClassPool;
import javassist.*;
import jasco.runtime.hotswap.*;
import jasco.util.generators.JavaGenerator;
import javassist.CtMethod;

/**
 * @author Wim
 */
public class EventTransformer extends MethodProcessor {

	/**
	 * 
	 */
	public EventTransformer() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param p
	 */
	public EventTransformer(ClassPool p) {
		super(p);
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param g
	 */
	public EventTransformer(ClassProcessor g) {
		super(g);
		// TODO Auto-generated constructor stub
	}

	/* (non-Javadoc)
	 * @see jasco.runtime.transform.MethodProcessor#processMethod(javassist.CtMethod, int)
	 */
	protected boolean processMethod(CtMethod method, int methodid)
			throws Exception {
		
		String name = method.getName();
		
		if(name.startsWith("add")&&name.endsWith("Listener")) {
			
			if(method.getParameterTypes().length!=1)
				return false;
			
			try {
			CtClass eventClass = method.getParameterTypes()[0];
			if(eventClass.isInterface())
				processEvent(eventClass);
			else return false;
			}
			catch(NotFoundException ex) {
				return false;
			}
		}
		return true;	
	}
	
	protected void processEvent(CtClass clz) throws Exception {
		
		
		String name = clz.getName();
		
		String addname = "add"+clz.getSimpleName();
		String removename = "remove"+clz.getSimpleName();
		
		CtMethod addmethod = assertMethodExists(addname,clz);
		CtMethod removemethod = assertMethodExists(removename,clz);
		if(addmethod==null||removemethod==null)
			return;
		
		String newaddname = copymethod(addmethod);
		String newremovename = copymethod(removemethod);
		
		generateNewAddOrRemoveCode("add",addmethod,name);
		generateNewAddOrRemoveCode("remove",removemethod,name);
		
		CtClass eventInnerClass = generateEventInnerClass(clz);
		adaptConstructotToRegister(eventInnerClass,newaddname);
	}
	
	protected void adaptConstructotToRegister(CtClass eventInnerClass, String addname) throws Exception {
		
		CtConstructor[] constrs = getTargetClass().getConstructors();
		for(int i=0;i<constrs.length;++i) {
			CtConstructor constr = constrs[i];
			constr.insertAfter(addname+"(new "+eventInnerClass.getName()+"($0));");
		}
	}
	
	protected String getFireMethodName(CtClass eventClass,CtMethod method) {
		return "fireJAsCo"+eventClass.getName().replace('.','_')+"_"+method.getName();
	}
	
	public static final String getFireMethod(String name, CtClass clz) {
		if(!name.startsWith("fireJAsCo"))
			return null;
		
		int index = name.lastIndexOf("_");
		String thename = name.substring(index+1);
		String result = "onevent "+clz.getName()+"."+thename;
		
		return result;
	}
	
	protected void generateFireMethod(CtMethod event, CtClass eventClass) throws Exception {
		
		CtMethod newmethod = CtNewMethod.copy(event,getFireMethodName(eventClass,event),getTargetClass(),null);
		int modifiers = 0;
		modifiers= Modifier.setPublic(modifiers);
		newmethod.setModifiers(modifiers);
		
		String code = "{";
		code+="java.util.Vector v = jasco.runtime.connector.OnEventTable#getListeners($0,\""+eventClass.getName()+"\");"+JavaGenerator.NEWLINE;
		code+="java.util.Iterator i = v.iterator();"+JavaGenerator.NEWLINE;
		code+="while(i.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=eventClass.getName()+" listener = i.next();"+JavaGenerator.NEWLINE;
		code+="listener."+event.getName()+"($$);"+JavaGenerator.NEWLINE;
		code+="}return;}"+JavaGenerator.NEWLINE;
		newmethod.setBody(code);
		
		getTargetClass().addMethod(newmethod);
		
		ClassPreTransformer preTransformer = new ClassPreTransformer();
		preTransformer.setTargetClass(getTargetClass());
		preTransformer.processMethod(newmethod,-1);
		JAsCoClassTransformer transformer = new JAsCoClassTransformer(getTargetClass());
		transformer.insertTrapsAtRuntime(new Vector(),newmethod.getName(),null);
		/*
		TrapInserter trapInserter = new TrapInserter(this);
		Vector v = new Vector();
		v.add(newmethod);
		trapInserter.setProcessMethods(v);
		*/
	}
	
	protected CtClass generateEventInnerClass(CtClass eventClass) throws Exception {
		
		CtClass eventInnerClass = JAsCoClassLoader.getSingleton().makeClass(getTargetClass().getName()+"$_JAsCo_"+eventClass.getName().replace('.','_'));
		eventInnerClass.addInterface(eventClass);
		addField("private "+getTargetClass().getName()+" object=null;",eventInnerClass);
		CtMethod[] methods = eventClass.getDeclaredMethods();
		for(int i=0;i<methods.length;++i) {
			CtMethod eventmethod = methods[i];
			CtMethod newmethod = CtNewMethod.copy(eventmethod,eventInnerClass,null);
			int modifiers = 0;
			modifiers= Modifier.setPublic(modifiers);
			newmethod.setModifiers(modifiers);
			
			generateFireMethod(newmethod,eventClass);
			
			String code = "{object."+getFireMethodName(eventClass,newmethod)+"($$);return;}";
			newmethod.setBody(code);
			
			eventInnerClass.addMethod(newmethod);
		}
		
		CtClass[] params = new CtClass[] {getTargetClass()};
		CtConstructor constr = CtNewConstructor.make(params,new CtClass[0],"this.object=$1;",eventInnerClass);
		
		eventInnerClass.addConstructor(constr);
		
		generateClass(eventInnerClass,jasco.options.Options.getTempDir());
		
		return eventInnerClass;
	}
	
	protected void generateClass(CtClass clz, File dir) throws Exception {
		
		String filedesc = clz.getName().replace('.',File.separatorChar)+".class";
		File file = new File(dir,filedesc);
		TransformerConstants.registerClassFile(file);
		
		makeDir(dir,clz);
		
		FileOutputStream stream = new FileOutputStream(file);
		stream.write(clz.toBytecode());
		stream.close();
		clz.defrost();
		
		TransformerConstants.registerClassFile(file);
	}
	
	protected void generateNewAddOrRemoveCode(String todo, CtMethod toChange, String listenername) throws Exception {
		
		String code="";
		code+="jasco.runtime.connector.OnEventTable#"+todo+"Listener($0, \""+listenername+"\", $1);"+JavaGenerator.NEWLINE;
		
		toChange.setBody(code);
		
	}
	
	protected String copymethod(CtMethod method) throws Exception {
		
		String newname = getNewMethodName(method)+"ForEvents";
		
		CtMethod copy = CtNewMethod.copy(method,newname,getTargetClass(),null);
		int modifiers = copy.getModifiers();
		modifiers=javassist.Modifier.setPublic(modifiers);
		copy.setModifiers(modifiers);
		//copy.setBody(null);
		getTargetClass().addMethod(copy);
		
		return newname;
	}
	
	protected CtMethod assertMethodExists(String name, CtClass eventClass) throws Exception {
		
		CtMethod[] methods = getTargetClass().getMethods();
		for(int i=0;i<methods.length;++i) {
			if(methods[i].getName().equals(name)&&!Modifier.isStatic(methods[i].getModifiers())) {
				CtClass[] params = methods[i].getParameterTypes();
				if(params.length==1&&params[0].equals(eventClass))
					return methods[i];
			}
		}
		
		return null;		
	}

}
