/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.util.generators.JavaGenerator;

import java.io.File;
import java.io.FileOutputStream;
import java.lang.reflect.Method;

import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.CtNewMethod;

public class FieldGenerator extends MethodProcessor  {
	
	private StringBuffer initCode = new StringBuffer();
	private ClassLoader loader = null;
	
	public FieldGenerator(ClassProcessor g,ClassLoader loader)  {
	
		super(g);
		this.loader=loader;
	}
	
	protected ClassLoader getClassLoader() {
		
		return loader;
	}
	
	public boolean processMethod(CtMethod method, int methodid) throws Exception {
		
		 
		Long id = new Long(methodid);
		String methodString = getMethodString(method,getTargetClass());
		String jascoTrapField = "public static final java.lang.String _Jasco_JascoMethod"+id.toString()+"ID = \""+methodString+"\";";	
		addField(jascoTrapField,getFieldClass());
		
		String callBackObject = generateCallbackObject(method);
		
		initCode.append(generateInitCode(method,"_Jasco_JascoMethod"+id.toString()));
		
		String jascoMethodField = "public static final jasco.runtime.JascoMethod _Jasco_JascoMethod"+id.toString()+" ";
		jascoMethodField+=" = new "+generateJascoMethodType()+"(\""+methodString+"\",  \""+method.getName()+"\" , \""+getTargetClass().getName()+"\", "+id.toString()+", "+callBackObject+", "+method.getModifiers()+", true);"+JavaGenerator.NEWLINE;
		addField(jascoMethodField,getFieldClass());
	
		return true;
	}
	
	public String generateInitCode(CtMethod method, String methodvar) throws Exception {
		
		String res = "";
		CtClass[] fargs = method.getParameterTypes();
		CtClass[] excs = method.getExceptionTypes();
		res+="{String[] fargs = new String["+fargs.length+"];"+JavaGenerator.NEWLINE;
		for(int i=0;i<fargs.length;++i)
			res+="fargs["+i+"] = \""+fargs[i].getName()+"\";"+JavaGenerator.NEWLINE;
		res+=methodvar+".setFormalArgumentTypes(fargs);"+JavaGenerator.NEWLINE;
		res+="String[] excs = new String["+excs.length+"];"+JavaGenerator.NEWLINE;
		for(int i=0;i<excs.length;++i)
			res+="excs["+i+"] = \""+excs[i].getName()+"\";"+JavaGenerator.NEWLINE;
		res+=methodvar+".setExceptionTypes(excs);}"+JavaGenerator.NEWLINE;
		res+=methodvar+".initClassLoader(loader);"+JavaGenerator.NEWLINE;
		return res;
	}
	
	public void initClassLoader(ClassLoader loader) throws Exception{
		
		if(loader==null||loader==ClassLoader.getSystemClassLoader())
			return;
		Class clz = ClassLoader.getSystemClassLoader().loadClass(getFieldClass().getName());
		Method m = clz.getMethod("init", new Class[] {ClassLoader.class});
		m.invoke(null,new Object[] {getClassLoader()});
	}
	
	public String generateCallbackObject(CtMethod method)  {
		 

		return TransformerConstants.getClallBackObject(this,getUniqueClassID(),method);
	}
	
	public String generateJascoMethodType()  {
	
		return "jasco.runtime.JascoMethod";
	}
	
	public void generateCallbackClassField() throws Exception {
	
		String thefield = "private static final jasco.runtime.JAsCoBeanCallback INSTANCE = new "+TransformerConstants.getCallbackClassFullname(getTargetClass(),getUniqueClassID())+"();"+JavaGenerator.NEWLINE;
		//String thefield = "public static jasco.runtime.JAsCoBeanCallback INSTANCE;"; // new "+TransformerConstants.getCallbackClassFullname(getTargetClass(),getUniqueClassID())+"();"+JavaGenerator.NEWLINE;
		addField(thefield,getFieldClass());
	}
	
	
	public void generateClass(File dir) throws Exception {
		
		makeDir(dir,getFieldClass());
		addNoAspects(getFieldClass()); 
		
		File fieldClasFile = new File(dir.getPath()+File.separator + getFieldClass().getName().replace('.','/') + ".class");
		
		FileOutputStream stream = new FileOutputStream(fieldClasFile);
		getFieldClass().stopPruning(true);
		stream.write(getFieldClass().toBytecode());
		stream.close();
		getFieldClass().defrost();
		
		TransformerConstants.registerClassFile(fieldClasFile);
		
		initClassLoader(loader);
	}
	
	public boolean startProcessing(CtClass clz, String name) throws Exception {
		
		//has to be before process methods!!!! otherwise null
		generateCallbackClassField();
		super.startProcessing(clz,name);
		generateStaticInit();
		
		return true;
	}
	
	protected void generateStaticInit() throws Exception{
		
		String code = "public static void init(ClassLoader loader) {"+initCode.toString()+"}";
		CtMethod method= CtNewMethod.make(code,getFieldClass());
		getFieldClass().addMethod(method);
		
		if(getClassLoader()==null||getClassLoader()==ClassLoader.getSystemClassLoader()) {
			
			CtConstructor c = getFieldClass().makeClassInitializer();
			c.setBody("init(ClassLoader.getSystemClassLoader());");
		
		}
	}
	
	protected String getMethodString(CtMethod theMethod, CtClass javaClass, String newName) throws Exception {	
		
		boolean fireMethod=false;
		String start = EventTransformer.getFireMethod(newName,javaClass);
		String res = "";
		if(start==null) {
			res+=getSignatureSourceName(theMethod.getReturnType());
			res+=" "+javaClass.getName()+"."+newName;
		} else res=start;
		res+="(";
		CtClass[] params = theMethod.getParameterTypes();
		for(int j=0; j<params.length; ++j)   {
			res+=getSignatureSourceName(params[j]);
			if(j<params.length-1)
				res+=", ";
		}
		res+=")";
		
			
		return res;	
	}
	
	protected String getMethodString(CtMethod theMethod, CtClass javaClass) throws Exception {	
		
		return getMethodString(theMethod,javaClass,theMethod.getName());
	}
}