/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Iterator;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtMethod;

public class TransformerConstants  {
	
	public static final String CALL_DES =  "_Jasco_CALL_NEW_";

	public static String getCallbackClass(long theID)  {
	
		return "CallbackClass"+theID;
	}
	
	public static String getCallbackClassFullname(CtClass clazz, long id)  {
	
		return clazz.getName()+"$"+getCallbackClass(id);
	}
	
	protected static String getClallBackObject(ClassProcessor clazz, long id, CtMethod method)  {
	
		return clazz.getFieldClass().getName()+".INSTANCE";
	}
	
	public static String getMethodField(long methodid)  {
	
		return "_Jasco_JascoMethod"+methodid;
	}
	
	public static String getCallMethodField(long methodid)  {
		
			return "_Jasco_JascoMethodCALL"+methodid;
		}
	
	public static String getTrapIDField(long methodid)  {
	
		return "_Jasco_JascoMethod"+methodid+"ID";
	}
	
	public static String getNewMethodName(String method, String clz)  {
		
		if(isForCall(method)) {
			String methodname = method.substring(CALL_DES.length());
			return "_Jasco_NEW_"+methodname;
		}
		else return "_Jasco_NEW_"+clz+method;
	}
	
	public static String getNewCallMethodName(String method, String clz)  {
		
			return CALL_DES+clz+method;
	}
	
	public static boolean addedByJAsCo(CtMethod method) {
		
		return method.getName().startsWith("_Jasco");
	}
	public static boolean isForCall(String methodname) {
		
		return methodname.startsWith(CALL_DES);
	}
	
	public static String getNewCallMethodName(CtMethod method)  {
		return TransformerConstants.getNewCallMethodName(method.getName(),method.getDeclaringClass().getSimpleName());
	}
	
	public static String getNewMethodName(CtMethod method)  {
		return TransformerConstants.getNewMethodName(method.getName(),method.getDeclaringClass().getSimpleName());
	}
	
	public static String getCallbackPerMethodName(CtMethod method)  throws Exception {
	
		return "jasco.runtime.callback."+method.getDeclaringClass().getName()+"_"+method.getName()+"_"+method.getParameterTypes().length;
	}
	
	public static String getClallBackPerMethodObject(CtMethod method) throws Exception  {
	
		return getCallbackPerMethodName(method)+"#INSTANCE";
	}
	
	private static Vector classFiles = new Vector();
	
	public static synchronized void registerClassFile(File f)  {
		
		classFiles.add(f);
	}
	
	public static synchronized void unregisterAllTempFiles() {
		
		classFiles.clear();
	}
	
	
	public static synchronized void deleteClassFiles()  {
		
		Iterator i = classFiles.iterator();
		while (i.hasNext())  {
			File f = (File) i.next();
			if(!f.delete())
				;//System.out.println("FAILED DELETING: "+f.getAbsolutePath());;
		}
		
		File f = jasco.options.Options.getTempDir();
		deleteEmptyDirs(f);
	}
	
	private static void deleteEmptyDirs(File dir) {
		// This is a rather dangerous method
		// FIXME it shouldn't be applied in dirs JAsCo has not generated itself
		// In the meantime, prevent it from screwing up an svn working copy:
		
		if(dir.isHidden())
			return;
		
		if(!dir.isDirectory())
			return;
		try{
		if(dir.getCanonicalPath().contains(".svn"))
			return;
		}catch(Exception ex){}
		File[] files = dir.listFiles();
		if(files==null||files.length==0) {
			//System.out.println("DEL"+dir.getAbsolutePath());
			dir.delete();
			return;
		}
		
		//System.out.println(dir.getAbsolutePath()+files.length);
		for(int i=0;i<files.length;++i) {
			deleteEmptyDirs(files[i]);
		}
		
		files = dir.listFiles();
		if(files==null||files.length==0) {
			//System.out.println("DEL"+dir.getAbsolutePath());
			dir.delete();
		}
	}
	
	private static int id=0;
	
	public static synchronized void moveClassFiles(File dest)  {
		
		Iterator i = classFiles.iterator();
		while (i.hasNext())  {
			try {
			File f = (File) i.next();
			
			String name = f.getName();
			
			
			FileInputStream str = new FileInputStream(f);
			BufferedInputStream str2 = new BufferedInputStream(str);
			File newDest = new File(dest,name);
			while(newDest.exists())
				newDest=new File(dest,++id+name);
			
			FileOutputStream output = new FileOutputStream(newDest);
			BufferedOutputStream output2 = new BufferedOutputStream(output);
			
			byte[] buffer = new byte[1024];
			while(true) {
				int length = str2.read(buffer);
				if(length<=0)
					break;
				output2.write(buffer,0,length);
			}
			
			str2.close();
			output2.close();
			 
			}catch(java.io.IOException ex){
				jasco.util.logging.Logger.getInstance().showError(ex);
			}
		} 
		
		TransformerConstants.deleteClassFiles();
	}

}