/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 20-okt-2004
 */
package jasco.testing;

import jasco.options.Options;
import jasco.util.logging.Logger;

import java.io.File;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import junit.framework.Test;
import junit.framework.TestResult;

/**
 * @author Wim
 */

//TODO add following tests: run system with all compiled connectors, adapt runJuggler to add/remove deleter connector
public class JAsCoTestMain implements Test {
	
	private static Vector tests = new Vector();
	private static HashMap depends = new HashMap();
	
	private static boolean redoTests = false;
	private static Vector doneTests = new Vector();
	
	private static final String pacoSuiteDirString = "jasco.testing.pacosuitedir";
	private static final String benchBaseResultString = "jasco.testing.benchbaseresult";
	
	private static String pacoSuiteDir = null;
	private static int benchBaseResult = RunBenchTest.DONT_CHECK;
	
	public static Test suite() {
		emptyFolder(new File(DefaultJAsCoTest.getOutputDir()),false);
		redoTests = false;
		initOptions();
		initTests();
		
		return new JAsCoTestMain();
	}
	
	public static void initOptions() {
		
		String tmp = System.getProperty(pacoSuiteDirString);
		if(tmp!=null&&!tmp.equals("")) {
			File f = new File(tmp);
			if(!f.exists())
				Logger.getInstance().showWarning("Illegal directory for PacoSuite");
			else pacoSuiteDir=tmp;
		}
		
		tmp = System.getProperty(benchBaseResultString);
		
		if(tmp!=null&&!tmp.equals("")) {
			try {
				int tmpint  = new Integer(tmp).intValue();
				benchBaseResult = tmpint;
			}
			catch(NumberFormatException ex) {
				Logger.getInstance().showWarning("Illegal bench base result");
			}
		}
	}
	
	public int countTestCases() {
		int count= 0;
		
		Iterator i = tests.iterator();
		while(i.hasNext()) {
			Test test= (Test)i.next();
			count += test.countTestCases();
		}
		Vector counted = new Vector();
		counted.addAll(tests);
		i = depends.entrySet().iterator();
		while(i.hasNext()) {
			Map.Entry entry = (Map.Entry)i.next();
			Vector v = (Vector) entry.getValue();
			Iterator j = v.iterator();
			while(j.hasNext()) {
				Test test = (Test)j.next();
				if(!counted.contains(test)) { //not already counted
					count += test.countTestCases();
					counted.add(test);
				}
			}
		}
		return count;
	}
	
	public void run(TestResult result) {
		Iterator i = tests.iterator();
		while(i.hasNext()) {
			if (result.shouldStop())
				break;
			executeJunit((IJAsCoTest)i.next(),result,false);
		}
	}

	protected void executeJunit(IJAsCoTest test, TestResult result, boolean alwaysExecute) {
		
		if(!alwaysExecute&&test.succeededPreviously()&&!redoTests)
			return;
		
		if(doneTests.contains(test))
			return;
		
		Iterator depends = getDepends(test);
		while(depends.hasNext()) {
			IJAsCoTest dependsTest = (IJAsCoTest) depends.next();
			executeJunit(dependsTest,result,true);
		}
		test.run(result);
		doneTests.add(test);
	}
	
	public static void main(String[] args) throws Exception{
		
		Logger.getInstance().showOutput("Starting JAsCo testing framework "+Options.getVersion());
		
		String logdir = DefaultJAsCoTest.getLogDir();
		Logger.getInstance().showOutput("Logs are generated in "+logdir);
		
		StopThread stopThread = new StopThread();
		stopThread.start();
		Thread.sleep(10);
		Logger.getInstance().showOutputNoNewLine("Press 'r' within three seconds to redo all tests instead of failed tests "); 
		Thread.sleep(1000);
		Logger.getInstance().showOutputNoNewLine(".");
		Thread.sleep(1000);
		Logger.getInstance().showOutputNoNewLine(".");
		Thread.sleep(1000);
		Logger.getInstance().showOutput(".");
		Thread.sleep(1000);
		
		if(redoTests) {
			Logger.getInstance().showOutput("Redoing all tests");
		}
		
		Logger.getInstance().showOutput("Cleaning temp dir: "+DefaultJAsCoTest.getOutputDir());
		emptyFolder(new File(DefaultJAsCoTest.getOutputDir()),false);
		
		Logger.getInstance().showOutput("Press 'q' at any time to quit"); 
		
		Logger.getInstance().showOutputNoNewLine("Preparing tests ..... ");
		initTests();
		Logger.getInstance().showOutput("done");
		Logger.getInstance().showOutput("");
		
		int failed = 0;
		int succes = 0;
		java.util.Iterator i = tests.iterator();
		while(i.hasNext()) {
			boolean b = executeTest((IJAsCoTest)i.next(),false);
			if(b)
				succes++;
			else failed++;
		}
		
		
		Logger.getInstance().showOutput("");
		if(failed==0) {
			Logger.getInstance().showOutput("DONE; All tests succeeded!");
			Logger.getInstance().showOutput("Cleaning temp dir: "+DefaultJAsCoTest.getOutputDir());
			emptyFolder(new File(DefaultJAsCoTest.getOutputDir()),false);
		}
		
		else Logger.getInstance().showOutput("DONE; "+succes+" tests succeeded, but "+failed+" tests failed.");
	}
	
	protected static boolean executeTest(IJAsCoTest test, boolean alwaysExecute) {
		
		if(!alwaysExecute&&test.succeededPreviously()&&!redoTests)
			return true;
		
		if(doneTests.contains(test))
			return true;
		
		Iterator depends = getDepends(test);
		while(depends.hasNext()) {
			IJAsCoTest dependsTest = (IJAsCoTest) depends.next();
			executeTest(dependsTest,true);
		}
		
		boolean  b = test.startTest();
		doneTests.add(test);
		return b;
	}
	
	protected static void emptyFolder(File dir) {
		emptyFolder(dir,true);
	}
	
	protected static void emptyFolder(File dir, boolean deleteDir)  {
		
			String[] files = dir.list();
			if(files==null)
				return;
			for(int i=0; i<files.length; ++i)  {
				File f = new File(dir,files[i]);
				if (f.isDirectory()) 
					emptyFolder(f);
				f.delete();
			}
			if(deleteDir)
				dir.delete();
	}
	
	public static void initTests() {
		
		tests.clear();
		
		CompileAspectTest testa1 = new CompileAspectTest("basic.asp");
		tests.add(testa1);
		
		CompileAspectTest testa2 = new CompileAspectTest("tester.asp");
		testa2.setTarget("1.5");
		tests.add(testa2);
		
		CompileLoggingAspectTest testa3 = new CompileLoggingAspectTest();
		tests.add(testa3);
		
		CompileAspectTest testa4 = new CompileAspectTest("extended.asp");
		addDepend(testa4,testa1);
		tests.add(testa4);
		
		CompileAspectTest testa5 = new CompileAspectTest("cflowClass.asp");
		tests.add(testa5);
		
		CompileAspectTest testa6 = new CompileAspectTest("deleter.asp");
		tests.add(testa6);
		
		CompileAspectTest testa16 = new CompileAspectTest("Stateful.asp");
		tests.add(testa16);
		
		CompileAspectTest testa17 = new CompileAspectTest("TargetTest.asp");
		tests.add(testa17);
		
		CompileAspectTest testa18 = new CompileAspectTest("AfterReturningTest.asp");
		tests.add(testa18);
		
		
		CompileConnectorTest testc3 = new CompileConnectorTest("aspectonaspecttracing.con");
		addDepend(testc3,testa3);
		tests.add(testc3);
	
		CompileConnectorTest testc30 = new CompileConnectorTest("extendedConnector.con");
		addDepend(testc30,testa4);
		tests.add(testc30);

		
		
		CompileTraversalTest testt1 = new CompileTraversalTest("TestTraversal.trv");
		addDepend(testt1,testa1);
		tests.add(testt1);
		
		SimpleTransformTest testtransform = new SimpleTransformTest("java.awt.Component");
		testtransform.addOption("-noevents");
		tests.add(testtransform);
		
		SimpleTransformTest testtransform2 = new SimpleTransformTest("javax.swing.JComponent");
		testtransform2.addOption("-A");
		tests.add(testtransform2);
		
		//run juggler tests.... 
		Vector dependsRunJuggler = new Vector();
		RemoveConnectors removeConnectorsJuggler = new RemoveConnectors();
		
		CompileConnectorTest testc1 = new CompileConnectorTest("basicConnector.con");
		addDepend(testc1,testa1);
		
		CompileConnectorTest testc123 = new CompileConnectorTest("subtypeConnector.con");
		addDepend(testc123,testa1);
		
		CompileConnectorTest testc4 = new CompileConnectorTest("buttonDelConnector.con");
		testc4.addOption("-target:1.5");
		addDepend(testc4,testa6);
		
		
		CompileConnectorTest testc9 = new CompileConnectorTest("basic2Connector.con");
		addDepend(testc9,testa2);
		addDepend(testc9,testa5);
		
		
		CompileConnectorTest testc11 = new CompileConnectorTest("basicConnectorArray.con");
		addDepend(testc11,testa1);
		
		
		CompileConnectorTest testc13 = new CompileConnectorTest("TargetConnector.con");
		addDepend(testc13,testa17);
		
		
		CompileConnectorTest testc14 = new CompileConnectorTest("AfterReturningConnector.con");
		addDepend(testc14,testa18);
		
		
		dependsRunJuggler.add(removeConnectorsJuggler);
		dependsRunJuggler.add(testc4);
		dependsRunJuggler.add(testc1);
		dependsRunJuggler.add(testc9);
		dependsRunJuggler.add(testc14);
		dependsRunJuggler.add(testc13);
		dependsRunJuggler.add(testc11);
		dependsRunJuggler.add(testc123);
		

		RunJugglerTest runjugglerbench1 = new RunJugglerTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(runjugglerbench1,dependsRunJuggler);
		tests.add(runjugglerbench1);
		
		RunJugglerTest runjugglerbench2 = new RunJugglerTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(runjugglerbench2,dependsRunJuggler);
		tests.add(runjugglerbench2);
		
//		prio tests
		CompileConnectorTest testc10 = new CompileConnectorTest("basicConnectorCopy.con");
		addDepend(testc10,testa1);
		tests.add(testc10);
		Vector prioDepend = new Vector();
		prioDepend.add(testc10);
		prioDepend.add(testc1);
		
		RunConnectorPriorityTest runprio1 = new RunConnectorPriorityTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(runprio1,prioDepend);
		tests.add(runprio1);
		
		RunConnectorPriorityTest runprio2 = new RunConnectorPriorityTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(runprio2,prioDepend);
		tests.add(runprio2);
		
		RunConnectorPriorityTest runprio3 = new RunConnectorPriorityTest(RunPacoSuiteTest.NO_JUTTA);
		addDepend(runprio3,prioDepend);
		tests.add(runprio3);
		
		
				
		//meta data tests
		Vector dependsMetaData = new Vector();
		RemoveConnectors removeConnectorsMetaData = new RemoveConnectors();
		
		
		CompileConnectorTest testc12 = new CompileConnectorTest("MetaDataConnector.con");
		addDepend(testc12,testa1);
		dependsMetaData.add(removeConnectorsMetaData);
		dependsMetaData.add(testc12);
		
		RunMetaDataTest metadata1 = new RunMetaDataTest(RunPacoSuiteTest.HOTSWAP1);
		addDepend(metadata1,dependsMetaData);
		tests.add(metadata1);
		
		
		RunMetaDataTest metadata2 = new RunMetaDataTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(metadata2,dependsMetaData);
		tests.add(metadata2);
		
		RunMetaDataTest metadata3 = new RunMetaDataTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(metadata3,dependsMetaData);
		tests.add(metadata3);
		 
		
		tests.add(testc12);

		
		//transforms juggler and JButton, has to be after run juggler tests...
		CompileTransformConnectorTest testc2 = new CompileTransformConnectorTest();
		addDepend(testc2,testa2);
		tests.add(testc2);
		
		//benchmark tests
		{
		RemoveConnectors removeConnectors1 = new RemoveConnectors();
		
		CompileAspectTest testb1 = new CompileAspectTest("VoidAspect.asp");
		
		CompileConnectorTest testb2 = new CompileConnectorTest("BenchConnector.con");
		
		Vector dependsRun = new Vector();
		dependsRun.add(removeConnectors1);
		dependsRun.add(testb1);
		dependsRun.add(testb2);
		
		RunJAsCoProgramTest runbench2 = new RunBenchTest(RunPacoSuiteTest.HOTSWAP1, 400);
	    addDepend(runbench2,dependsRun);
		tests.add(runbench2);
		
		RunJAsCoProgramTest runbench3 = new RunBenchTest(RunPacoSuiteTest.HOTSWAP2, 350);
		addDepend(runbench3,dependsRun);
		tests.add(runbench3);
		
		RunJAsCoProgramTest runbench4 = new RunBenchTest(RunPacoSuiteTest.RUNTIME_WEAVER, 175);
		addDepend(runbench4,dependsRun);
		tests.add(runbench4);
	}
		//counting tests
		{
		RemoveConnectors removeConnectors4 = new RemoveConnectors();
		
		CompileAspectTest testcount1 = new CompileAspectTest("Counting.asp");
		
		
		CompileConnectorTest testcount2 = new CompileConnectorTest("CountingConnector.con");
		
		
		Vector dependsRunCount = new Vector();
		dependsRunCount.add(removeConnectors4);
		dependsRunCount.add(testcount1);
		dependsRunCount.add(testcount2);
		
		RunJAsCoProgramTest runcount1 = new RunCountTest(RunPacoSuiteTest.NO_JUTTA);
	    addDepend(runcount1,dependsRunCount);
		tests.add(runcount1);
		
		RunJAsCoProgramTest runcount2 = new RunCountTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(runcount2,dependsRunCount);
		tests.add(runcount2);
		
		RunJAsCoProgramTest runcount3 = new RunCountTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(runcount3,dependsRunCount);
		tests.add(runcount3);
	}
		{	
		// after throwing/returning tests
		Vector dependsThrowingReturning = new Vector();
		RemoveConnectors removeConnectorsThrowingReturning = new RemoveConnectors();
		
		CompileAspectTest testthrowingret1 = new CompileAspectTest("Around.asp");
		CompileConnectorTest testthrowingret2 = new CompileConnectorTest("aroundRTConnector.con");
		
		dependsThrowingReturning.add(removeConnectorsThrowingReturning);
		dependsThrowingReturning.add(testthrowingret1);
		dependsThrowingReturning.add(testthrowingret2);
		
		
		RunJAsCoProgramTest runthrowingret1 = new RunAroundThrowingReturningTest(RunPacoSuiteTest.NO_JUTTA);
	    addDepend(runthrowingret1,dependsThrowingReturning);
		tests.add(runthrowingret1);
		
		RunJAsCoProgramTest runthrowingret2 = new RunAroundThrowingReturningTest(RunPacoSuiteTest.HOTSWAP2);
		addDepend(runthrowingret2,dependsThrowingReturning);
		tests.add(runthrowingret2);
		
		RunJAsCoProgramTest runthrowingret3 = new RunAroundThrowingReturningTest(RunPacoSuiteTest.RUNTIME_WEAVER);
		addDepend(runthrowingret3,dependsThrowingReturning);
		tests.add(runthrowingret3);
		}
		
		{	
			// subtype test...
			Vector dependsJumping = new Vector();
			RemoveConnectors removeConnectorsJumping = new RemoveConnectors();
			
			CompileAspectTest testjumping1 = new CompileAspectTest("deleter.asp");
			CompileConnectorTest testjumping2 = new CompileConnectorTest("subtypeDelConnector.con");
			
			dependsJumping.add(removeConnectorsJumping);
			dependsJumping.add(testjumping1);
			dependsJumping.add(testjumping2);
			
			RunJAsCoProgramTest runjumping = new RunSubtypeTest(RunPacoSuiteTest.RUNTIME_WEAVER);
			addDepend(runjumping,dependsJumping);
			tests.add(runjumping);
			
			RunJAsCoProgramTest runjumping2 = new RunSubtypeTest(RunPacoSuiteTest.HOTSWAP2);
			addDepend(runjumping2,dependsJumping);
			tests.add(runjumping2);
			
			RunJAsCoProgramTest runjumping3 = new RunSubtypeTest(RunPacoSuiteTest.HOTSWAP1);
			addDepend(runjumping3,dependsJumping);
			tests.add(runjumping3);
			}
		
		{	
			// jumping aspects tests...
			Vector dependsJumping = new Vector();
			RemoveConnectors removeConnectorsJumping = new RemoveConnectors();
			
			CompileAspectTest testjumping1 = new CompileAspectTest("Jumping.asp");
			CompileConnectorTest testjumping2 = new CompileConnectorTest("JumpingLogging.con");
			
			dependsJumping.add(removeConnectorsJumping);
			dependsJumping.add(testjumping1);
			dependsJumping.add(testjumping2);
			
			RunJAsCoProgramTest runjumping = new RunJumpingAspectsTest();
			addDepend(runjumping,dependsJumping);
			tests.add(runjumping);
			}
		{	
			//aspects on aspects test
			Vector dependsAonA = new Vector();
			RemoveConnectors removeConnectorsAonA = new RemoveConnectors();
			
			CompileAspectTest testAonA1 = new CompileAspectTest("logging.asp");
			CompileConnectorTest testAonA2 = new CompileConnectorTest("JumpingTestLoggingConnector.con");
			addDepend(testAonA2,testa1);
			CompileConnectorTest testAonA3 = new CompileConnectorTest("aspectonaspecttracing.con");
			addDepend(testAonA3,testAonA1);
			addDepend(testAonA3,testa1);
			
			dependsAonA.add(removeConnectorsAonA);
			dependsAonA.add(testAonA2);
			dependsAonA.add(testAonA3);
			
//			test aspects on aspects trap adding at runtime:
			/* tests fail because aspects are loaded by boot classpath loader and are thus not subject for hotswapping!
			 * 
			 * RunJAsCoProgramTest runAonA1 = new RunAspectOnAspectTest(RunJAsCoProgramTest.HOTSWAP2,true);
			addDepend(runAonA1,dependsAonA);
			tests.add(runAonA1);
			
			RunJAsCoProgramTest runAonA2 = new RunAspectOnAspectTest(RunJAsCoProgramTest.RUNTIME_WEAVER,true);
			addDepend(runAonA2,dependsAonA);
			tests.add(runAonA2);*/
			
//			test aspects on aspects trap adding at compile time:
			CompileAspectTest testAonA0 = new CompileAspectTest("basic.asp");
			testAonA0.addOption("-X");
			
			CompileConnectorTest testAonA5 = new CompileConnectorTest("JumpingTestLoggingConnector.con");
			addDepend(testAonA5,testAonA0);
			CompileConnectorTest testAonA4 = new CompileConnectorTest("aspectonaspecttracing.con");
			addDepend(testAonA4,testAonA1);
			addDepend(testAonA4,testAonA0);
	
			dependsAonA.clear();
			dependsAonA.add(new RemoveConnectors());
			dependsAonA.add(testAonA4);
			dependsAonA.add(testAonA5);
			
			RunJAsCoProgramTest runAonA3 = new RunAspectOnAspectTest(RunJAsCoProgramTest.HOTSWAP2,false);
			addDepend(runAonA3,dependsAonA);
			tests.add(runAonA3);
			
			RunJAsCoProgramTest runAonA4 = new RunAspectOnAspectTest(RunJAsCoProgramTest.RUNTIME_WEAVER,false);
			addDepend(runAonA4,dependsAonA);
			tests.add(runAonA4);
			}
		
//		run-time tests
		
		if(getPacoSuiteDir()!=null)  {//no pacosuite tests
			RemoveConnectors removeConnectors2 = new RemoveConnectors();
			
			TransformBeanTest testtransform3 = new TransformBeanTest();
			tests.add(testtransform3);
		
			RunPacoSuiteTest paco1 = new RunPacoSuiteTest(RunPacoSuiteTest.TRANSFORM_BEAN);
			addDepend(paco1,removeConnectors2);
			addDepend(paco1,testtransform3);
			tests.add(paco1);
		
			RunPacoSuiteTest paco2 = new RunPacoSuiteTest(RunPacoSuiteTest.HOTSWAP1);
			addDepend(paco2,removeConnectors2);
			tests.add(paco2);
		
			RunPacoSuiteTest paco3 = new RunPacoSuiteTest(RunPacoSuiteTest.HOTSWAP2);
			addDepend(paco3,removeConnectors2);
			tests.add(paco3);
		
			RunPacoSuiteTest paco4 = new RunPacoSuiteTest(RunPacoSuiteTest.RUNTIME_WEAVER);
			addDepend(paco1,removeConnectors2);
			tests.add(paco4);

			//			app server tests
			RunPacoSuiteTest paco5 = new RunPacoSuiteTest(RunPacoSuiteTest.HOTSWAP2);
			addDepend(paco5,removeConnectors2);
			paco5.addOption("-Djasco.hotswap.applicationserver=true");
			tests.add(paco5); 
		} else Logger.getInstance().showWarning("jasco.testing.pacosuiteDir not set: skipping PaocSuite tests");
		
		
		
		RunJAsCoInterfaceTest interfaceTest = new RunJAsCoInterfaceTest();
		RemoveConnectors removeConnectorsI = new RemoveConnectors();
		addDepend(interfaceTest,removeConnectorsI);
		tests.add(interfaceTest);
		
		 
		//TODO: comb strat and precedence test
		//TODO: test heuristsics of inline compiler
	}
	
	public static String getPacoSuiteDir() {
		
		return pacoSuiteDir;
	}
	
	public static int getBenchBaseResult() {
		
		return benchBaseResult;
	}
	
	public static void addDepend(IJAsCoTest test, Vector dependsOn) {
		
		 Vector v = (Vector) depends.get(test);
		 if(v==null) {
		 	v=new Vector();
		 	depends.put(test,v);
		 }
		v.addAll(dependsOn);
	}
	
	public static void addDepend(IJAsCoTest test, IJAsCoTest dependsOn) {
		
		 Vector v = (Vector) depends.get(test);
		 if(v==null) {
		 	v=new Vector();
		 	depends.put(test,v);
		 }
		 v.add(dependsOn);
	}
	
	public static Iterator getDepends(IJAsCoTest test) {
		
		Vector v =  (Vector) depends.get(test);
		if(v==null) v=new Vector();
		
		return v.iterator();
	}
	
	static class StopThread extends Thread {
		
		StopThread() {
			setDaemon(true);
		}
		
		public void run() {
			try {
			InputStream is = System.in;
			while(true) {
				int i = is.read();
				if(i==-1)
					break;
				if(i=='q') {
					Logger.getInstance().showOutput("User intervention detected, halting processes .....");
					System.exit(0);
				}
				if(i=='r') {
					redoTests=true;
				}
					
			
			}
			}catch(Exception ex) {}
		}
	}
}
