/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 25-okt-2004
 */
package jasco.testing;

import jasco.options.Options;

import java.io.File;

/**
 * @author Wim
 */
public class RunJAsCoProgramTest extends DefaultJAsCoTest {

	
	public RunJAsCoProgramTest(String mainClass, String descr, int type, boolean debug) {
		super("illegal desctiption", "java ");
		
		String output = getOutputDir();
		String desc = "Running "+descr+" ";
		String orig_classpath = System.getProperty("java.class.path");
		String classpath = "\""+orig_classpath+Options.PATH_SEPARATOR+"."+
			Options.PATH_SEPARATOR+Options.getLibPaths()+output+
			Options.PATH_SEPARATOR+ new File(getWorkingDir(),"classes").getAbsolutePath()+
			Options.PATH_SEPARATOR+ new File(getWorkingDir(),"testing-libs.jar").getAbsolutePath()+
			"\"";
		
		switch(type) {
		
		case TRANSFORM_BEAN:
			desc+="TransformBean";
			break;
		
		case HOTSWAP1:
			desc+="HotSwap1";
			mainClass="jasco.HotSwap "+mainClass;
			break;
			
		case NO_JUTTA:
			desc+="NoJutta ";
			addOption("-Djasco.jutta=false");
			
		case HOTSWAP2:
			desc+="HotSwap2";
			addOption("-javaagent:"+new File(Options.getInstallPath(),"jasco.jar").getAbsolutePath());
			break;
			
		case RUNTIME_WEAVER:
			desc+="Run-time Weaver";
			addOption("-javaagent:"+new File(Options.getInstallPath(),"jasco.jar").getAbsolutePath());
			addOption("-Djasco.hotswap.inlinecompiler=true");
			break;
			
		default:
			try {
			getGlobalLog().write("Error illegal weaver type: "+type);
			} catch(Exception ex){}
		}
		
		if(debug) {
			addOption("-Djasco.debug=true");
			addOption("-Djasco.keep.tempfiles=true");
		}
			 
		addOption("-Djasco.installpath=\""+jasco.options.Options.getInstallPath()+"\"");
		addOption("-classpath "+classpath);
		addOption(mainClass);
		setDescription(desc);
		
	}
	

	/**
	 * Checks whether the given array of strings all occur in the StringBuffer f. The array of strings have 
	 * to occur in the order they occur in the array.
	 * 
	 * @param f target string (haystack)
	 * @param shouldoccur search strings (needles) 
	 * @return
	 */
	
	public boolean checkOutputShouldOccur(StringBuffer f, String[] shouldoccur) {
		
		return checkOutputShouldOccur(f,shouldoccur,true);
	}
	/**
	 * Checks whether the given array of strings all occur in the StringBuffer f.
	 * 
	 * @param f target string (haystack)
	 * @param shouldoccur search strings (needles)
	 * @param ordered shouldoccur strings need to occur in the order specified or not 
	 * @return
	 */
	public boolean checkOutputShouldOccur(StringBuffer f, String[] shouldoccur, boolean ordered) {
		
		String res = f.toString();
		for(int i=0;i<shouldoccur.length;++i) {
			int j = res.indexOf(shouldoccur[i]);
			if(j==-1)
				return false;
			
			if(ordered)
				res=res.substring(j+1);
		}
		
		return true;
	}
	
	/**
	 * Checks whether the given array of strings all NOT occur in the StringBuffer f.
	 * 
	 * @param f target string (haystack)
	 * @param shouldnotoccur search strings that cannot occur in the target string f (needles) 
	 * @return
	 */
	public boolean checkOutputShouldNotOccur(StringBuffer f, String[] shouldnotoccur) {
		
		String res = f.toString();
		for(int i=0;i<shouldnotoccur.length;++i) {
			int j = res.indexOf(shouldnotoccur[i]);
			if(j!=-1)
				return false;
		}
		
		return true;
	}

}
