/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.aspectparser;

import jasco.options.Options;
import jasco.runtime.aspect.ConstructorMethod;
import jasco.runtime.aspect.HookContextProperty;
import jasco.runtime.aspect.PCutpointConstructorApplicationDesignator;
import jasco.runtime.aspect.PCutpointConstructorCall;
import jasco.runtime.aspect.PCutpointConstructorExecute;
import jasco.runtime.aspect.PCutpointConstructorRegular;
import jasco.runtime.aspect.PCutpointConstructorTarget;
import jasco.runtime.aspect.Transition;
import jasco.tools.jascoparser.PImport;
import jasco.tools.jascoparser.PJavaCode;
import jasco.tools.jascoparser.PMethod;
import jasco.util.generators.ClassGenerator;
import jasco.util.generators.GImport;
import jasco.util.generators.GPackage;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;
import jasco.util.generators.VariableGenerator;
import jasco.util.javacompiler.CompileError;
import jasco.util.javacompiler.JavaCompiler;
import jasco.util.logging.Logger;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

public class AspectClassesGenerator 
{
	PAspect aspect = null;
	public static final String CONSTRUCTOR_ARGS =  "_Jasco_constructorArgs";
	public static final String CONSTRUCTOR_REP =  "_Jasco_cutpointconstructorRep";
	public static final String HOOKCONTEXTPROP = "_Jasco_getHookContextPropertyForAdvice";
	
	private Vector lastErrors;
	private String targetObjectType = "java.lang.Object";
	
	public static final String ABSTRACT_METHODS =  "_Jasco_abstractMethods";
	public static final String ABSTRACT_METHODS_IMPL =  "_Jasco_abstractMethodsToImplement";
	public static final String MULTIPLE_ELEMENTS =  "multipleelements";
	
	private static final String CONST_thisJoinPoint = "thisJoinPoint";
	private static final String CONST_thisJoinPointObject = "thisJoinPointObject";
	private static final String CONST_proceed = "proceed";
	private static final String CONST_args = "args";
	
	public AspectClassesGenerator(PAspect aAspect)
	{
		aspect = aAspect;
		
	}
	
	public void generate()  {
	
		try {
			generateImpl();
			
		}
		catch(IllegalArgumentException ex)  {
			Logger.getInstance().showError(ex.getMessage());
		}
	}
	
	protected void generateImpl() 
	{
		
		generateAspect();
		
		
		try 
		{	
			//retrieve the source and classpath
			String classpath = Options.getClassPath();
			String packageDir="";
			if(! (aspect.getPackage().getPackageName().equals("")))
				packageDir=aspect.getPackage().getPackageName().replace('.',File.separatorChar) +Options.FILE_SEPARATOR;
			//String source = (new File(Options.getOutputDir())).getCanonicalPath() + Options.FILE_SEPARATOR + packageDir+"*.java";
			String source = (new File(Options.getOutputDir())).getCanonicalPath() + Options.FILE_SEPARATOR + packageDir+aspect.getAspectName()+".java";

			//compile all files
			JavaCompiler compile = new JavaCompiler();
		
			compile.compile(source,classpath);
			Vector errors = compile.getErrors();
			lastErrors=errors;
			if(Options.deleteTempFiles())
				new File(source).delete();
		
			//translate the errors to the aspect source instead of the class source
			translateAndPrint(errors);
		}
		catch(Exception e)
		{
			Logger.getInstance().showFullError(e);
		}
		
		generateCutpoints();
		
		if (Options.transformHooks())   {
			
			String packageName = "";
			if(! (aspect.getPackage().getPackageName().equals("")))
				packageName=aspect.getPackage().getPackageName();
			jasco.Jasco.transformBeanInOutputDir(packageName+"."+aspect.getAspectName(),new Vector());
			transformCutpoints();
		}
	}
	
	public Iterator getLastErrors()  {
		
		if(lastErrors!=null)
			return lastErrors.iterator();
		else return null;	
	}
	
	private boolean translateAndPrint(Vector aErrors)
	{
		String source = aspect.getAspectName() + ".asp";
		
		for (int i = 0 ; i < aErrors.size() ; i++) 
		{
			//Get error and errorstring;
			CompileError error = (CompileError)aErrors.elementAt(i);
			String errorstring = error.getErrorString();
			
		
			
			//Get linenumber of error in original file
			int documentationindex = errorstring.lastIndexOf("//");
		
			if (documentationindex==-1)  {
				//System.out.println("!!!!!!!! JAsCo System Error: !!!!!!!!");
				//System.out.println(error.toString());
				Logger.getInstance().showError(error);
					
				continue;
			}
			//System.out.println("OK"+errorstring);
			int restindex = errorstring.indexOf(Options.NEWLINE,documentationindex);
			String linenumber = errorstring.substring(documentationindex + 2,restindex);
			
			int reallinenumber = (new Integer(linenumber)).intValue();
			
			
			//Get errormessage by extracting documentation
			String realerror = errorstring.substring(0,documentationindex) + "\n" + errorstring.substring(restindex + 1); 
			//System.out.println("OK"+errorstring+reallinenumber);
			CompileError newerror = new CompileError(source,reallinenumber,realerror);
				
			//System.out.print("FULL"+newerror.toString());
			Logger.getInstance().showError(newerror);
		}
		if (aErrors.size() != 0)
		{
			if (aErrors.size() == 1)
			{
				//System.out.println(1 + " error");
				Logger.getInstance().showOutput(1 + " error");
			}
			else 
			{
				//System.out.println(aErrors.size() + " errors");
				Logger.getInstance().showOutput(aErrors.size() + " errors");
			}
			
			return false;
		}
		else return true;
	}
	
	public void generateAspect() 
	{
		String aspectname = aspect.getAspectName();
		String packageName = aspect.getPackage().getPackageName();
		
		if(packageName.equals(""))
			System.err.println("Warning: declaring an aspect without a package causes classpath problems when compiling a connector!");
		
		String globalName = aspectname;//+"_global";
	
		ClassGenerator globalClass = new ClassGenerator(globalName,new GPackage(packageName, "" + aspect.getPackage().getLineNumber()));
		
		globalClass.setModifiers(aspect.getModifiers());
		
		globalClass.addImportPackage(new GImport("java.lang.*"));
		globalClass.addImportPackage(new GImport("java.util.*"));
		globalClass.addImportPackage(new GImport("jasco.runtime.aspect.*"));
		globalClass.addImportPackage(new GImport("jasco.runtime.connector.*"));
		globalClass.addImportPackage(new GImport("jasco.util.logging.*"));
		globalClass.addImportPackage(new GImport("jasco.runtime.*"));
		
		//add imported packages
		for (int i = 0 ; i < aspect.getImports().getImportsSize() ; i++) 
		{
			PImport imported = aspect.getImports().getImport(i);
			globalClass.addImportPackage(new GImport(imported.getImportName(),"" + imported.getLineNumber()));
		} 
		
		VariableGenerator hooks = new VariableGenerator("Vector","_Jasco_hooks");
		hooks.addModifier(Modifier.PRIVATE);
		hooks.addModifier(Modifier.STATIC);
		hooks.setInitialiser("new Vector()");
		globalClass.addVariable(hooks);
			
			
		MethodGenerator getCutpoints = new MethodGenerator("_Jasco_getHooks","Vector");
		getCutpoints.addModifier(Modifier.PUBLIC);
		getCutpoints.addModifier(Modifier.STATIC);
		String impCode = "";
		impCode+="return _Jasco_hooks;"+JavaGenerator.NEWLINE;
		getCutpoints.setImplementation(impCode);
		globalClass.addMethod(getCutpoints);
			
		globalClass.addModifier(Modifier.PUBLIC);
		
		aspect.initClassGenerator(globalClass);
		
		if (aspect.getExtendingAspectName().equals(""))  
		{ 	
			String staticHooksInit =  "";
			PCutpoints cutpoints = aspect.getCutpoints();
			for (int i = 0; i < cutpoints.getCutpointsSize() ; i++) 
			{
				staticHooksInit+="_Jasco_hooks.addElement(\""+cutpoints.getCutpoint(i).getCutpointName()+"\");"+JavaGenerator.NEWLINE;
			}
			globalClass.addStaticInit(staticHooksInit);
		}
		else  {		
			globalClass.setSuperClass(aspect.getExtendingAspectName());
			String clString = findClass(aspect.getExtendingAspectName(),globalClass);
			if(clString==null)
				throw new IllegalArgumentException("Class not found error: "+aspect.getExtendingAspectName());
			
			Vector v = null;
			try {
			Class cl = Options.loadClass(clString);
			Method getHooks = cl.getMethod("_Jasco_getHooks",new Class[] {});
			v=(Vector)getHooks.invoke(null,new Object[]  {});
			}
			catch(Exception ex)  {
				//System.err.println("!!!!! JASCO SYSTEM ERROR !!!!!!! "+ex.getMessage());
				//ex.printStackTrace();
				//Logger.getInstance().showError("!!!!! JASCO SYSTEM ERROR !!!!!!! ");
				Logger.getInstance().showFullError(ex);
				throw new IllegalArgumentException("!!!!! JASCO SYSTEM ERROR !!!!!!!");
			}
			Vector cp = (Vector)v.clone();
			Vector cp2 = (Vector)v.clone();
			PCutpoints cutpoints = aspect.getCutpoints();
			for (int i = 0; i < cutpoints.getCutpointsSize() ; i++) 
			{
				
				String ext = cutpoints.getCutpoint(i).getExtendingCutpointName();
				String cpname = cutpoints.getCutpoint(i).getCutpointName();
				if(!cp.contains(cpname)) 
					cp.add(cpname);
					
				if (ext!=null)  {
					cp.remove(ext);
					cp2.remove(ext);
				}
				else  {
					cp2.remove(cpname);
				}
				
				
			}
			String staticHooksInit =  "";
			
			Iterator i = cp.iterator();
			while (i.hasNext())  {
				String next = (String) i.next();
				staticHooksInit+="_Jasco_hooks.addElement(\""+next+"\");"+JavaGenerator.NEWLINE;
			}
			globalClass.addStaticInit(staticHooksInit);
			
			generateDummyHooks(cp2);
			
		}
			

		
		try  {
			
			String packageDir = "";
			if(!packageName.equals(""))
				packageDir=packageName.replace('.','/')+"/";
			File filetowrite = new File(Options.getOutputDir() + "/"+packageDir+globalName+".java");
			if(filetowrite.exists())
				filetowrite.delete();
			else if(filetowrite.getParent()!=null)
				new File(filetowrite.getParent()).mkdirs();	
			BufferedWriter out = new BufferedWriter(new FileWriter(filetowrite));
			out.write(globalClass.toString());
			out.close();
		}
		catch(Exception e)
		{	
			//System.err.println(e.getMessage()); e.printStackTrace(); 
			Logger.getInstance().showFullError(e);
		}
		
		
	}
	
	protected void generateDummyHooks(Vector hooks)  {
		
		try  {
			
			
		Iterator i = hooks.iterator();
		while (i.hasNext())  {
			String hook = (String)i.next();
			String globalName = aspect.getAspectName()+"_"+hook;
			String packageName = aspect.getPackage().getPackageName();
			
			PCutpoint cutpoint = new PCutpoint();
			cutpoint.setCutpointName(hook);
			cutpoint.setExtendingCutpointName(hook);
			
			aspect.addCutpoint(cutpoint);
			
			/*
			ClassGenerator dummy = new ClassGenerator(globalName,new GPackage(packageName, "" + aspect.getPackage().getLineNumber()));
			dummy.setSuperClass(aspect.getExtendingAspectName()+"_"+hook);
			
			
			File filetowrite = new File(Options.getOutputDir() + "/"+packageName.replace('.','/')+"/"+globalName+".java");
			if(filetowrite.exists())
				filetowrite.delete();
			else if(filetowrite.getParent()!=null)
				new File(filetowrite.getParent()).mkdirs();	
			BufferedWriter out = new BufferedWriter(new FileWriter(filetowrite));
			out.write(dummy.toString());
			out.close();
			*/
		}
		}
		catch(Exception e)
		{	
			//System.err.println(e.getMessage()); e.printStackTrace(); 
			Logger.getInstance().showFullError(e);
		}
	}

	public void generateCutpoints()
	{
		PCutpoints cutpoints = aspect.getCutpoints();
		for (int i = 0; i < cutpoints.getCutpointsSize() ; i++) 
		{
			generateCutpoint(cutpoints.getCutpoint(i));
		}
		
	
	}
	
	protected void transformCutpoints()  {
		
		PCutpoints cutpoints = aspect.getCutpoints();
		for (int i = 0; i < cutpoints.getCutpointsSize() ; i++) 
		{	
			PCutpoint aCutpoint = cutpoints.getCutpoint(i);
			String aspectname = aspect.getAspectName();
			String packageName = aspect.getPackage().getPackageName();
			String cutpointname = aCutpoint.getCutpointName();
			String cutpointClassName = aspectname+"_"+cutpointname;
			String extendingcutpointname = aCutpoint.getExtendingCutpointName();
			jasco.Jasco.transformBeanInOutputDir(packageName+"."+cutpointClassName,new Vector());
		}
	}
	
	public void generateCutpoint(PCutpoint aCutpoint)
	{
		String aspectname = aspect.getAspectName();
		String packageName = aspect.getPackage().getPackageName();
		String cutpointname = aCutpoint.getCutpointName();
		String cutpointClassName = aspectname+"_"+cutpointname;
		String extendingcutpointname = aCutpoint.getExtendingCutpointName();
		
		String realCutpointName = cutpointClassName;
		if(!packageName.equals(""))
			realCutpointName=packageName+"."+realCutpointName;
		
		ClassGenerator cutpointClass =  new ClassGenerator(cutpointClassName,new GPackage(packageName));
		cutpointClass.addModifier(Modifier.PUBLIC);
		cutpointClass.addImportPackage(new GImport("java.lang.*"));
		cutpointClass.addImportPackage(new GImport("java.util.*"));
		cutpointClass.addImportPackage(new GImport("java.lang.reflect.*"));
		cutpointClass.addImportPackage(new GImport("jasco.runtime.aspect.*"));
		cutpointClass.addImportPackage(new GImport("jasco.runtime.connector.*"));
		cutpointClass.addImportPackage(new GImport("jasco.runtime.*"));
		cutpointClass.addImportPackage(new GImport("jasco.util.logging.*"));
		cutpointClass.addInterface("Cutpoint");
		
		//add imported packages
		for (int i = 0 ; i < aspect.getImports().getImportsSize() ; i++) 
		{
			PImport imported = aspect.getImports().getImport(i);
			cutpointClass.addImportPackage(new GImport(imported.getImportName(),"" + imported.getLineNumber()));
		} 
		Vector genadvices = new Vector();
		if (aCutpoint.hasSuperCutpoint()) 
			{
				String extendedaspectname = aspect.getExtendingAspectName();
				if (extendedaspectname.equals(""))  
					throw new IllegalArgumentException("A hook can not inherit without that its enclosing aspect bean inherits!");
				
				checkNoConsturctor(aCutpoint);
				cutpointClass.setSuperClass(extendedaspectname+"_"+extendingcutpointname);
				
				
				generateConstructor(aCutpoint,cutpointClass);
				generateConstructorField(aCutpoint,cutpointClass);
				
				
				genadvices.addAll(generateBefore(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateAfter(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateReplace(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateAfterThrowing(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateAfterReturning(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateAroundThrowing(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateAroundReturning(aCutpoint,cutpointClass,true));
				genadvices.addAll(generateTestStmt(aCutpoint,cutpointClass,true));
				
				addJPVars(genadvices,aCutpoint,cutpointClass);
				
				generateRefinableMethods(aCutpoint,cutpointClass);
				
	
				generateLocalJavaCode(aCutpoint,cutpointClass);
				
				
				generateGlobalKeyword(aCutpoint,cutpointClass);
				
				
			}
			else 
			{	
				PCutpointConstructor constructor = aCutpoint.getCutpointConstructor();
				if(constructor==null)
					throw new IllegalArgumentException("hook "+aCutpoint.getCutpointName()+" doesnt define a constructor!!!");
				
				generateFields(aCutpoint,cutpointClass);
				generateConstructor(aCutpoint,cutpointClass);
				generateConstructorField(aCutpoint,cutpointClass);
				//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
				generateCupointSetNamingMethod(aCutpoint,cutpointClass);
				generateCupointGetNamingMethod(aCutpoint,cutpointClass);
				generateRefinableMethods(aCutpoint,cutpointClass);
				generateConnectorSettingMethod(aCutpoint,cutpointClass); 
				generateConnectorGettingMethod(aCutpoint,cutpointClass);
				//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
				genadvices.addAll(generateBefore(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateAfter(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateReplace(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateAfterThrowing(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateAfterReturning(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateAroundThrowing(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateAroundReturning(aCutpoint,cutpointClass,false));
				genadvices.addAll(generateTestStmt(aCutpoint,cutpointClass,false));
				 
				addJPVars(genadvices,aCutpoint,cutpointClass);
				
				generateCalledMethod(aCutpoint,cutpointClass); //see generateCalledMethod comment
				//generateInitialise(aCutpoint,cutpointClass);
				
				generateLocalJavaCode(aCutpoint,cutpointClass);
				generateProceed(aCutpoint,cutpointClass);
				
				generateGlobalKeyword(aCutpoint,cutpointClass);
				
				
				
			}
		try
		{	
			String packageDir = "";
			if(!packageName.equals(""))
				packageDir=packageName.replace('.','/')+"/";
			File filetowrite = new File(Options.getOutputDir() + "/"+ packageDir+cutpointClassName + ".java");
			if(filetowrite.exists())
				filetowrite.delete();
			else if(filetowrite.getParent()!=null)
				new File(filetowrite.getParent()).mkdirs();	
			BufferedWriter out = new BufferedWriter(new FileWriter(filetowrite));
			out.write(cutpointClass.toString());
			out.close();
			
			//compile all files
			JavaCompiler compile = new JavaCompiler();
			String classpath = Options.getClassPath();
			compile.compile(filetowrite.getCanonicalPath(),classpath);
			Vector errors = compile.getErrors();
			boolean succes = translateAndPrint(errors);
			
			if(Options.deleteTempFiles())
				filetowrite.delete();
			
			if(succes)
				setHookContextProperty(aCutpoint,realCutpointName);
		}
		catch(Exception e)
		{
			Logger.getInstance().showFullError(e);
			
		}
	}
	
	private void checkNoConsturctor(PCutpoint aCutpoint) {
		
		if(aCutpoint.getCutpointConstructor()!=null)
			throw new IllegalArgumentException("A subhook cannot redefine the constructor: "+aCutpoint.getCutpointName());
	}
	
	private String generateTargetObjectType(PCutpoint aCutpoint) {
		
		PCutpointConstructor constr = aCutpoint.getCutpointConstructor();
		PCutpointConstructorApplicationDesignator app = null;
		if(constr==null) {
			
			String fullname = aCutpoint.getFullExtendingCutpointName(aspect);
   			try {
   			Class clz = Options.loadClass(fullname);
   			java.lang.reflect.Field f2 = clz.getField(CONSTRUCTOR_REP);
   			PCutpointConstructorApplicationDesignator designator = (PCutpointConstructorApplicationDesignator) f2.get(null);
   			app=designator;
   			} catch(Exception ex) {
   				Logger.getInstance().showError("Super aspect failed loading for target fetching ("+fullname+")");
   				Logger.getInstance().showError(ex);
   				return "Object";
   			}
		}
		else app = constr.getConstructorApplicationDesignator();
		
		PCutpointConstructorTarget target = findTarget(app); 
		
		String targetObjectType = "Object";
		if(target!=null)
			targetObjectType=target.getTargetClassName();
		
		return targetObjectType;
			
	}
	
	private PCutpointConstructorTarget findTarget(PCutpointConstructorApplicationDesignator app) {
		
		if(app==null) 
			return null;
		if(app instanceof PCutpointConstructorTarget)
			return (PCutpointConstructorTarget) app;
		
		PCutpointConstructorTarget target = findTarget(app.getParameter());
		
		if(target!=null)
			return target;
		 
		target = findTarget(app.getParameter2());
		
		if(target!=null)
			return target;
		else return null;
	}
	
	private void setHookContextProperty(PCutpoint aCutpoint, String cname)  {
		
		HookContextProperty orig = aCutpoint.getHookContextProperty();
		
		try  {
		ContextPropertySetter setter = new ContextPropertySetter(Options.getOutputDir());
		setter.setProperty(cname,orig);
		}
		catch(Exception ex)  {
			Logger.getInstance().showFullError(ex);
		}
	}
	
	private void generateContextProperty(PCutpoint cp, Vector advices, Vector types, Vector names, Vector used, ClassGenerator cutpointClass)  {
	
		MethodGenerator method = new MethodGenerator(HOOKCONTEXTPROP,"HookContextProperty");
		method.addVariable("String","advice");
		method.addModifier(Modifier.PUBLIC);
		
		int index = names.indexOf(CONST_thisJoinPoint);
		Boolean thisJPused = (Boolean) used.elementAt(index);
		
		index = names.indexOf(CONST_thisJoinPointObject);
		Boolean thisJPObjectUsed = (Boolean) used.elementAt(index);
		
		index = types.indexOf(MULTIPLE_ELEMENTS);
		Boolean argsArrayUsed = new Boolean(false);
		if(index>=0)
			argsArrayUsed=(Boolean) used.elementAt(index);
		
		HookContextProperty prop = new HookContextProperty(true,argsArrayUsed.booleanValue(),thisJPObjectUsed.booleanValue(),
				thisJPused.booleanValue(),true,true); 
		
		String code = "return "+HOOKCONTEXTPROP+"(advice, "+prop.toSourceCode()+");"+JavaGenerator.NEWLINE;
		method.setImplementation(code);
		
		cutpointClass.addMethod(method);
		
		MethodGenerator method2 = new MethodGenerator(HOOKCONTEXTPROP,"HookContextProperty");
		method2.addVariable("String","advice");
		method2.addVariable("HookContextProperty","thisProperty");
		method2.addModifier(Modifier.PUBLIC);
		 
		code="";
	
		if(cp.hasSuperClass()) {
			code+="if (advice.equals(HookContextProperty.GLOBAL)) return HookContextProperty.intersection(thisProperty, super."+HOOKCONTEXTPROP+"(advice))"+";"+JavaGenerator.NEWLINE;
		}
		else code+="if (advice.equals(HookContextProperty.GLOBAL)) return thisProperty;"+JavaGenerator.NEWLINE;
		
		Iterator advicesI = advices.iterator();
		while(advicesI.hasNext()) {
			MethodGenerator gen = (MethodGenerator) advicesI.next();
			
			code+="else ";
			
			code+="if (advice.equals(\""+gen.getName()+"\")) return thisProperty;"+JavaGenerator.NEWLINE;
		}
		
		code+="else ";
		
		if(cp.hasSuperClass()) {
			
			code+="return super."+HOOKCONTEXTPROP+"(advice);"+JavaGenerator.NEWLINE;
		}
		else code+="return null;"+JavaGenerator.NEWLINE;
		 
		method2.setImplementation(code);
		cutpointClass.addMethod(method2);
		
		cp.setHookContextProperty(prop);
	}
	
	private void generateConstructorField(PCutpoint aCutpoint,ClassGenerator cutpointClass)  {
		
		PCutpointConstructor constructor = aCutpoint.getCutpointConstructor();
		if(constructor==null)
		return;
		
		VariableGenerator cpconstructorRep = new VariableGenerator("PCutpointConstructorApplicationDesignator","_Jasco_cutpointconstructorRep");
		cpconstructorRep.addModifier(Modifier.PUBLIC);
		cpconstructorRep.addModifier(Modifier.STATIC);
		cpconstructorRep.addModifier(Modifier.FINAL);
		cutpointClass.addVariable(cpconstructorRep);
		
		MethodGenerator cpconstructorRepGet = new MethodGenerator("_Jasco_getConstructorRep","PCutpointConstructorApplicationDesignator");
		cpconstructorRepGet.addModifier(Modifier.PUBLIC);
		cpconstructorRepGet.setImplementation("return _Jasco_cutpointconstructorRep;");
		cutpointClass.addMethod(cpconstructorRepGet);
		
		//cpconstructorRep.setInitialiser("null");
		
		if (aCutpoint.getCutpointConstructor().getConstructorApplicationDesignator() instanceof PCutpointConstructorRegular)  {
			//Logger.getInstance().showOutput("JASCO WARNING : hook "+aCutpoint.getCutpointName()+" is not compatible yet with Jutta because of the stateful constructor, using Jutta with this hook is not possible!");
			cutpointClass.addInterface("jasco.runtime.aspect.StatefulAspect");
		}
		else if (aCutpoint.getCutpointConstructor().getConstructorApplicationDesignator().doNotCache())  {
			//Logger.getInstance().showOutput("JASCO WARNING : hook "+aCutpoint.getCutpointName()+" is not Jutta cachable because cflow or withincode is used or it is a stateful aspect");
			cutpointClass.addInterface("jasco.runtime.DoNotCache");
		}
		
		cutpointClass.addStaticInit(aCutpoint.getCutpointConstructor().getConstructorApplicationDesignator().generateCreateCode("_Jasco_cutpointconstructorRep"));
		
		
		VariableGenerator cpconstructorArgs = new VariableGenerator("ConstructorMethod[]",CONSTRUCTOR_ARGS);
		cpconstructorArgs.addModifier(Modifier.PUBLIC);
		cpconstructorArgs.addModifier(Modifier.STATIC);
		cpconstructorArgs.addModifier(Modifier.FINAL);
		cutpointClass.addVariable(cpconstructorArgs);
		cpconstructorArgs.setInitialiser("new ConstructorMethod["+constructor.getNoCutpointConstructorMethods()+"]");
		
		String code="";
		for(int i=0;i<constructor.getNoCutpointConstructorMethods();++i)  {
			PCutpointConstructorMethod method = constructor.getMethodForTag(i);
			String args = "new Class[] {";
			boolean multipleElements = false;
			Iterator iter = method.getArgumentTypes();
			while (iter.hasNext())  {
				String next = (String) iter.next();
				if(isMultipleElements(next))
					multipleElements=true;
				else args+=next+".class";
				if(iter.hasNext())
					args+=", ";
			}
			String argnames = "new String[] {";
			iter = method.getArgumentNames();
			while (iter.hasNext())  {
				String next = (String) iter.next();
				argnames+="\""+next+"\"";
				if(iter.hasNext())
					argnames+=", ";
			}
			
			String returnType = "null";
			if(method.hasReturnType())
				returnType=method.getReturnType()+".class";
			args+="}";
			argnames+="}";
			code+=CONSTRUCTOR_ARGS+"["+i+"]= new ConstructorMethod("+returnType+ ", "+args+", "+argnames+", "+multipleElements+");"+Options.NEWLINE;
		}
		
		cutpointClass.addStaticInit(code);
		
		String calledMethodS = findCalledMethod(constructor.getConstructorApplicationDesignator());
		if(calledMethodS==null)
			return;
		PCutpointConstructorMethod constructormethod = (PCutpointConstructorMethod) constructor.getMethodForTag(calledMethodS);
		
		
		for (int i = 0; i < constructormethod.getArgumentsSize() ; i++)
		{	
			String type=constructormethod.getArgumentType(i);
			if(isMultipleElements(type))
				type="Object[]";
			VariableGenerator arg = new VariableGenerator(type,constructormethod.getArgumentName(i));
			arg.addModifier(Modifier.PROTECTED);
			cutpointClass.addVariable(arg);
		}
	}
	
	private void generateGlobalKeyword(PCutpoint aCutpoint,ClassGenerator cutpointClass)  {
		
		String globalName = aspect.getAspectName();//+"_global";
		VariableGenerator global = new VariableGenerator(globalName,"global");
		global.addModifier(Modifier.PROTECTED);
		cutpointClass.addVariable(global);
		
	}
		
	private void generateFields(PCutpoint aCutpoint,ClassGenerator cutpointClass)
	{	
		//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
		VariableGenerator cutpointname = new VariableGenerator("String","_Jasco_cutpointname");
		cutpointname.addModifier(Modifier.PRIVATE);
		cutpointClass.addVariable(cutpointname);
		//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
		
		//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
		VariableGenerator connectorref = new VariableGenerator("Connector","_Jasco_connector");
		connectorref.addModifier(Modifier.PRIVATE);
		cutpointClass.addVariable(connectorref);
		
		/*
		VariableGenerator thisJP = new VariableGenerator("JascoMethod",CONST_thisJoinPoint);
		thisJP.addModifier(Modifier.PROTECTED);
		cutpointClass.addVariable(thisJP);*/
		
		VariableGenerator abstractmethods = new VariableGenerator("Vector",ABSTRACT_METHODS);
		abstractmethods.addModifier(Modifier.PUBLIC);
		abstractmethods.addModifier(Modifier.FINAL);
		abstractmethods.addModifier(Modifier.STATIC);
		abstractmethods.setInitialiser("new Vector()");
		cutpointClass.addVariable(abstractmethods);
		
		VariableGenerator abstractmethodsImpl = new VariableGenerator("Vector",ABSTRACT_METHODS_IMPL);
		abstractmethodsImpl.addModifier(Modifier.PUBLIC);
		abstractmethodsImpl.addModifier(Modifier.FINAL);
		abstractmethodsImpl.addModifier(Modifier.STATIC);
		abstractmethodsImpl.setInitialiser("new Vector()");
		cutpointClass.addVariable(abstractmethodsImpl);
		
		String globalName = aspect.getAspectName();
		MethodGenerator cutpointVectorMethod = new MethodGenerator("_Jasco_getGlobal","Object");
		cutpointVectorMethod.addModifier(Modifier.PUBLIC);
		cutpointVectorMethod.setImplementation("return global;");
		cutpointClass.addMethod(cutpointVectorMethod);
		
		MethodGenerator setglobal = new MethodGenerator("_Jasco_setGlobal","void");
		setglobal.addVariable(globalName,"aGlobal");
		setglobal.addModifier(Modifier.PUBLIC);
		setglobal.setImplementation("global=aGlobal;");
		cutpointClass.addMethod(setglobal);
	
		
		
	
		
		
	}
	
	public static boolean isMultipleElements(String s)  {
	
		return s.equals(MULTIPLE_ELEMENTS);
	}
	
	private void generateConstructor(PCutpoint aCutpoint,ClassGenerator cutpointClass)
	{	
		String extendingcutpointname = aCutpoint.getExtendingCutpointName();
		
		
		String code= "";
		
		if(aCutpoint.hasSuperCutpoint())
			code+="super(aGlobal);"+JavaGenerator.NEWLINE;
		code+="global=aGlobal;"+JavaGenerator.NEWLINE;	
		
	
		String globalName = aspect.getAspectName();//+"_global";
		MethodGenerator constructor = new MethodGenerator(cutpointClass.getName(),"");
		constructor.addVariable(globalName,"aGlobal");
		constructor.addModifier(Modifier.PUBLIC);
		constructor.setImplementation(code);
		cutpointClass.addMethod(constructor);
		
		MethodGenerator constructor2 = new MethodGenerator(cutpointClass.getName(),"");
		constructor2.addModifier(Modifier.PUBLIC);
		constructor2.setImplementation(" ");
		cutpointClass.addMethod(constructor2);
		
	
		
	
	}
	
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	private void generateCupointSetNamingMethod(PCutpoint aCutpoint,ClassGenerator cutpointClass) 
	{
		MethodGenerator cutpointnaming = new MethodGenerator("_Jasco_setCutpointName","void");
		cutpointnaming.addModifier(Modifier.PUBLIC);
		cutpointnaming.addVariable("String","aCutpointName");
		String implementation = "_Jasco_cutpointname = aCutpointName;";
		cutpointnaming.setImplementation(implementation);
		cutpointClass.addMethod(cutpointnaming);
	}
	private void generateCupointGetNamingMethod(PCutpoint aCutpoint,ClassGenerator cutpointClass) 
	{
		MethodGenerator cutpointnaming = new MethodGenerator("_Jasco_getCutpointName","String");
		cutpointnaming.addModifier(Modifier.PUBLIC);
		String implementation = "return _Jasco_cutpointname;";
		cutpointnaming.setImplementation(implementation);
		cutpointClass.addMethod(cutpointnaming);
	}
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	private void generateConnectorSettingMethod(PCutpoint aCutpoint,ClassGenerator cutpointClass) 
	{
		MethodGenerator cutpointnaming = new MethodGenerator("_Jasco_setConnector","void");
		cutpointnaming.addModifier(Modifier.PUBLIC);
		cutpointnaming.addVariable("Connector","aConnector");
		String implementation = "_Jasco_connector = aConnector;";
		cutpointnaming.setImplementation(implementation);
		cutpointClass.addMethod(cutpointnaming);
	}
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	private void generateConnectorGettingMethod(PCutpoint aCutpoint,ClassGenerator cutpointClass) 
	{
		MethodGenerator cutpointnaming = new MethodGenerator("_Jasco_getConnector","Connector");
		cutpointnaming.addModifier(Modifier.PUBLIC);
		String implementation = "return _Jasco_connector;";
		cutpointnaming.setImplementation(implementation);
		cutpointClass.addMethod(cutpointnaming);
	}
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	
	//todo synchronized!!!!!!!!!!!!!!!!!
	protected int checkAccessor(String s)  {
	
		if(s.equals("public"))
			return Modifier.PUBLIC;
		else if(s.equals("protected"))
			return Modifier.PROTECTED;
		else if(s.equals("private")) 
			return Modifier.PRIVATE;
		else return -1;
	}	
			
	private void generateRefinableMethods(PCutpoint aCutpoint,ClassGenerator cutpointClass) 
	{	
		Iterator methods = aCutpoint.getMethods();
		while(methods.hasNext()) {
			PMethod m = (PMethod) methods.next();
			MethodGenerator mimpl = m.getMethodInfo();
			
			if(AspectParser.isRefinable(mimpl.getModifiers())) {
				processRefinableMethod(mimpl,cutpointClass);
				cutpointClass.addStaticInit(ABSTRACT_METHODS+".add(\""+mimpl.getName()+"\");");
			}
		}
	} 
	
	protected void processRefinableMethod(MethodGenerator method, ClassGenerator cutpointClass) {
		
		String invokeDefaultMethodName = method.getName()+"_Jasco_DEFAULT";
		MethodGenerator invokeDefaultMethod = method.copy(invokeDefaultMethodName);
		if(method.hasImplementation())
			cutpointClass.addMethod(invokeDefaultMethod);
		
		String methodname = method.getName();
		Iterator j = method.getVariableTypes();
		Iterator i = method.getVariableNames();
		String arguments = "new Object[] {";
		String classes = "new Class[] {";
		while(j.hasNext())
		{
			String type = (String) j.next();
			String name = (String) i.next();
			arguments = arguments + makePrimitiveObject(type,name);
			arguments = arguments + ",";
			String prim = getPrimitiveType(type);
			if(prim!=null) {
				type=prim;
				classes+=type+".TYPE, ";
			}
			else classes+=type+".class, ";
		}
		arguments += "this }";
		classes+="this.getClass() }";
		
		
		String code="String error=\"\";"+JavaGenerator.NEWLINE;
		code = code + "try {" + JavaGenerator.NEWLINE;
		code = code + "Class theclass = _Jasco_getConnector().getClass();" + JavaGenerator.NEWLINE;
		code = code + "Method themethod = theclass.getMethod(\"" +method.getName()+ "\"" + " + _Jasco_getCutpointName(),"+classes+");" + JavaGenerator.NEWLINE;
		
		code = code + "Object returnvalue = null;" + JavaGenerator.NEWLINE;
		
		String returntype = method.getReturnType();
		
		if (returntype.equals("void"))   
		{  
			code = code + "themethod.invoke(_Jasco_getConnector()," + arguments + ");" + JavaGenerator.NEWLINE;
			code+="return;"+ JavaGenerator.NEWLINE;
		}
		else 
		{
			code = code + "returnvalue = themethod.invoke(_Jasco_getConnector()," + arguments + ");" + JavaGenerator.NEWLINE;
			
		}	
		if (!returntype.equals("void")) 
		{
			code = code + "return " + makePrimitive(returntype,"returnvalue") + ";"+JavaGenerator.NEWLINE;
		}
			
		code += "}" + JavaGenerator.NEWLINE;
		code += "catch(IllegalAccessException e) {System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
		code += "catch(IllegalArgumentException e) {System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
		code += "catch(java.lang.reflect.InvocationTargetException invoc) {"+ JavaGenerator.NEWLINE;
		code += JavaGenerator.TAB+"Throwable b = invoc.getTargetException();"+ JavaGenerator.NEWLINE;
		code+=   JavaGenerator.TAB+"if(b!= null) {"+JavaGenerator.NEWLINE;
		boolean runtimeExDone=false;
		boolean addClassNotFoundEx = false;
		String runtimeEx="java.lang.RuntimeException";
		Iterator iter = method.getExceptions();
		while(iter.hasNext()) {
			String next = iter.next().toString();
			code+=JavaGenerator.TAB+JavaGenerator.TAB+"if(b instanceof "+next+")";
			code+=" throw ("+next+") b;"+JavaGenerator.NEWLINE;
			if(next.equals(runtimeEx))
				runtimeExDone=true;
		}
		if (!runtimeExDone)  {
			code+=JavaGenerator.TAB+JavaGenerator.TAB+"if(b instanceof "+runtimeEx+")";
			code+=" throw ("+runtimeEx+") b;"+JavaGenerator.NEWLINE;
		}
		code+=JavaGenerator.TAB+"}"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		if(addClassNotFoundEx) 
			code += "catch(ClassNotFoundException e) {System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
		
		String catchNoSuchMethod = "";
		if(method.hasImplementation()) { //abstract method with default impl
			if(!method.getReturnType().equals("void"))
				catchNoSuchMethod+="return ";
			catchNoSuchMethod+=invokeDefaultMethod.generateInvocation()+";";
			if(method.getReturnType().equals("void"))
				catchNoSuchMethod+="return;";
		}
		else {
			catchNoSuchMethod="System.err.println(\"Abstract Method not Implemented in connector: "+methodname+"\");System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();";
			cutpointClass.addStaticInit(ABSTRACT_METHODS_IMPL+".add(\""+method.getName()+"\");");
		}
		
		code += "catch(NoSuchMethodException e) {"+catchNoSuchMethod+"}"+ JavaGenerator.NEWLINE;
		
		code+= "throw new IllegalArgumentException(\"!!!!!JASCO SYSTEM ERROR!!!!!  this should not occur in refinable \");"+ JavaGenerator.NEWLINE;
		
		method.setImplementation(code);
		
	}
		
		
		
		
		
		
		/*PJavaCodes abstracts = aCutpoint.getAbstractJavaCode();
		for (int i = 0; i < abstracts.getJavaCodeSize(); i++) 
		{
			PJavaCode javacode = abstracts.getJavaCode(i); 
			String tempcode = javacode.getJavaCode(0);
			
			// Do some parsing to get Data needed for abstract method implementation
			StringTokenizer t = new StringTokenizer(tempcode," ");
			String accesser = t.nextToken(); // Get Accessortype
			String returntype;
			int modifier = checkAccessor(accesser);
			if(modifier==-1) //no accessor, package default
				returntype=accesser;
			else returntype = t.nextToken();
			
			String tempreturntype = t.nextToken();
			while (tempreturntype .equals(".")) 
			{
				returntype = returntype + "." + t.nextToken();
				tempreturntype  = t.nextToken();
			}
			
			String methodname = tempreturntype ;  // Get methodname
			
			t.nextToken(); // skip haakje
			
			Vector args = new Vector();
			String temp = t.nextToken();
			while (!temp.equals(")")) 
			{
				String argtype = temp;
			    String tempargtype = t.nextToken();
				while (tempargtype .equals(".")) 
				{
					argtype = argtype + "." + t.nextToken();
				    tempargtype  = t.nextToken();
				}
				args.add(argtype);
				temp = t.nextToken();
				if (temp.equals(",")) 
				{
					temp = t.nextToken();
				}
			}
			
			MethodGenerator abstractMethod = new MethodGenerator(methodname,returntype);
			if(modifier!=-1)
				abstractMethod.addModifier(modifier);
				
			//parse exceptions
			String throwsclause = t.nextToken();
			String currentType="";
			while (t.hasMoreTokens())  {
				String ex1 = t.nextToken();
				String next = t.nextToken();
				currentType+=ex1;
				if (next.equals("."))  {
					currentType+=".";
				}
				else  {
					abstractMethod.addException(currentType);
					currentType="";
				}
				
			}
			
					
			
			
			String methodargsfinder="";	
			
			boolean addClassNotFoundEx =false;
			Iterator argsI = args.iterator();
			int j= 0;
			while (argsI.hasNext())  {
				
				String type = argsI.next().toString();
				abstractMethod.addVariable(type,"args"+j++);
				
				String newType = testPrimitive(type);
				if (newType==null)  {
					addClassNotFoundEx=true;
					newType=findClass(type,cutpointClass);
					newType="Class.forName(\""+newType+"\")";
				}
				
				methodargsfinder+=newType+",";
			}
			
			methodargsfinder+="this.getClass()";
		
			String code="String error=\"\";"+JavaGenerator.NEWLINE;
			code = code + "try {" + JavaGenerator.NEWLINE;
			code = code + "Class theclass = _Jasco_getConnector().getClass();" + JavaGenerator.NEWLINE;
			code = code + "Method themethod = theclass.getMethod(\"" +methodname+ "\"" + " + _Jasco_getCutpointName(),new Class[] {"+methodargsfinder+"});" + JavaGenerator.NEWLINE;
			
				
			String arguments = "new Object[] {";
			for(int k = 0 ; k < args.size() ; k++) 
			{
				String type = (String)args.elementAt(k);
				arguments = arguments + makePrimitiveObject(type,"args" + k);
				
				arguments = arguments + ",";
			
			}
			arguments += "this }";
			
			code = code + "Object returnvalue = null;" + JavaGenerator.NEWLINE;
		
			
			if (returntype.equals("void"))   
			{  
				code = code + "themethod.invoke(_Jasco_getConnector()," + arguments + ");" + JavaGenerator.NEWLINE;
				code+="return;"+ JavaGenerator.NEWLINE;
			}
			else 
			{
				code = code + "returnvalue = themethod.invoke(_Jasco_getConnector()," + arguments + ");" + JavaGenerator.NEWLINE;
				
			}	
			if (!returntype.equals("void")) 
			{
				code = code + "return " + makePrimitive(returntype,"returnvalue") + ";"+JavaGenerator.NEWLINE;
			}
				
			code += "}" + JavaGenerator.NEWLINE;
			code += "catch(IllegalAccessException e) {System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
			code += "catch(IllegalArgumentException e) {System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
			code += "catch(java.lang.reflect.InvocationTargetException invoc) {"+ JavaGenerator.NEWLINE;
			code += JavaGenerator.TAB+"Throwable b = invoc.getTargetException();"+ JavaGenerator.NEWLINE;
			code+=   JavaGenerator.TAB+"if(b!= null) {"+JavaGenerator.NEWLINE;
			boolean runtimeExDone=false;
			String runtimeEx="java.lang.RuntimeException";
			Iterator iter = abstractMethod.getExceptions();
			while(iter.hasNext()) {
				String next = iter.next().toString();
				code+=JavaGenerator.TAB+JavaGenerator.TAB+"if(b instanceof "+next+")";
				code+=" throw ("+next+") b;"+JavaGenerator.NEWLINE;
				if(next.equals(runtimeEx))
					runtimeExDone=true;
			}
			if (!runtimeExDone)  {
				code+=JavaGenerator.TAB+JavaGenerator.TAB+"if(b instanceof "+runtimeEx+")";
				code+=" throw ("+runtimeEx+") b;"+JavaGenerator.NEWLINE;
			}
			code+=JavaGenerator.TAB+"}"+JavaGenerator.NEWLINE;
			code+="}"+JavaGenerator.NEWLINE;
			if(addClassNotFoundEx) 
				code += "catch(ClassNotFoundException e) {System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
			code += "catch(NoSuchMethodException e) {System.err.println(\"Abstract Method not Implemented in connector: "+methodname+"\");System.err.println(\"!!!!!JASCO SYSTEM ERROR!!!!! \"+e.getMessage());e.printStackTrace();error=e.getMessage();}"+ JavaGenerator.NEWLINE;
			
			code+="throw new IllegalArgumentException(\"!!!JASCO SYSTEM ERROR!!!\"+error);"+ JavaGenerator.NEWLINE;
			
			abstractMethod.setImplementation(code);
			cutpointClass.addMethod(abstractMethod);
			//cutpointClass.addStaticInit(ABSTRACT_METHODS+".add("+methodname+");");
		}*/
	
	

	protected String findClass(String str,ClassGenerator cutpointClass)  {
		
		String currentName = str;
		
		try {
		Options.loadClass(currentName);
		return currentName;
		}
		catch(Throwable ex) {
		}
		
		Vector packages = new Vector();
		Iterator i = cutpointClass.getImportPackages();
		while (i.hasNext())  {
			GImport next = (GImport) i.next();
			String pack =  next.getImportName();
			if(pack.endsWith("*"))
				packages.add(pack.substring(0,pack.length()-1));
			else if (pack.endsWith(currentName)) 	 
				return pack;
		}
		packages.add(cutpointClass.getPackageName()+".");
		packages.add("java.lang.");
		
		i = packages.iterator();
		while (i.hasNext())  {
			String next = i.next().toString();
			try {
			//System.out.println("testing:"+next+currentName);
			Options.loadClass(next+currentName);
			return next+currentName;
			}
			catch(Throwable ex) {
			}
		}
		
		throw new IllegalArgumentException("Class not found error: "+str);
	}
	
	private String makePrimitiveObject(String aClass,String aObject) 
	{
		if(aClass.equals("boolean"))
			return "new Boolean(" + aObject + ")";
		else if(aClass.equals("int"))
			return "new Integer(" + aObject + ")";
		else if(aClass.equals("float"))
			return "new Float(" + aObject + ")";
		else if(aClass.equals("double"))
			return "new Double(" + aObject + ")";
		else if(aClass.equals("char"))
			return "new Character(" + aObject + ")";
		else if(aClass.equals("byte"))
			return "new Byte(" + aObject + ")";
		else if(aClass.equals("short"))
			return "new Short(" + aObject + ")";
		else if(aClass.equals("long"))
			return "new Long(" + aObject + ")";
		else return aObject;
		
	}
	
	private String testPrimitive(String aClass) 
	{
		if(aClass.equals("boolean"))
			return "Boolean.TYPE";
		else if(aClass.equals("int"))
			return "Integer.TYPE";
		else if(aClass.equals("float"))
			return "Float.TYPE";
		else if(aClass.equals("double"))
			return "Double.TYPE";
		else if(aClass.equals("char"))
			return "Character.TYPE";
		else if(aClass.equals("byte"))
			return "Byte.TYPE";
		else if(aClass.equals("short"))
			return "Short.TYPE";
		else if(aClass.equals("long"))
			return "Long.TYPE";
		else return null;
		
	}
	
	private String makePrimitive(String aClass,String aObject) 
	{
		if(aClass.equals("boolean"))
			return "((Boolean)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("int"))
			return "((Integer)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("float"))
			return "((Float)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("double"))
			return "((Double)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("char"))
			return "((Character)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("byte"))
			return "((Boolean)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("short"))
			return "((Short)" + aObject + ")" + "." + aClass + "Value()";
		else if(aClass.equals("long"))
			return "((Long)" + aObject + ")" + "." + aClass + "Value()";
		else 
		{
			return "(" + aClass + ")" + aObject;
		}
	}
	private Vector generateBefore(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{
		String defaultImpl = null;
		if(!extended)
			defaultImpl=" ;";
		return generateAdvices(aCutpoint,cutpointClass,defaultImpl,aCutpoint.getBeforeCutpointAdaptation(),PCutpointAdaptation.BEFORE,"void");	
	}

	private Vector generateAfter(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		String defaultImpl = null;
		if(!extended)
			defaultImpl=" ;";
		return generateAdvices(aCutpoint,cutpointClass,defaultImpl,aCutpoint.getAfterCutpointAdaptation(),PCutpointAdaptation.AFTER,"void");
		
	}
	
	private Vector  generateTestStmt(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		String defaultImpl = null;
		if(!extended)
			defaultImpl="return true;";
		Vector adv = generateAdvices(aCutpoint,cutpointClass,defaultImpl,aCutpoint.getTest(),PCutpointTest.TYPE,"boolean");
		if(cutpointClass.containsVariable("_Jasco_hasIsApplicable"))
			cutpointClass.addInterface("jasco.runtime.DoNotCache");
		return adv;
	}
	
	private void generateProceed(PCutpoint aCutpoint,ClassGenerator cutpointClass) 
	{
		//System.out.println("boe");
		MethodGenerator replace = new MethodGenerator("proceed","Object");
		replace.addVariable("jasco.runtime.JascoMethod","calledmethod");
		replace.addModifier(Modifier.PUBLIC);
		replace.addVariable("Object","obj");
		replace.addVariable("Object[]","args");
		replace.addException("java.lang.Exception");
		replace.setImplementation("return calledmethod.invoke(obj,args);");
		replace.addModifier(Modifier.FINAL);
		cutpointClass.addMethod(replace);
		
		MethodGenerator replace2 = new MethodGenerator("proceed","Object");
		replace2.addModifier(Modifier.PUBLIC);
		replace2.addVariable("jasco.runtime.JascoMethod","calledmethod");
		replace2.addException("java.lang.Exception");
		replace2.setImplementation("return calledmethod.invokeJAsCoMethod();");
		replace2.addModifier(Modifier.FINAL);
		cutpointClass.addMethod(replace2);
	}
	
	private Vector generateMultipleArgsAdvices(PCutpoint aCutpoint,ClassGenerator cutpointClass,Hashtable advices,boolean extended,String adviceName,String globalAdviceName)
	{	
		Vector generatedAdv = new Vector();
		String returnType="void";
		boolean doesReturn = false;
		boolean throwsException = false;
		
		String defaultArgs = CONST_thisJoinPoint+", ";
		defaultArgs+=CONST_thisJoinPointObject+", ";
		defaultArgs+=CONST_args;
		
		String defaultImpl = null;
		if(adviceName.equals(PCutpointAdaptation.AROUND_RETURNING)||adviceName.equals(PCutpointAdaptation.AROUND_THROWING)) {
			returnType="Object";
			doesReturn=true;
		}
		if(adviceName.equals(PCutpointAdaptation.AROUND_THROWING))
			throwsException=true;
		
		if(extended) {
			defaultImpl="super."+adviceName+"(arg,"+defaultArgs+");"+JavaGenerator.NEWLINE;
			if(doesReturn)
				defaultImpl="return "+defaultImpl;
		}
		else { 
			defaultImpl=" ;"+JavaGenerator.NEWLINE;
			if(doesReturn)
				defaultImpl="return arg"+defaultImpl;
			if(throwsException)
				defaultImpl="throw (Exception) arg;";
		}

		String code=" ";
		boolean first=true;
		boolean objectFound = false;
		java.util.TreeSet set = new java.util.TreeSet();
		Iterator entries = advices.entrySet().iterator();
		while (entries.hasNext())  {
			Map.Entry entry = (Map.Entry) entries.next();
			String type = (String) entry.getKey();
			if(type.equals("java.lang.Object"))
				objectFound=true;
			
			Vector v = (Vector) entry.getValue();
			if(set.contains(type)) {
				throw new IllegalArgumentException("Duplicate "+adviceName+" definition: "+type);
			}
			else set.add(type);
			if(!first)
				code+="else ";
			else first=false;
			
			
			if(testPrimitive(type)==null) { //primitives are bound directly, no test works
				
				//code+="if(Class.forName(\""+type+"\").isAssignableFrom(arg.getClass()))"+JavaGenerator.NEWLINE;
				code+="if(arg instanceof "+type+")"+JavaGenerator.NEWLINE;
				if(doesReturn) code+="return ";
				code+=JavaGenerator.TAB+adviceName+"("+"("+type+")arg, "+defaultArgs+");"+JavaGenerator.NEWLINE;
			}
			else {
				code+="if(arg instanceof "+getPrimitiveType(type)+")"+JavaGenerator.NEWLINE;
				if(doesReturn) code+="return ";
				code+=JavaGenerator.TAB+adviceName+"("+makePrimitive(type,"arg")+", "+defaultArgs+");"+JavaGenerator.NEWLINE;
			}
			
			Iterator sameIterator = v.iterator();
			generatedAdv.addAll(generateAdvices(aCutpoint,cutpointClass,defaultImpl,sameIterator,adviceName,returnType,type));
		}
		
		if (!first)  { //after throwing/returning defined...
			 {
			String field = "_Jasco_has"+JavaGenerator.upperCaseFirst(adviceName);
			VariableGenerator hasAdvice = new VariableGenerator("boolean",field);
			hasAdvice.addModifier(Modifier.PUBLIC);
			hasAdvice.setInitialiser("true");
			hasAdvice.addModifier(Modifier.STATIC);
			hasAdvice.addModifier(Modifier.FINAL);
			cutpointClass.addVariable(hasAdvice);
			 } {
			 String field = "_Jasco_has"+JavaGenerator.upperCaseFirst(globalAdviceName);
			VariableGenerator hasAdvice = new VariableGenerator("boolean",field);
			hasAdvice.addModifier(Modifier.PUBLIC);
			hasAdvice.setInitialiser("true");
			hasAdvice.addModifier(Modifier.STATIC);
			hasAdvice.addModifier(Modifier.FINAL);
			if(cutpointClass.getVariable(field)==null)
				cutpointClass.addVariable(hasAdvice);
			 } 
		}
		
		if(!objectFound) {
			MethodGenerator method=new MethodGenerator(adviceName,returnType);
			method.addException("java.lang.Exception");
			method.addModifier(Modifier.PUBLIC); 
			
			method.addVariable("Object","arg");	
			method.addVariable("jasco.runtime.JascoMethod",CONST_thisJoinPoint);
			method.addVariable("Object",CONST_thisJoinPointObject);
			method.addVariable("Object[]",CONST_args);
			
			if(first)
				code=defaultImpl;
			else code+="else "+defaultImpl;
				
			method.setImplementation(code);
		
			cutpointClass.addMethod(method);
		}
		
		return generatedAdv;
	}

	private Vector generateAfterThrowing(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		return generateMultipleArgsAdvices(aCutpoint,cutpointClass,aCutpoint.getAfterThrowingCutpointAdaptation(),extended,PCutpointAdaptation.AFTER_THROWING,PCutpointAdaptation.AFTER);
	}
	
	private Vector generateAfterReturning(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		return generateMultipleArgsAdvices(aCutpoint,cutpointClass,aCutpoint.getAfterReturningCutpointAdaptation(),extended,PCutpointAdaptation.AFTER_RETURNING,PCutpointAdaptation.AFTER);
	}
	
	private Vector generateAroundThrowing(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		return generateMultipleArgsAdvices(aCutpoint,cutpointClass,aCutpoint.getAroundThrowingCutpointAdaptation(),extended,PCutpointAdaptation.AROUND_THROWING,PCutpointAdaptation.REPLACE);
	}
	
	private Vector generateAroundReturning(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		return generateMultipleArgsAdvices(aCutpoint,cutpointClass,aCutpoint.getAroundReturningCutpointAdaptation(),extended,PCutpointAdaptation.AROUND_RETURNING,PCutpointAdaptation.REPLACE);
	}

	private Vector generateReplace(PCutpoint aCutpoint,ClassGenerator cutpointClass,boolean extended)
	{	
		String defaultImpl = null;
		if(!extended)
			defaultImpl="return proceed();"+JavaGenerator.NEWLINE;
		Vector advices = generateAdvices(aCutpoint,cutpointClass,defaultImpl,aCutpoint.getReplaceCutpointAdaptation(),PCutpointAdaptation.REPLACE,"Object");
		return advices;
	}
	
	private void patchProceed(MethodGenerator advice) {
		
		StringBuffer impl = new StringBuffer(advice.getImplementation());
		int lastindex=0;
		while(true) {
			int newindex = impl.indexOf(CONST_proceed,lastindex);
			if(newindex<0)
				break;
			newindex = impl.indexOf("(",newindex);
			if(newindex<0)
				break; //syntax error...
			newindex++;
			impl.insert(newindex,CONST_thisJoinPoint);
			lastindex=newindex;
		}
		advice.setImplementation(impl.toString());
	}
	
	private void checkName(PCutpoint aCutpoint, String name, String type,Vector names)  {
	
		if(names.contains(name))
			throw new IllegalArgumentException(type+" "+name+" can occur only once!");
		
	
		PCutpointConstructorRegular regularConstructor = null; 
		if (!aCutpoint.hasSuperCutpoint())  {
			PCutpointConstructorApplicationDesignator designator = aCutpoint.getCutpointConstructor().getConstructorApplicationDesignator();
			if(!(designator instanceof PCutpointConstructorRegular))
				throw new IllegalArgumentException(type+" "+name+" not found, "+"the hook does not define a stateful constructor");
			else  {
				regularConstructor=(PCutpointConstructorRegular) designator;
				if(!regularConstructor.definesName(name))
					throw new IllegalArgumentException(type+" "+name+" not found in the hook's stateful constructor");
			}
		} 
	}
	

	private Vector generateAdvices(PCutpoint aCutpoint,ClassGenerator cutpointClass, String defaultImpl, Iterator advices, String type, String returnType)  {
		return generateAdvices(aCutpoint,cutpointClass,defaultImpl,advices,type,returnType,null);
	}
	
	private Vector generateAdvices(PCutpoint aCutpoint,ClassGenerator cutpointClass, String defaultImpl, Iterator advices, String type, String returnType, String argType)  {
		
		Vector generatedAdvices = new Vector();
		
		boolean fullFound = false;
		boolean doesReturn = true;
		if(returnType.equals("void"))
			doesReturn=false;
		
		boolean passArgs=argType!=null;
		String argName = "";
		if(passArgs)
			argName="arg";
			
		Vector names = new Vector();
		
		while (advices.hasNext())  {
			PCutpointAdaptation adaptation = (PCutpointAdaptation) advices.next();
			
			if (!adaptation.hasName())  {
				if(fullFound) 
					throw new IllegalArgumentException("Error: multiple "+type+" declarations");
				else fullFound=true;
			}
			else  {
				checkName(aCutpoint,adaptation.getName(), type,names);
				names.add(adaptation.getName());
					
			}
			generatedAdvices.add(generateAdvice(adaptation,aCutpoint,cutpointClass,argType));
		}
		
		
		
		if (!fullFound)  {
			
			String code = "";
			boolean isFirst= true;
			Iterator namesI = names.iterator();
			
			
			
			if((!namesI.hasNext())&&defaultImpl==null) //no impl necessary
				return generatedAdvices;

			while (namesI.hasNext())  {
				String name = (String) namesI.next();
				if(!isFirst)
					code+="else ";
				else  { 
					isFirst=false;
					code+="PCutpointConstructorRegular regular = (PCutpointConstructorRegular) _Jasco_cutpointconstructorRep;"+JavaGenerator.NEWLINE;
					code+="String currentStateID = regular.getCurrentStateID();"+JavaGenerator.NEWLINE; 
					
					 
				}
				code+="if(currentStateID.equals(\""+name+"\"))"+JavaGenerator.NEWLINE;
				if(doesReturn)
					code+=JavaGenerator.TAB+"return "+type+"_"+name+"($$);"+JavaGenerator.NEWLINE;
				else code+=JavaGenerator.TAB+type+"_"+name+"($$);"+JavaGenerator.NEWLINE;
			}	
			if(!isFirst)
				code+="else ";
			code+=defaultImpl;
			
			MethodGenerator advice = new MethodGenerator(type,returnType);
			
			advice.addModifier(Modifier.PUBLIC);
			if(passArgs)
				advice.addVariable(argType,argName);
			if(!type.equals(PCutpointTest.TYPE))
				advice.addException("java.lang.Exception");
			cutpointClass.addMethod(advice);
			advice.setImplementation(code);
			
			generatedAdvices.add(advice);
		}
		else if(names.size()>0)
			throw new IllegalArgumentException("it is not possible to define both a normal "+type+" and a stateful one");
		
		if(argType!=null)
			return generatedAdvices;
		
		if (fullFound||names.size()>0)  {
			String field = "_Jasco_has"+JavaGenerator.upperCaseFirst(type);
			VariableGenerator hasAdvice = new VariableGenerator("boolean",field);
			hasAdvice.addModifier(Modifier.PUBLIC);
			hasAdvice.setInitialiser("true");
			hasAdvice.addModifier(Modifier.STATIC);
			hasAdvice.addModifier(Modifier.FINAL);
			if(cutpointClass.getVariable(field)==null)
				cutpointClass.addVariable(hasAdvice);
		}	
		
		return generatedAdvices;
	}		
	
	private MethodGenerator generateAdvice(PCutpointAdaptation adaptation, PCutpoint aCutpoint,ClassGenerator cutpointClass, String argType)  {
		
		String type = adaptation.getType();
		String returnType = adaptation.getReturnType();
		
		String argName = adaptation.getArgName();
		
		if (adaptation.hasName())  {
			String name = adaptation.getName();
			type+="_"+name;
		}
		
		MethodGenerator advice = new MethodGenerator(type,returnType);
		advice.setLineNo(adaptation.getLine());
		if(argName!=null&&argType!=null)
			advice.addVariable(argType,argName);
		advice.addModifier(Modifier.PUBLIC);
		if(adaptation.throwsException())
			advice.addException("java.lang.Exception");
		cutpointClass.addMethod(advice);
		String implementation = "";
		
		PJavaCode code = adaptation.getCutpointAdaptationImplementation();
		for (int j = 0; j < code.getJavaCodeSize() ; j++) 
		{
			implementation = implementation + code.getJavaCode(j) + " //" + code.getLineNumber(j) + JavaGenerator.NEWLINE;
		}
		advice.setImplementation(implementation);
		
		if(Options.isJavaVMTargetAtLeast(1,5))
			advice.addAnnotation("@jasco.runtime.aspect.annotation.Advice(type=\""+type+"\")");
		
		return advice;
	}
	
	protected void addAll(java.util.Collection c, Object[] args) {
		
		for(int i=0;i<args.length;++i)
			c.add(args[i]);
	}
	
	protected void addAll(java.util.Collection c, Iterator args) {
		
		while(args.hasNext())
			c.add(args.next());
	}
	
   protected boolean getConstructorMethodArguments(PCutpoint hook, Vector names, Vector types) {
   		
   		PCutpointConstructor constructor = hook.getCutpointConstructor();
   		if(constructor==null) { //check constructor in super...
   			
   			String fullname = hook.getFullExtendingCutpointName(aspect);
   			try {
   			Class clz = Options.loadClass(fullname);
   			java.lang.reflect.Field f = clz.getField(CONSTRUCTOR_ARGS);
   			Object o = clz.newInstance();
   			ConstructorMethod[] cms = (ConstructorMethod[]) f.get(o);
   			java.lang.reflect.Field f2 = clz.getField(CONSTRUCTOR_REP);
   			PCutpointConstructorApplicationDesignator designator = (PCutpointConstructorApplicationDesignator) f2.get(o);
   			String calledMethodS = findCalledMethod(designator);
   			if(calledMethodS==null)
   				return false;
   			
   			ConstructorMethod cm = cms[new Integer(calledMethodS).intValue()];
   			addAll(names,cm.getArgNames());
   			for(int i=0;i<cm.getArgTypes().length;++i)
   				types.add(cm.getArgTypes()[i].getName());
   			return true;
   			} catch(Exception ex) {
   				Logger.getInstance().showError("Super aspect failed loading ("+fullname+")");
   				Logger.getInstance().showError(ex);
   				return false;
   			}
   		}

		String calledMethodS = findCalledMethod(constructor.getConstructorApplicationDesignator());
		if(calledMethodS==null)
			return false;
		
		PCutpointConstructorMethod constructormethod = (PCutpointConstructorMethod) constructor.getMethodForTag(calledMethodS);
		addAll(names,constructormethod.getArgumentNames());
		addAll(types,constructormethod.getArgumentTypes());
		return true;
   }
	
	//known bug: ..args at end still contains all arguments
	protected void addJPVars(Vector advices, PCutpoint hook, ClassGenerator cpClass) {
		
		Vector types = new Vector();
		Vector names = new Vector();
		types.add("jasco.runtime.JascoMethod");
		types.add(generateTargetObjectType(hook));
		names.add(CONST_thisJoinPoint);
		names.add(CONST_thisJoinPointObject);
		
		boolean result = getConstructorMethodArguments(hook,names,types);
		if(!result) {
			Logger.getInstance().showFullError("error in finding args for advice!!!");
			return;
		}
		
		Iterator advicesI = advices.iterator();
		Vector useargs = new Vector();
		for(int i=0;i<names.size();++i)
			useargs.add(new Boolean(false));
		
		//check whether args are used by all advices...
		while(advicesI.hasNext()) {
			MethodGenerator advice = (MethodGenerator) advicesI.next();
			
			int index=0;
			Iterator namesI = names.iterator();
			Iterator typesI = types.iterator();
			while(namesI.hasNext()) {
				String argname = (String) namesI.next();
				String argtype = MULTIPLE_ELEMENTS;
				if(typesI.hasNext()) { //if multiple elements, no next...
					argtype = (String) typesI.next();
				}
//				//important for generateHookContextProperty, should be last so no check for concurrent mod necessarry
				else types.add(MULTIPLE_ELEMENTS); 
				
				boolean usearg = checkUse(advice,argname,argtype);
				if(usearg)
					useargs.setElementAt(new Boolean(true),index);
				++index;
			}
		}
		
		generateContextProperty(hook,advices,types,names,useargs,cpClass);
				
		//generate advice copy with default arguments (calledmethod,calledobject,args array)
		//+ change original advice to take only the required arguments + correct types of jp args
		advicesI = advices.iterator();
		while(advicesI.hasNext()) {
			String invokeCode = "";
			int index = -2;
			boolean mutlipleElementsFirst = false;
			MethodGenerator advice = (MethodGenerator) advicesI.next();
			MethodGenerator adviceCopy = advice.copy(advice.getName());
			boolean isReplaceAdvice = advice.getName().startsWith(PCutpointAdaptation.REPLACE);
			
			Iterator i = advice.getVariableNames();
			while(i.hasNext()) { 
				invokeCode+=i.next().toString();
				if(i.hasNext())
					invokeCode+=", ";
			}
			
			Iterator namesI = names.iterator();
			Iterator typesI = types.iterator();
			Iterator useargsI = useargs.iterator();
			while(namesI.hasNext()) {
				String argname = (String) namesI.next();
				String argtype = MULTIPLE_ELEMENTS;
				Boolean usearg = (Boolean) useargsI.next();
				
				if(typesI.hasNext()) //if multiple elements, no next...
					argtype = (String) typesI.next();
				if(usearg.booleanValue()||(argname.equals(CONST_thisJoinPoint)&&isReplaceAdvice)||(argname.equals(CONST_thisJoinPointObject)&&isReplaceAdvice)) {
					if(!invokeCode.equals(""))
						invokeCode+=", ";
					if(index<0) {//CONST_thisJoinPoint or CONST_thisJoinPointObject
						advice.addVariable(argtype,argname);
						if(argname.equals(CONST_thisJoinPointObject)&&!argtype.equals("Object"))
							invokeCode+="("+argtype+")";
						invokeCode+=argname;
					}
					else if (isMultipleElements(argtype)) {
						mutlipleElementsFirst=(index==0);
						advice.addVariable("Object[]",argname);
						invokeCode+=CONST_args;
					}
					else { 
						advice.addVariable(argtype,argname);
						String prim = getPrimitiveType(argtype);
						if (prim != null)  
							invokeCode+="(("+prim+")";
						else invokeCode+="("+argtype+")";	
						invokeCode+= CONST_args+"["+index+"]";
						if(prim!=null)
							invokeCode+=")."+argtype+"Value()";
					}
				}
				index++;
			}
			
			if(isReplaceAdvice)
				patchProceed(advice);
			
			adviceCopy.addVariable("jasco.runtime.JascoMethod",CONST_thisJoinPoint);
			adviceCopy.addVariable("Object",CONST_thisJoinPointObject);
			adviceCopy.addVariable("Object[]",CONST_args);
			
			//	advice has same signature as advice copy, dont add advice copy
			if(mutlipleElementsFirst&&advice.getVariableCount()==adviceCopy.getVariableCount()) {
				/*
				if(isReplaceAdvice) { //add jp to class for proceeds to work
					String code = advice.getImplementation();
					code="this."+CONST_thisJoinPoint+" = "+CONST_thisJoinPoint+";"+JavaGenerator.NEWLINE+code;
					advice.setImplementation(code);
				}*/
				continue;
			}
			
			String code = "";
			/*
			if(isReplaceAdvice) { //add jp to class for proceeds to work
				code+="this."+CONST_thisJoinPoint+" = "+CONST_thisJoinPoint+";"+JavaGenerator.NEWLINE;
			}*/
			
			code+=advice.getName()+"("+invokeCode+");"+JavaGenerator.NEWLINE;
			if(!adviceCopy.getReturnType().equals("void"))
				code="return "+code;
			
			adviceCopy.setImplementation(code);
			cpClass.addMethod(adviceCopy);
		}
	}
	
	protected boolean checkUse(MethodGenerator method, String argname, String argtype) {
		
		return method.getImplementation().contains(argname);
	}
		
	
	protected String findCalledMethod(PCutpointConstructorApplicationDesignator d)  {
	
		jasco.util.Queue q = new 	jasco.util.Queue();
		q.enqueue(d);
		
		while (!q.isEmpty())  {
			PCutpointConstructorApplicationDesignator des =(PCutpointConstructorApplicationDesignator) q.dequeue();
			if(des instanceof PCutpointConstructorExecute)
				return ((PCutpointConstructorExecute)des).getCalledMethod();
			else if(des instanceof PCutpointConstructorCall)
				return ((PCutpointConstructorCall)des).getCalledMethod();
			else if (des instanceof PCutpointConstructorRegular)  {
				Iterator i = ((PCutpointConstructorRegular)des).allTransitions().iterator();
				while (i.hasNext())  {
					Transition tr = (Transition) i.next();
					q.enqueue(tr.getParam());
				}
			}
			
			Object o = des.getParameter();
			if(o!=null)
				q.enqueue(o);
			o=des.getParameter2();	
			if(o!=null)
				q.enqueue(o);
		}
		return null;
	}	
	
	//execute(method) -> no method(args) supported any longer!
	private void generateCalledMethod(PCutpoint aCutpoint,ClassGenerator cutpointClass)
	{
			PCutpointConstructor constructor = aCutpoint.getCutpointConstructor();
			String calledMethod = findCalledMethod(constructor.getConstructorApplicationDesignator());
			if(calledMethod==null)
				return;
			PCutpointConstructorMethod constructormethod = (PCutpointConstructorMethod) constructor.getMethodForTag(calledMethod);
			
			MethodGenerator constructorMethod = new MethodGenerator("proceed","Object");
			constructorMethod.addModifier(Modifier.PROTECTED);
			constructorMethod.addException("java.lang.Exception");
			constructorMethod.addVariable("jasco.runtime.JascoMethod",CONST_thisJoinPoint);
			
			// no method() when no args, should use proceed() instead, is optimized...
			if (constructormethod.getArgumentsSize()==0)  {
				return;
			}
			
			String transCode = "";
			String endCode = "";
			
			
			Vector callargs = new Vector();
			for (int i = 0; i < constructormethod.getArgumentsSize(); i++) 
			{
				  String type = constructormethod.getArgumentType(i);
				  String argname = "var" + i;
				  if (isMultipleElements(type))  {
				  	type="Object[]";
					constructorMethod.addVariable(type, argname);
					endCode+="Object[] _newArgs = new Object["+i+"+"+argname+".length];"+JavaGenerator.NEWLINE;
					for(int j=0;j<callargs.size();++j)  {
						endCode+="_newArgs["+j+"] = "+callargs.elementAt(j)+";"+JavaGenerator.NEWLINE;
					}
					endCode+="for(int j=0;j<"+argname+".length;++j)"+JavaGenerator.NEWLINE;	
					endCode+=JavaGenerator.TAB+"_newArgs["+i+"+j] = "+argname+"[j];"+JavaGenerator.NEWLINE;
					break;
				  }	
				  constructorMethod.addVariable(type, argname);
				  String prim = getPrimitiveType(type);		 
				  if (prim!=null)  {
				  	String newArgname = argname+"Boxed";
				  	transCode+=prim+" "+newArgname+" = new "+prim+"("+argname+");"+JavaGenerator.NEWLINE;
					callargs.add(newArgname);
				  }
				  else callargs.add(argname);
			}
			
			if (endCode.equals(""))  {
				endCode+="Object[] _newArgs = new Object[] {";
				Iterator callargsI = callargs.iterator();
				while (callargsI.hasNext())  {
					endCode+=callargsI.next();
					if(callargsI.hasNext())
						endCode+=", ";
				}
				endCode+="};"+	JavaGenerator.NEWLINE;
			}
			
			String code = "";
		
			code+=transCode;
			code+=endCode;
			code+= "return "+CONST_thisJoinPoint+".invoke("+CONST_thisJoinPoint+".getCalledObject(),_newArgs);"+JavaGenerator.NEWLINE;
			
			constructorMethod.setImplementation(code);
			cutpointClass.addMethod(constructorMethod);
	}
	
	protected String getPrimitiveType(String type)  {

		if(type.equals("boolean"))
			return "Boolean";
		else if(type.equals("int"))
			return "Integer";
		else if(type.equals("float"))
			return "Float";
		else if(type.equals("double"))
			return "Double";
		else if(type.equals("char"))
			return "Character";
		else if(type.equals("byte"))
			return "Byte";
		else if(type.equals("short"))
			return "Short";
		else if(type.equals("long"))
			return "Long";
		else return null;	
	}
	
	private void generateLocalJavaCode(PCutpoint aCutpoint, ClassGenerator cutpointClass)
	{	
		aCutpoint.initClassGenerator(cutpointClass);
		
		boolean testPropertyImplemented = cutpointClass.getMethod("globalPropertyChangeEvent")!=null;
		
		if (!testPropertyImplemented&&!aCutpoint.hasSuperCutpoint())  {
			MethodGenerator prop = new MethodGenerator("globalPropertyChangeEvent","void");
			prop.addVariable("HookPropertyChangeEvent","event");
			prop.addModifier(Modifier.PUBLIC);
			cutpointClass.addMethod(prop);
			prop.setImplementation(" ");
		}

	}
}
