/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.aspectparser;


import jasco.runtime.aspect.HookContextProperty;
import jasco.tools.jascoparser.PJavaClass;
import jasco.tools.jascoparser.PJavaCode;
import jasco.tools.jascoparser.PJavaCodes;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

public class PCutpoint extends PJavaClass
{
	private PCutpointConstructor constructor = null;
	private Vector before = new Vector();
	private Vector after = new Vector();
	private Vector replace = new Vector();
	private Hashtable afterThrowing = new Hashtable();
	private Hashtable afterReturning = new Hashtable();
	private Hashtable aroundThrowing = new Hashtable();
	private Hashtable aroundReturning = new Hashtable();
	
	private PJavaCodes cutpointjavacode = new PJavaCodes();
	private PJavaCodes abstractjavacode = new PJavaCodes();
	private Vector test = new Vector();
	private HookContextProperty contextProp;

	public PCutpoint()
	{
		
	}
	
	public void addTest(PCutpointTest test)  {
	
		this.test.add(test);
	}
	
	
	public Iterator getTest()  {
	
		return test.iterator(); 
	}
	
	public void setHookContextProperty(HookContextProperty context) {
		this.contextProp=context;
	}
	
	public HookContextProperty getHookContextProperty() {
		
		return contextProp;
	}
	
	public void setCutpointName(String aCutpointName) 
	{
		setName(aCutpointName);
	}
	
	public String getCutpointName() 
	{
		return getName();
	}
	
	public boolean hasSuperCutpoint()  {
	
		return super.hasSuperClass();
	}
	
	public void setExtendingCutpointName(String aCutpointName) 
	{
		setSuperClass(aCutpointName);
	}
	
	public String getExtendingCutpointName() 
	{
		return getSuperClassSimple();
	}
	
	public void setCutpointConstructor(PCutpointConstructor aCutpointConstructor)
	{
		constructor = aCutpointConstructor;
	}
	
	public PCutpointConstructor getCutpointConstructor()
	{
		return constructor;
	}
	
	public void addBeforeCutpointAdaptation(PCutpointAdaptation aCutpointAdaptation)
	{
		before.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.BEFORE);
	}
	
	public void addAfterReturningCutpointAdaptation(String argname, PCutpointAdaptation aCutpointAdaptation)
	{
		Vector v = (Vector) afterReturning.get(argname);
		if (v==null)  {
			v=new Vector();
			afterReturning.put(argname,v);
		}
		v.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.AFTER_RETURNING);
	}
	
	public Hashtable getAfterReturningCutpointAdaptation() 
	{
		return afterReturning;
	}
	
	public void addAroundReturningCutpointAdaptation(String argname, PCutpointAdaptation aCutpointAdaptation)
	{
		Vector v = (Vector) aroundReturning.get(argname);
		if (v==null)  {
			v=new Vector();
			aroundReturning.put(argname,v);
		}
		v.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.AROUND_RETURNING);
	}
	
	public Hashtable getAroundReturningCutpointAdaptation() 
	{
		return aroundReturning;
	}
	
	public void addAfterThrowingCutpointAdaptation(String exception_name, PCutpointAdaptation aCutpointAdaptation)
	{
		Vector v = (Vector) afterThrowing.get(exception_name);
		if (v==null)  {
			v=new Vector();
			afterThrowing.put(exception_name,v);
		}
		v.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.AFTER_THROWING);
	}
	
	public Hashtable getAfterThrowingCutpointAdaptation() 
	{
		return afterThrowing;
	}
	
	public void addAroundThrowingCutpointAdaptation(String exception_name, PCutpointAdaptation aCutpointAdaptation)
	{
		Vector v = (Vector) aroundThrowing.get(exception_name);
		if (v==null)  {
			v=new Vector();
			aroundThrowing.put(exception_name,v);
		}
		v.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.AROUND_THROWING);
	}
	
	public Hashtable getAroundThrowingCutpointAdaptation() 
	{
		return aroundThrowing;
	}
	
	public Iterator getBeforeCutpointAdaptation() 
	{
		return before.iterator();
	}
	
	public void addAfterCutpointAdaptation(PCutpointAdaptation aCutpointAdaptation)
	{
		after.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.AFTER);
	}
	
	public Iterator getAfterCutpointAdaptation() 
	{
		return after.iterator();
	}
	
	public void addReplaceCutpointAdaptation(PCutpointAdaptation aCutpointAdaptation)
	{
		replace.add(aCutpointAdaptation);
		aCutpointAdaptation.setType(PCutpointAdaptation.REPLACE);
	}
	
	public Iterator getReplaceCutpointAdaptation() 
	{
		return replace.iterator();
	}
	
	public void addCutpointJavaCode(PJavaCode aJavaCode)
	{
		cutpointjavacode.addJavaCode(aJavaCode);
	}
	
	public PJavaCodes getCutpointJavaCode()
	{
		return cutpointjavacode;
	}
	
	public void addAbstractJavaCode(PJavaCode aJavaCode)
	{
		abstractjavacode.addJavaCode(aJavaCode);
	}
	
	public PJavaCodes getAbstractJavaCode()
	{
		return abstractjavacode;
	}
	
	public String getFullExtendingCutpointName(PAspect aspect) {
		
		String name = getExtendingCutpointName();
		return aspect.getSuperClass()+"_"+name;
	}
}