/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.connectorparser;

import jasco.options.Options;
import jasco.runtime.aspect.Cutpoint;
import jasco.runtime.aspect.HookHelper;
import jasco.runtime.aspect.PCutpointConstructorApplicationDesignator;
import jasco.runtime.aspect.PCutpointConstructorCall;
import jasco.runtime.aspect.PCutpointConstructorExecute;
import jasco.runtime.connector.CutpointExecutionElement;
import jasco.tools.aspectparser.AspectClassesGenerator;
import jasco.tools.jascoparser.PImport;
import jasco.tools.jascoparser.PJavaCode;
import jasco.tools.jascoparser.PJavaCodes;
import jasco.util.Queue;
import jasco.util.generators.ClassGenerator;
import jasco.util.generators.GImport;
import jasco.util.generators.GPackage;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.JavaNameGenerator;
import jasco.util.generators.MethodGenerator;
import jasco.util.generators.VariableGenerator;
import jasco.util.javacompiler.CompileError;
import jasco.util.javacompiler.JavaCompiler;
import jasco.util.logging.Logger;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

public class ConnectorClassesGenerator 
{	
	private static int index = 0;
	private PConnector connector = null;
	private Vector generatedglobals = new Vector();
	private Vector cutpointnames = new Vector();
	private Vector cutpointbefore = new Vector();
	private Vector cutpointafter = new Vector();
	private Vector cutpointreplace = new Vector();
	private HashMap nameMap = new HashMap();
	private Vector loadCode = new Vector();
	private Vector staticInitCode = new Vector();
	

	
	private static final String ASPECT_INIT_METHOD = "initAspects";
	
	private Vector lastErrors;
	
	public ConnectorClassesGenerator(PConnector aConnector) 
	{
		connector = aConnector;
	
	}
	
	public void generate() 
	{
		try {
		generateConnector();
		}
		catch(IllegalArgumentException ex)  {
			Logger.getInstance().showError(ex.getMessage());
		}
	}
	
	public void generateConnector() 
	{
		String connectorname = connector.getConnectorName();
		
		
		
		ClassGenerator connectorClass = new ClassGenerator(connectorname, new GPackage(Options.getConnectorPackage()));
		connectorClass.addInterface("Connector");
		connectorClass.addImportPackage(new GImport("java.lang.*"));
		connectorClass.addImportPackage(new GImport("java.util.*"));
		connectorClass.addImportPackage(new GImport("jasco.runtime.aspect.*"));
		connectorClass.addImportPackage(new GImport("jasco.runtime.connector.*"));
		connectorClass.addImportPackage(new GImport("jasco.runtime.*"));
		
		connectorClass.addModifier(Modifier.PUBLIC);
		
		VariableGenerator methods = new VariableGenerator("Vector","methods");
		methods.addModifier(Modifier.PRIVATE);
		//methods.addModifier(Modifier.STATIC);
		methods.setInitialiser("new Vector()");
		connectorClass.addVariable(methods);
		
		// Generation of combinationstrategies vector
		VariableGenerator combinationstrategies = new VariableGenerator("Vector","combinationStrategies");
		combinationstrategies.addModifier(Modifier.PRIVATE);
		combinationstrategies.setInitialiser("new "+"Vector"+"()");
		connectorClass.addVariable(combinationstrategies);
		
		VariableGenerator instances = new VariableGenerator("Vector","instances");
		instances.addModifier(Modifier.PRIVATE);
		//instances.addModifier(Modifier.STATIC);
		instances.setInitialiser("new Vector()");
		connectorClass.addVariable(instances);
		
		MethodGenerator addCombinationstrategy = new MethodGenerator("addCombinationStrategy","void");
		addCombinationstrategy.addVariable("CombinationStrategy","aStrategy");
		String code="myinstance.combinationStrategies.addElement(aStrategy);"+JavaGenerator.NEWLINE;
		addCombinationstrategy.setImplementation(code);						  
		addCombinationstrategy.addModifier(Modifier.PRIVATE);
		addCombinationstrategy.addModifier(Modifier.STATIC);
		connectorClass.addMethod(addCombinationstrategy);
		
		MethodGenerator addInstance = new MethodGenerator("addInstance","void");
		addInstance.addVariable("Object","object");
		code="instances.addElement(object);"+JavaGenerator.NEWLINE;
		addInstance.setImplementation(code);						  
		addInstance.addModifier(Modifier.PUBLIC);
		//addInstance.addModifier(Modifier.STATIC);
		connectorClass.addMethod(addInstance);
		
		MethodGenerator removeInstance = new MethodGenerator("removeInstance","void");
		removeInstance.addVariable("Object","object");
		code="instances.removeElement(object);"+JavaGenerator.NEWLINE;
		removeInstance.setImplementation(code);
		removeInstance.addModifier(Modifier.PUBLIC);
		//removeInstance.addModifier(Modifier.STATIC);
		connectorClass.addMethod(removeInstance);
		

		VariableGenerator myinstance = new VariableGenerator(connectorname,"myinstance");
		myinstance.addModifier(Modifier.PRIVATE);
		myinstance.addModifier(Modifier.STATIC);
		myinstance.setInitialiser("new "+connectorname+"()");
		connectorClass.addVariable(myinstance);
		
		MethodGenerator getconnector = new MethodGenerator("getConnector",connectorname);
		code="return myinstance;"+JavaGenerator.NEWLINE;
		getconnector.setImplementation(code);
		getconnector.addModifier(Modifier.PUBLIC);
		getconnector.addModifier(Modifier.STATIC);
		connectorClass.addMethod(getconnector);
		
		VariableGenerator isClassAdapter = new VariableGenerator("boolean","isClassAdapter");
		isClassAdapter.addModifier(Modifier.PRIVATE);
		//isClassAdapter.addModifier(Modifier.STATIC);
		isClassAdapter.setInitialiser("true");
		connectorClass.addVariable(isClassAdapter);
		
		MethodGenerator isClassAdapterM = new MethodGenerator("adaptOnClasses","boolean");
		code="return isClassAdapter;"+JavaGenerator.NEWLINE;
		isClassAdapterM.setImplementation(code);
		isClassAdapterM.addModifier(Modifier.PUBLIC);
		//isClassAdapterM.addModifier(Modifier.STATIC);
		connectorClass.addMethod(isClassAdapterM);
		
		MethodGenerator getName = new MethodGenerator("getName","String");
		code="return \""+connectorname+"\";"+JavaGenerator.NEWLINE;
		getName.setImplementation(code);
		getName.addModifier(Modifier.PUBLIC);
		//isClassAdapterM.addModifier(Modifier.STATIC);
		connectorClass.addMethod(getName);
		
		MethodGenerator setAdaptOnClasses = new MethodGenerator("setAdaptOnClasses","void");
		setAdaptOnClasses.addVariable("boolean","bol");
		code="setAdaptOnClassesImpl(bol);"+JavaGenerator.NEWLINE;
		code+="ConnectorPersistance.saveConnector(this);"+JavaGenerator.NEWLINE;
		//code+="firePropertyChangeEvent(new HookPropertyChangeEvent(HookPropertyChangeEvent.IS_STATIC,new Boolean(bol)));"+JavaGenerator.NEWLINE;
		setAdaptOnClasses.setImplementation(code);
		setAdaptOnClasses.addModifier(Modifier.PUBLIC);
		//setAdaptOnClasses.addModifier(Modifier.STATIC);
		connectorClass.addMethod(setAdaptOnClasses);
		
		MethodGenerator setAdaptOnClassesImpl = new MethodGenerator("setAdaptOnClassesImpl","void");
		setAdaptOnClassesImpl.addVariable("boolean","bol");
		code="isClassAdapter = bol;"+JavaGenerator.NEWLINE;
		code+="firePropertyChangeEvent(new HookPropertyChangeEvent(HookPropertyChangeEvent.IS_STATIC,new Boolean(bol)));"+JavaGenerator.NEWLINE;
		
		setAdaptOnClassesImpl.setImplementation(code);
		setAdaptOnClassesImpl.addModifier(Modifier.PUBLIC);
		//setAdaptOnClasses.addModifier(Modifier.STATIC);
		connectorClass.addMethod(setAdaptOnClassesImpl);
		
		MethodGenerator getMethods = new MethodGenerator("getMethods","Enumeration");
		code="return methods.elements();"+JavaGenerator.NEWLINE;
		getMethods.setImplementation(code);
		getMethods.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getMethods);
		
		MethodGenerator getMethodsHash = new MethodGenerator("getMethodHashtable","Hashtable");
		getMethodsHash.addVariable("int","i");
		code="return (Hashtable) methodHashtableVector.elementAt(i);"+JavaGenerator.NEWLINE;
		getMethodsHash.setImplementation(code);
		getMethodsHash.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getMethodsHash);
		
		MethodGenerator getMethodsHashC = new MethodGenerator("getMethodHashtable","Hashtable");
		getMethodsHashC.addVariable("CutpointElement","c");
		code="";
		code+="int pos = cutpoints.indexOf(c);"+JavaGenerator.NEWLINE;
		code+="return (Hashtable) methodHashtableVector.elementAt(pos);"+JavaGenerator.NEWLINE;
		getMethodsHashC.setImplementation(code);
		getMethodsHashC.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getMethodsHashC);
		
		MethodGenerator getIndex = new MethodGenerator("indexOf","int");
		getIndex.addVariable("CutpointElement","c");
		code="return cutpoints.indexOf(c);"+JavaGenerator.NEWLINE;
		getIndex.setImplementation(code);
		getIndex.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getIndex);
		
		MethodGenerator setEnabled = new MethodGenerator("setEnabled","void");
		setEnabled.addVariable("boolean","b");
		code="setEnabledImpl(b);"+JavaGenerator.NEWLINE;
		code+="ConnectorPersistance.saveConnector(this);"+JavaGenerator.NEWLINE;
		//code+="firePropertyChangeEvent(new HookPropertyChangeEvent(HookPropertyChangeEvent.IS_ENABLED,new Boolean(b)));"+JavaGenerator.NEWLINE;
		setEnabled.setImplementation(code);
		setEnabled.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(setEnabled);
		
		MethodGenerator setEnabledImpl = new MethodGenerator("setEnabledImpl","void");
		setEnabledImpl.addVariable("boolean","b");
		code="isEnabled=b;"+JavaGenerator.NEWLINE;
		code+="firePropertyChangeEvent(new HookPropertyChangeEvent(HookPropertyChangeEvent.IS_ENABLED,new Boolean(b)));"+JavaGenerator.NEWLINE;
		setEnabledImpl.setImplementation(code);
		setEnabledImpl.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(setEnabledImpl);
		
		MethodGenerator firePropertyChangeEvent = new MethodGenerator("firePropertyChangeEvent","void");
		firePropertyChangeEvent.addVariable("HookPropertyChangeEvent","event");
		code="Enumeration i = getCutpoints();"+JavaGenerator.NEWLINE;
		code+="while(i.hasMoreElements()) {"+JavaGenerator.NEWLINE;
		String tmpcode = "CutpointElement next = (CutpointElement) i.nextElement();"+JavaGenerator.NEWLINE;
		tmpcode+="next.getCutpoint().globalPropertyChangeEvent(event);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.addTabsBeforeLines(tmpcode);
		code+="}"+JavaGenerator.NEWLINE;
		firePropertyChangeEvent.setImplementation(code);
		firePropertyChangeEvent.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(firePropertyChangeEvent);
		
		MethodGenerator isEnabled = new MethodGenerator("isEnabled","boolean");
		code="return isEnabled;"+JavaGenerator.NEWLINE;
		isEnabled.setImplementation(code);
		isEnabled.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(isEnabled);
		
		VariableGenerator isEnabledeVar = new VariableGenerator("boolean","isEnabled");
		isEnabledeVar.addModifier(Modifier.PRIVATE);
		isEnabledeVar.setInitialiser("true");
		connectorClass.addVariable(isEnabledeVar);
		
		VariableGenerator doNotCacheF = new VariableGenerator("boolean","doNotCache");
		doNotCacheF.addModifier(Modifier.PRIVATE);
		doNotCacheF.setInitialiser("false");
		connectorClass.addVariable(doNotCacheF);
		
		MethodGenerator doNotCache = new MethodGenerator("doNotCache","boolean");
		code="return doNotCache;"+JavaGenerator.NEWLINE;
		doNotCache.setImplementation(code);
		doNotCache.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(doNotCache);
		
		MethodGenerator setDoNotCache = new MethodGenerator("setDoNotCache","void");
		setDoNotCache.addVariable("boolean","b");
		code="doNotCache=b;"+JavaGenerator.NEWLINE;
		setDoNotCache.setImplementation(code);
		setDoNotCache.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(setDoNotCache);
				
		MethodGenerator getCutpoints = new MethodGenerator("getCutpoints","Enumeration");
		code="return cutpoints.elements();"+JavaGenerator.NEWLINE;
		getCutpoints.setImplementation(code);
		getCutpoints.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getCutpoints);
		
		MethodGenerator getCombinationStrategies = new MethodGenerator("getCombinationStrategies","Enumeration");
		code="return combinationStrategies.elements();"+JavaGenerator.NEWLINE;
		getCombinationStrategies.setImplementation(code);
		getCombinationStrategies.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getCombinationStrategies);
		
		MethodGenerator getHooks = new MethodGenerator("getHooks","Enumeration");
		code="return getCutpoints();"+JavaGenerator.NEWLINE;
		getHooks.setImplementation(code);
		getHooks.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getHooks);
		
		MethodGenerator getInstances = new MethodGenerator("getInstances","Enumeration");
		code="return instances.elements();"+JavaGenerator.NEWLINE;
		getInstances.setImplementation(code);
		getInstances.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getInstances);
		
		VariableGenerator cutpoints = new VariableGenerator("Vector","cutpoints");
		cutpoints.addModifier(Modifier.PRIVATE);
		//cutpoints.addModifier(Modifier.STATIC);
		cutpoints.setInitialiser("new Vector()");
		connectorClass.addVariable(cutpoints);
		
		
		VariableGenerator methodHash = new VariableGenerator("Vector","methodHashtableVector");
		methodHash.addModifier(Modifier.PRIVATE);
		//cutpoints.addModifier(Modifier.STATIC);
		methodHash.setInitialiser("new Vector()");
		connectorClass.addVariable(methodHash);
		
		VariableGenerator filename = new VariableGenerator("String","filename");
		filename.addModifier(Modifier.PRIVATE);
		//isClassAdapter.addModifier(Modifier.STATIC);
		//isClassAdapter.setInitialiser("");
		connectorClass.addVariable(filename);
		
		MethodGenerator filenameGM = new MethodGenerator("getFilename","String");
		code="return filename;"+JavaGenerator.NEWLINE;
		filenameGM.setImplementation(code);
		filenameGM.addModifier(Modifier.PUBLIC);
		//isClassAdapterM.addModifier(Modifier.STATIC);
		connectorClass.addMethod(filenameGM);
		
		MethodGenerator filenameSM = new MethodGenerator("setFilename","void");
		filenameSM.addVariable("String","filename");
		code="this.filename=filename;"+JavaGenerator.NEWLINE;
		filenameSM.setImplementation(code);
		filenameSM.addModifier(Modifier.PUBLIC);
		//isClassAdapterM.addModifier(Modifier.STATIC);
		connectorClass.addMethod(filenameSM);
		
		MethodGenerator getPrio = new MethodGenerator("getPriority","int");
		code="return priority;"+JavaGenerator.NEWLINE;
		getPrio.setImplementation(code);
		getPrio.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(getPrio);
		
		MethodGenerator setPrio = new MethodGenerator("setPriority","void");
		setPrio.addVariable("int","p");
		code="this.priority=p;"+JavaGenerator.NEWLINE;
		code+="ConnectorRegistry.reOrderConnectors();"+JavaGenerator.NEWLINE;
		setPrio.setImplementation(code);
		setPrio.addModifier(Modifier.PUBLIC);
		connectorClass.addMethod(setPrio);
		
		VariableGenerator prio = new VariableGenerator("int","priority");
		prio.addModifier(Modifier.PRIVATE);
		connectorClass.addVariable(prio);
		
		PCutpointDeclarations declarations = connector.getCutpointDeclarations();
		for (int i = 0; i < declarations.getCutpointDeclarationsSize() ; i++) 
		{
			PCutpointDeclaration declaration = declarations.getCutpointDeclarations(i);
			generateCutpointDeclaration(declaration,connectorClass);
			nameMap.put(declaration.getCutpointDeclarationName(),declaration);
		}
		
		generateAnnotations(connectorClass);
		generateCutpointMethodExecutions(connectorClass);
		
		//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
		generateCutpointNaming(connectorClass);
		generateAbstractMethodImplemenations(connectorClass);
		generateSetConnectors(connectorClass);
		//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	
		for (int i = 0 ; i < connector.getImports().getImportsSize() ; i++)  
		 { 
		     PImport imported = connector.getImports().getImport(i); 
		     connectorClass.addImportPackage(new GImport(imported.getImportName(),"" + imported.getLineNumber())); 
	   }
	   
	  
		
		
		PCutpointExecutions executions = connector.getCutpointExecutions();
		for (int i = 0; i < executions.getCutpointExecutionsSize() ; i++) 
		{
			
			PCutpointExecution execution = executions.getCutpointExecution(i);
			generateCutpointExecution(execution,connectorClass);
			
		}
		
		for (int i = 0; i < declarations.getCutpointDeclarationsSize() ; i++) 
		{
			PCutpointDeclaration declaration = declarations.getCutpointDeclarations(i);
			String name = declaration.getCutpointDeclarationName();
			if (!cutpointnames.contains(name))  {
				
				cutpointnames.add(name); 
				cutpointbefore.add("");
				cutpointafter.add("");
				cutpointreplace.add("");
				PCutpointExecution execBefore = new PCutpointExecution();
				execBefore.setCutpointAdaptationType(PCutpointExecution.DEFAULT);
				execBefore.setExecutionNumber(Options.HOOKS_MAX_VALUE);
				execBefore.setCutpointExecutionName(name);
				addExecution(execBefore,connectorClass,cutpointnames.size()-1);
			} 	
			
		}
		
		generateCutpointElements(connectorClass);
		
		for (int i = 0; i < declarations.getCutpointDeclarationsSize() ; i++) 
		{	
			PCutpointDeclaration declaration = declarations.getCutpointDeclarations(i);
			generateMethodWithCutpoints(declaration,connectorClass);
		}
		
		addCutpointFinder(connectorClass);
		
		generateAspectInitMethod(connectorClass);
		
		generateLoadMethod(connectorClass);
		
		generateVerifyCombinationsMethod(connectorClass);
		
		
		
		try  {
			File filetowrite = new File(Options.getOutputDir() +"/"+Options.getConnectorPackage().replace('.','/') + "/"+connectorname +".java");
			if(filetowrite.exists())
				filetowrite.delete();
			else if(filetowrite.getParent()!=null)
				new File(filetowrite.getParent()).mkdirs();	
			BufferedWriter out = new BufferedWriter(new FileWriter(filetowrite));
			out.write(connectorClass.toString());
			out.close();
			
			
			//retrieve the source and classpath
			String classpath = Options.getClassPath();	
			String source = (new File(Options.getOutputDir()).getCanonicalPath()) + Options.FILE_SEPARATOR + Options.getConnectorPackage().replace('.',File.separatorChar) +Options.FILE_SEPARATOR+ connectorname + ".java";
						
			//compile all files
			JavaCompiler compile = new JavaCompiler();
			
			compile.compile(source,classpath);
			
			Vector errors = compile.getErrors();
			this.lastErrors=errors;
	
			//translate the errors to the aspect source instead of the class source
			translateAndPrint(errors);
			
			if(Options.deleteTempFiles())
				filetowrite.delete();
			
		}
		catch(Exception e)
		{
			Logger.getInstance().showFullError(e); 
		
		}
		
		if(!jasco.options.Options.transformBean())
			return;
		
		for (int i = 0; i < declarations.getCutpointDeclarationsSize() ; i++) 
		{	
			PCutpointDeclaration declaration = declarations.getCutpointDeclarations(i);
			adaptMethods(declaration,connectorClass);
		}	
	}
	
	private void generateAnnotations(ClassGenerator connectorClass) {
		
		Iterator i = connector.getAnnotations();
		while(i.hasNext()) {
			connectorClass.addAnnotation(i.next().toString());
		}
	}
	
	public Iterator getLastErrors()  {
		
		if(lastErrors!=null)
			return lastErrors.iterator();
		else return null;	
	}
	
	protected PCutpointDeclaration getCutpointDelcaration4name(String name)  {
		
		PCutpointDeclarations declarations = connector.getCutpointDeclarations();
		for (int i = 0; i < declarations.getCutpointDeclarationsSize() ; i++) 
			if(declarations.getCutpointDeclarations(i).getCutpointDeclarationName().equals(name)) 
				return declarations.getCutpointDeclarations(i);
		
		throw new IllegalArgumentException("Problem with finding declaration: "+name);
		//return null;
	}	
	
	protected void generateVerifyCombinationsMethod(ClassGenerator cg)  {
		
		String code = "Enumeration combinations = combinationStrategies.elements();"+JavaGenerator.NEWLINE;
		code+="while(combinations.hasMoreElements()) {" + JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"CombinationStrategy strategy = (CombinationStrategy) combinations.nextElement();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"hooks=strategy.validateCombinations(hooks);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		code+="return hooks;"+JavaGenerator.NEWLINE;
					
		MethodGenerator checkCode = new MethodGenerator("verifyCombinations","HookList");
		checkCode.addVariable("HookList","hooks");
		checkCode.addModifier(Modifier.PROTECTED);
		checkCode.setImplementation(code);
		cg.addMethod(checkCode);
	}
			
	
	protected void generateCutpointMethodExecutions(ClassGenerator cg)  
    {  
          String code = generateJavaCode(); 
          Vector methodexecutions = connector.getCutpointMethodExecutions();   
          for (int i = 0; i < methodexecutions.size() ; i++)   
          {    
		  	String exe = (String)methodexecutions.elementAt(i);
		
            code = code + "myinstance" + "." + exe + JavaGenerator.NEWLINE;  
			int index = exe.indexOf(".");  
			String name = exe.substring(0,index);
			PCutpointDeclaration declaration = (PCutpointDeclaration) nameMap.get(name);
			if (declaration.getPerObjectModifier())  {
				String icode = "";
				String fullClassName = declaration.getClassName();
			
				icode+="myinstance" + "." + name + "_cutpoint." + "addProperty(\"\"+DynamicCutpointElement.INITIALIZER+"+getNextIndex()+", new HookInitializer() {"+ JavaGenerator.NEWLINE;
				icode+=JavaGenerator.TAB+"public void init(Cutpoint c) {"+ JavaGenerator.NEWLINE;
				icode+=JavaGenerator.TAB+JavaGenerator.TAB+fullClassName+" "+name+" = ("+fullClassName+")c;"+JavaGenerator.NEWLINE;
				icode+=JavaGenerator.TAB+JavaGenerator.TAB+exe+JavaGenerator.NEWLINE;
				icode+="}});"+ JavaGenerator.NEWLINE;
				
				code+=icode;
			}	
          }  
          addAspectInitCode(code);
		  
    }
	
	protected void addAspectInitCode(String code)  {
		
		staticInitCode.add(code);
	}
		
	
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	protected void generateCutpointNaming(ClassGenerator cg) 
	{
		String code = "";
		PCutpointDeclarations declarations = connector.getCutpointDeclarations();
		for (int i = 0 ; i < declarations.getCutpointDeclarationsSize(); i++) 
		{
			PCutpointDeclaration c = declarations.getCutpointDeclarations(i);
			String name = c.getCutpointDeclarationName();
			
			code = code + "myinstance" + "." + name + "." + "_Jasco_setCutpointName(\"" + name + "\");" + JavaGenerator.NEWLINE;
			
			PCutpointDeclaration declaration = (PCutpointDeclaration) nameMap.get(name);
			if(declaration.getPerObjectModifier())
				code+="myinstance" + "." + name + "_cutpoint." + "addProperty(\"_Jasco_setCutpointName\", \"" + name + "\");" + JavaGenerator.NEWLINE;
		}
		cg.addStaticInit(code);
	}
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	protected void generateSetConnectors(ClassGenerator cg) 
	{
		String code = "";
		PCutpointDeclarations declarations = connector.getCutpointDeclarations();
		for (int i = 0 ; i < declarations.getCutpointDeclarationsSize(); i++) 
		{
			PCutpointDeclaration c = declarations.getCutpointDeclarations(i);
			String name = c.getCutpointDeclarationName();
			code = code + "myinstance" + "." + name + "." + "_Jasco_setConnector(myinstance);" + JavaGenerator.NEWLINE;
			PCutpointDeclaration declaration = (PCutpointDeclaration) nameMap.get(name);
			if (declaration.getPerObjectModifier())  {
				code+="myinstance" + "." + name + "_cutpoint." + "addProperty(\"_Jasco_setConnector\",  myinstance );" + JavaGenerator.NEWLINE;
				code+="myinstance" + "." + name + "_cutpoint." + "addProperty(\"_Jasco_setGlobal\", myinstance" + "." + name + "._Jasco_getGlobal());" + JavaGenerator.NEWLINE;
			}
		}
		addAspectInitCode(code);
	}
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
	
	
	
	
	
	//should be in connectorparser!!!!!!!!!!//
	protected void generateAbstractMethodImplemenations(ClassGenerator cg) 
	{	
		PCutpointDeclarations declarations = connector.getCutpointDeclarations();
		for (int i = 0 ; i < declarations.getCutpointDeclarationsSize(); i++) 
		{
			PCutpointDeclaration c = declarations.getCutpointDeclarations(i);
			generateAbstractMethodImplemenations(c,cg);
		}
	}
	
	protected void generateAbstractMethodImplemenations(PCutpointDeclaration decl, ClassGenerator cg) 
	{	
		Vector abstracts = null;
		Vector abstractsToImplement = null;
		try {
			Field f = Options.loadClass(decl.getFullClassName()).getField(AspectClassesGenerator.ABSTRACT_METHODS);
			if(f!=null) { //backward compatible...
				Vector abstractsorig = (Vector) f.get(null);
				abstracts = (Vector) abstractsorig.clone();
			}
			Field f2 = Options.loadClass(decl.getFullClassName()).getField(AspectClassesGenerator.ABSTRACT_METHODS_IMPL);
			if(f2!=null) { //backward compatible...
				Vector abstractsorig = (Vector) f2.get(null);
				abstractsToImplement = (Vector) abstractsorig.clone();
			}
		} catch(Exception ex) {}
		 
		Iterator abstractmethods = decl.getAbstractImplementations();
		while(abstractmethods.hasNext()) {
			PAbstractImplementation method = (PAbstractImplementation) abstractmethods.next();
			MethodGenerator methodimpl = method.getAbstractImplementation().getMethodInfo();
			int modifiers = methodimpl.getModifiers();
			modifiers = javassist.Modifier.setPublic(modifiers);
			methodimpl.setModifiers(modifiers);
			
			String methodname = methodimpl.getName();
			methodimpl.addVariable(method.getCutpointType(),method.getCutpointName());
			methodimpl.setName(methodname+decl.getCutpointDeclarationName());
			cg.addMethod(methodimpl);
			
			if(abstracts!=null&&!abstracts.contains(methodname))
				throw new IllegalArgumentException("Refinable method "+methodname+" of "+decl.getCutpointDeclarationName()+" does not exist!");
			else if(abstractsToImplement!=null) abstractsToImplement.remove(methodname);
		}
		
		if(abstractsToImplement!=null) {
		Iterator i = abstractsToImplement.iterator();
		while(i.hasNext()) {
			throw new IllegalArgumentException("Refinable method "+i.next()+" of "+decl.getCutpointDeclarationName()+" is not implemented!");
		}
		}
		
	}
	//!!!!!!!!!!!!!!!!!!!!new CODE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
			
	protected String generateJavaCode()  
    { 
         PJavaCodes javacode = connector.getAspectJavaCode(); 
         String code = "//Java Code"+JavaGenerator.NEWLINE; 
         for (int i = 0 ; i < javacode.getJavaCodeSize(); i++)  
         { 
		 	
              PJavaCode codes = javacode.getJavaCode(i); 
              for (int j = 0; j < codes.getJavaCodeSize() ; j++)  
              { 
                   code = code + codes.getJavaCode(j) + "//" + codes.getLineNumber(j) + "\n"; 
              } 
         } 
         return code; 
    }

	
	protected void generateLoadMethod(ClassGenerator cg)  {
		
		StringBuffer lcode = new StringBuffer();
		lcode.append("synchronized(myinstance) {");
		lcode.append("getConnector().setFilename(filename);"+JavaGenerator.NEWLINE);
		
		Iterator i = loadCode.iterator();
		while (i.hasNext())  {
			lcode.append(i.next().toString());
		}
		lcode.append("ConnectorRegistry.registerConnector(myinstance);"+JavaGenerator.NEWLINE);
		lcode.append("myinstance.firePropertyChangeEvent(new HookPropertyChangeEvent(HookPropertyChangeEvent.IS_ADDED));"+JavaGenerator.NEWLINE);
		lcode.append("}");
		
		
		MethodGenerator load = new MethodGenerator("load","void");
		load.addVariable("String","filename");
		load.setImplementation(lcode.toString());
		load.addModifier(Modifier.PUBLIC);
		load.addModifier(Modifier.STATIC);
		//load.addModifier(Modifier.SYNCHRONIZED);
		cg.addMethod(load);
	}
	
	private void translateAndPrint(Vector aErrors)
	{
	
		String source = connector.getConnectorName() + ".con";
		
		for (int i = 0 ; i < aErrors.size() ; i++) 
		{
			//Get error and errorstring;
			CompileError error = (CompileError)aErrors.elementAt(i);
			String errorstring = error.getErrorString();
			
			
			//Get linenumber of error in original file
			int documentationindex = errorstring.lastIndexOf("//");
			if (documentationindex==-1)  {
				//System.out.println("!!!!!!!! JAsCo System Error: !!!!!!!!");
				//System.out.println(error.toString());
				//CompileError newerror = new CompileError(source,-1,errorstring);
				//Logger.getInstance().showFullError(newerror.toString());
				
				
				Logger.getInstance().showError(error);
				
				continue;
			}
			
			
			
			
			int restindex = errorstring.indexOf(Options.NEWLINE,documentationindex);
			String linenumber = errorstring.substring(documentationindex + 2,restindex);
			
			int reallinenumber = (new Integer(linenumber)).intValue();
			
			
			//Get errormessage by extracting documentation
			String realerror = errorstring.substring(0,documentationindex) + Options.NEWLINE + errorstring.substring(restindex + 1); 
			
			CompileError newerror = new CompileError(source,reallinenumber,realerror);
			
			//System.out.print(newerror.toString());
			Logger.getInstance().showError(newerror);
		}
		if (aErrors.size() != 0)
		{
			if (aErrors.size() == 1)
			{
				//System.out.println(1 + " error");
				Logger.getInstance().showOutput(1 + " error");
			}
			else 
			{
				//System.out.println(aErrors.size() + " errors");
				Logger.getInstance().showOutput(aErrors.size() + " errors");
			}
		}
	
	}
	/*
	private void translateAndPrint(Vector aErrors)
	{
		String source = connector.getConnectorName() + ".con";
		
		for (int i = 0 ; i < aErrors.size() ; i++) 
		{
			//Get error and errorstring;
			CompileError error = (CompileError)aErrors.elementAt(i);
			String errorstring = error.getErrorString();
			
			//Get linenumber of error in original file
			int documentationindex = errorstring.lastIndexOf("//");
			if (documentationindex==-1)  {
				System.out.println("!!!!!!!! JAsCo System Error: !!!!!!!!");
				System.out.println(error.toString());
				continue;
			}
			
			int restindex = errorstring.indexOf("\n",documentationindex);
			String linenumber = errorstring.substring(documentationindex + 2,restindex);
			
			int reallinenumber = (new Integer(linenumber)).intValue();
			
			
			//Get errormessage by extracting documentation
			String realerror = errorstring.substring(0,documentationindex) + "\n" + errorstring.substring(restindex + 1); 
			
			CompileError newerror = new CompileError(source,reallinenumber,realerror);
			
			System.out.print(newerror.toString());
		}
		if (aErrors.size() != 0)
		{
			if (aErrors.size() == 1)
			{
				System.out.println(1 + " error");
			}
			else 
			{
				System.out.println(aErrors.size() + " errors");
			}
		}
	}
	*/
	private void generateTranformMethod(ClassGenerator cl)  {
	
		MethodGenerator method = new MethodGenerator("transformMethods","void");
		method.addModifier(Modifier.PROTECTED);
		method.addModifier(Modifier.STATIC);
		
		String code = "";
		code = "Vector v = (Vector) myinstance.methods.elementAt(0);"+JavaGenerator.NEWLINE;
		code+="for(int i=0; i<v.size(); ++i)"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"myinstance.methodHashtable.put(i+\"\" , v.elementAt(i));"+JavaGenerator.NEWLINE;
		method.setImplementation(code);
		
		cl.addMethod(method);
	}
		
		
	
	
	private void generateCutpointDeclaration(PCutpointDeclaration aCutpointDeclaration, ClassGenerator connectorClass)
	{	
		if(!aCutpointDeclaration.getAspectPackage().equals(""))
			connectorClass.addImportPackage(new GImport(aCutpointDeclaration.getAspectPackage()+".*"));
		String aspecttype = aCutpointDeclaration.getAspectType();
		String cutpointtype = aCutpointDeclaration.getCutpointType();
		String cutpointname = aCutpointDeclaration.getCutpointDeclarationName();
		String theglobalaspect = aspecttype;// + "_global";
		if (!generatedGlobal(theglobalaspect))
		{ 
			addGeneratedGlobal(theglobalaspect);
			VariableGenerator global = new VariableGenerator(theglobalaspect,theglobalaspect);
			global.addModifier(Modifier.PUBLIC);
			//global.addModifier(Modifier.STATIC);
			global.setInitialiser("new "+theglobalaspect+"()");
			connectorClass.addVariable(global);
		}
		
		String cutpointClassNameWithoutPackage = aspecttype + "_" + cutpointtype;
		String cutpointClassName = aCutpointDeclaration.getFullClassName();
		VariableGenerator cutpoint = new VariableGenerator(cutpointClassNameWithoutPackage,cutpointname);
		cutpoint.addModifier(Modifier.PUBLIC);
		//cutpoint.addModifier(Modifier.STATIC);
		cutpoint.setInitialiser("new "+cutpointClassName+"("+theglobalaspect+")");
		connectorClass.addVariable(cutpoint);
		/*
		try 
		{
			String aspecttype = aCutpointDeclaration.getAspectType();
			String cutpointtype = aCutpointDeclaration.getCutpointType();
			String cutpointname = aCutpointDeclaration.getCutpointDeclarationName();
			String theglobalaspect = aspecttype + "_global";
			if (!generatedGlobal(theglobalaspect))
			{ 
				addGeneratedGlobal(theglobalaspect);
				aOut.write("\t" + "public static " + aspecttype + "_global " + aspecttype + "_global = new " + aspecttype + "_global();");
				aOut.newLine();
			}
			aOut.write("\t" + "public static " + aspecttype + "_" + cutpointtype + " " + cutpointname + " = new " + aspecttype + "_" + cutpointtype + "(" + aspecttype + "_global);");
			aOut.newLine();
		}
		catch (Exception e) 
		{
			System.out.println(e.getMessage()); e.printStackTrace();
		} */
	}
	
	private void generateCutpointExecution(PCutpointExecution aCutpointExecution, ClassGenerator connectorClass) 
     { 
          String cutpointname = aCutpointExecution.getCutpointExecutionName(); 
          for (int i = 0 ; i < cutpointnames.size(); i++)  
          { 
               if (cutpointname.equals(cutpointnames.elementAt(i)))  
               { 
                    addExecution(aCutpointExecution,connectorClass,i); 
                    return; 
               } 
          } 
          cutpointnames.add(cutpointname); 
          cutpointbefore.add(new CutpointExecutionElement(false,-1)); 
          cutpointafter.add(new CutpointExecutionElement(false,-1)); 
          cutpointreplace.add(new CutpointExecutionElement(false,-1)); 
          addExecution(aCutpointExecution,connectorClass,cutpointnames.size() - 1); 
     } 
      
     private void addExecution(PCutpointExecution aCutpointExecution,ClassGenerator connectorClass,int i)  
     { 
          int type = aCutpointExecution.getCutpointAdaptationType(); 
		  String cutpointname = aCutpointExecution.getCutpointExecutionName();
		  String cutpointType = getCutpointDelcaration4name(cutpointname).getFullClassName();
		  //String cutpointType = connectorClass.getVariable(cutpointname).getType();
		  
		  java.lang.Class testiep = null;
		  try {
		  //testiep = java.lang.Class.forName(cutpointType);
		  testiep = jasco.options.Options.loadClass(cutpointType);
		  
		  }catch(Exception ex)  {
		  	//System.err.println("Error loading hook: "+cutpointType);
		  	throw new IllegalArgumentException("Error loading hook: "+cutpointType);
		  }
		  
		  boolean hasBefore = HookHelper.definesAdvice(testiep, Cutpoint.BEFORE);
		  boolean hasAfter = HookHelper.definesAdvice(testiep, Cutpoint.AFTER);
		  boolean hasReplace = HookHelper.definesAdvice(testiep, Cutpoint.REPLACE);
		  
          switch (type)  { 
           
               case PCutpointExecution.BEFORE: 
			   	
			   		if (!hasBefore)  {
						//System.err.println(cutpointType+" does not define a before!");
						throw new IllegalArgumentException(cutpointType+" does not define a before!");
			   		}
                    cutpointbefore.setElementAt(new CutpointExecutionElement(true,aCutpointExecution.getExecutionNumber()),i); 
                    break; 
                
               case PCutpointExecution.AFTER: 
			   		if (!hasAfter)  {
						//System.err.println(cutpointType+" does not define an after!");
						throw new IllegalArgumentException(cutpointType+" does not define an after!");
			   		}
                    cutpointafter.setElementAt(new CutpointExecutionElement(true,aCutpointExecution.getExecutionNumber()),i); 
                    break; 
                
               case PCutpointExecution.REPLACE: 
			   	
			   		if (!hasReplace)  {
						//System.err.println(cutpointType+" does not define a replace!");
						throw new IllegalArgumentException(cutpointType+" does not define an around (aka replace) advice !");
			   		}
                    cutpointreplace.setElementAt(new CutpointExecutionElement(true,aCutpointExecution.getExecutionNumber()),i); 
                    break; 
                
               default: // What to do here? 
			   		
                    cutpointbefore.setElementAt(new CutpointExecutionElement(hasBefore,aCutpointExecution.getExecutionNumber()),i); 
                    cutpointafter.setElementAt(new CutpointExecutionElement(hasAfter,aCutpointExecution.getExecutionNumber()),i); 
                    cutpointreplace.setElementAt(new CutpointExecutionElement(hasReplace,aCutpointExecution.getExecutionNumber()),i); 
                    break; 
          } 
     }

	
	public void generateCutpointElements(ClassGenerator connectorClass) 
	{
		for (int i = 0; i < cutpointnames.size() ; i++) 
			{
			String name = (String) cutpointnames.elementAt(i);	
			
			PCutpointDeclaration declaration = (PCutpointDeclaration)nameMap.get(name);
			
				VariableGenerator cpElement = new VariableGenerator("CutpointElement",name + "_cutpoint");
				cpElement.addModifier(Modifier.PRIVATE);
				
				String cutpointType = "CutpointElement";
				if (declaration.getPerObjectModifier())   
					cutpointType="DynamicCutpointElement";
					
				//cpElement.addModifier(Modifier.STATIC);
				cpElement.setInitialiser("new "+cutpointType+" (" + (String)cutpointnames.elementAt(i) + ", " + cutpointbefore.elementAt(i) + ", " + cutpointafter.elementAt(i) + ", " + cutpointreplace.elementAt(i) + ", \""+(String)cutpointnames.elementAt(i)+"\")");
				connectorClass.addVariable(cpElement);
				
/*				aOut.write("\tpublic static CutpointElement " + (String)cutpointnames.elementAt(i) + "_cutpoint = new CutpointElement(" + (String)cutpointnames.elementAt(i) + ", " + cutpointbefore.elementAt(i) + ", " + cutpointafter.elementAt(i) + ", " + cutpointreplace.elementAt(i) + ");");
				aOut.newLine();*/
			
			}
		
	} 
	
	private void generateMethodWithCutpoints(PCutpointDeclaration aCutpointDeclaration, ClassGenerator connectorClass)
	{
		
		String bol = "";
		if (connector.isStaticConnector())  
			bol="true";
		else  { 
		   bol="false";	
		   Logger.getInstance().showWarning("the connector is declared non-static and only triggers on instance-level! (default=static)");
		}
		String code="myinstance.setAdaptOnClassesImpl("+bol+");"+JavaGenerator.NEWLINE;
		
	//	TreeMap methodmappings = new TreeMap();
		String aspecttype = aCutpointDeclaration.getAspectType();
		String cutpointtype = aCutpointDeclaration.getCutpointType();
		String packagename = aCutpointDeclaration.getAspectPackage();
		String cutpointClassName = aCutpointDeclaration.getFullClassName();

		Set calledmethodmatch = new TreeSet();
			
		try {
		java.lang.Class cl = Options.loadClass(cutpointClassName);
		PCutpointConstructorApplicationDesignator object = (PCutpointConstructorApplicationDesignator) cl.getField("_Jasco_cutpointconstructorRep").get(null);
		calledmethodmatch=findCalledMethod(object);
		}
		catch(Exception ex)  {
			
			System.err.println(ex.getMessage());
			ex.printStackTrace();
			throw new IllegalArgumentException("Failed loading aspect:"+packagename+"."+aspecttype);
		}
		ConnectorParameterInputCombinationMaker combinations = new ConnectorParameterInputCombinationMaker();
		
		PCutpointDeclarationParameters cutpointdeclarationparameters = aCutpointDeclaration.getCutpointDeclarationParameters();
		for(int k=0; k<cutpointdeclarationparameters.getCutpointDeclarationParametersSize();++k)  {
			PCutpointDeclarationParameter cutpointdeclarationparameter = cutpointdeclarationparameters.getCutpointDeclarationParameter(k);
			for (int j = 0 ; j < cutpointdeclarationparameter.getCutpointDeclarationParameterInputSize() ; j++) 
			{   
				PCutpointDeclarationParameterInput cutpointdeclarationparameterinput = cutpointdeclarationparameter.getCutpointDeclarationParameterInput(j); 
				/*
				String classname = cutpointdeclarationparameterinput.getClassName();
				String type = cutpointdeclarationparameterinput.getType();
				String methodname = cutpointdeclarationparameterinput.getMethodName();
				String cutpointname = type + " " + classname + "." + methodname + "(";
				if(cutpointdeclarationparameterinput.hasAnnotation())
					cutpointname="@"+cutpointdeclarationparameterinput.getAnnotation()+" "+type + " " + classname + "." + methodname + "(";
						
				for (int i = 0 ; i < cutpointdeclarationparameterinput.getMethodParametersSize() ; i++)
				{
					cutpointname = cutpointname + cutpointdeclarationparameterinput.getMethodParameter(i);
					if (i < (cutpointdeclarationparameterinput.getMethodParametersSize() - 1))
					{
						cutpointname = cutpointname + ", ";
					} 
				}
				cutpointname+=")";
				
				
				*/
				String pos = k+"";
				
				String nextVar = JavaNameGenerator.getDefault().getNextUniqueName();
				code+="MethodSignature "+nextVar+" = "+cutpointdeclarationparameterinput.generateMethodSignatureBuilder()+";"+JavaGenerator.NEWLINE;
				/* only required when matching on strings for cflow/withincode
				if(!calledmethodmatch.contains(pos))
					res=transformParam(cutpointname);*/
				combinations.putMapping(pos,nextVar);
			}
			combinations.nextRound();
		}	
		
		boolean first = true;
		Iterator combI = combinations.getResultingMappings();
		while (combI.hasNext())  {
			TreeMap methodmappings = (TreeMap) combI.next();
			
			
			code+="{Hashtable methodHashtable=new Hashtable();"+JavaGenerator.NEWLINE;
		
			Iterator i = methodmappings.entrySet().iterator();
			while (i.hasNext())  {
				Map.Entry next = (Map.Entry) i.next();
				code+="methodHashtable.put(\""+next.getKey()+"\", "+next.getValue()+");"+ JavaGenerator.NEWLINE;  
			}
			
			code+="myinstance.methodHashtableVector.addElement(methodHashtable);"+JavaGenerator.NEWLINE;
			Iterator calledmethodmatchI = calledmethodmatch.iterator();
			while (calledmethodmatchI.hasNext())  {
				String pos = (String) calledmethodmatchI.next();
				code+="myinstance.methods.addElement("+methodmappings.get(pos)+");"+JavaGenerator.NEWLINE;
				
			}
			
			if (!first)  {
				code+="CutpointElement _JAsCo_tempElement = myinstance."+ aCutpointDeclaration.getCutpointDeclarationName() + "_cutpoint.copy()" + ";"+JavaGenerator.NEWLINE;
				code+="myinstance.cutpoints.addElement(" + "_JAsCo_tempElement" + ");"+JavaGenerator.NEWLINE;
			}
			else  {
				code+="myinstance.cutpoints.addElement(" +"myinstance."+ aCutpointDeclaration.getCutpointDeclarationName() + "_cutpoint" + ");"+JavaGenerator.NEWLINE;			
				first=false;
			}	
			code+="myinstance."+aCutpointDeclaration.getCutpointDeclarationName() + "_cutpoint"+".setCutpointContext(myinstance."+aCutpointDeclaration.getCutpointDeclarationName()+"._Jasco_cutpointconstructorRep);"+JavaGenerator.NEWLINE;
			
			code+="}"+JavaGenerator.NEWLINE;
		}
		addAspectInitCode(code);
	}
	
	/*
	protected void generateAnnotationCode(PCutpointDeclaration aCutpointDeclaration, PCutpointDeclarationParameterInput p, String methodname) {
		
		if(!p.hasAnnotation())
			return;
		
		String annotationString = p.getAnnotation();
		String code = "";
		String getConstr = "myinstance."+aCutpointDeclaration.getCutpointDeclarationName() + "._Jasco_cutpointconstructorRep";
		code+=getConstr+" = new jasco.runtime.aspect.PCutpointConstructorAnd("+getConstr+", new jasco.runtime.aspect.PCutpointConstructorAnnotation(\""+annotationString+"\"" +
				", \""+methodname+"\"));"+JavaGenerator.NEWLINE;
		
		addAspectInitCode(code);
	}*/
	
	protected void generateAspectInitMethod(ClassGenerator cg)  {
	
		Iterator i = staticInitCode.iterator();
		StringBuffer buffer = new StringBuffer();
		
		if(connector.getAnnotations().hasNext())
			buffer.append("myinstance.priority=jasco.runtime.connector.ConnectorManager.initPriority(myinstance);"+JavaGenerator.NEWLINE);
		
		buffer.append("myinstance.methods.clear();"+JavaGenerator.NEWLINE);
		buffer.append("myinstance.combinationStrategies.clear();"+JavaGenerator.NEWLINE);
		//buffer.append("myinstance.instances.clear();"+JavaGenerator.NEWLINE); Pacosuite problemn...
		buffer.append("myinstance.cutpoints.clear();"+JavaGenerator.NEWLINE);
		buffer.append("myinstance.methodHashtableVector.clear();"+JavaGenerator.NEWLINE);
		
		
		while (i.hasNext())  {
			String s = (String) i.next();
			buffer.append(s);
		}
		
		
		
		
		
		MethodGenerator generateAspect = new MethodGenerator(ASPECT_INIT_METHOD,"void");
		generateAspect.addModifier(Modifier.PROTECTED);
		generateAspect.addModifier(Modifier.STATIC);
		generateAspect.setImplementation(buffer.toString());
		
		cg.addMethod(generateAspect);
		
		
		loadCode.insertElementAt(""+ASPECT_INIT_METHOD+"();"+JavaGenerator.NEWLINE,0);
	}
	
	protected String transformParam(String param)  {
		
		/*try {
		
		int i = param.indexOf("_");
		param = param.substring(i+1);
		i = param.indexOf("_");
		param =param.substring(0,i)+ "."+param.substring(i+1);
		i = param.indexOf("_");
		param = param.substring(0,i);
		
		}
		catch(Exception ex)  {
		System.out.println(ex.getMessage());
		ex.printStackTrace();
		}
		return param;*/
		
		int i = param.indexOf("(");
		param = param.substring(0,i);
		i = param.indexOf(" ");
		param = param.substring(i+1);
		return param;
	}
	
	protected Set findCalledMethod(PCutpointConstructorApplicationDesignator d)  {
	
		Queue q = new Queue();
		q.enqueue(d);
		Set res = new TreeSet();
		
		while (!q.isEmpty())  {
			PCutpointConstructorApplicationDesignator des =(PCutpointConstructorApplicationDesignator) q.dequeue();
			if(des instanceof PCutpointConstructorExecute)
				res.add(((PCutpointConstructorExecute)des).getCalledMethod());
			else if(des instanceof PCutpointConstructorCall)
				res.add(((PCutpointConstructorCall)des).getCalledMethod());
			
			Iterator i = des.allPointcuts().iterator();
			while (i.hasNext())  {
				Object o = i.next();
				if(!q.contains(o))
					q.enqueue(o);
			}
		}
		return res;
	}	
		
		
	public void addGeneratedGlobal(String aGlobal) 
	{
		generatedglobals.add(aGlobal);
	}
	
	public void addCutpointFinder(ClassGenerator connectorClass) 
	{
		
		MethodGenerator method = new MethodGenerator("findCutpointForMethod","Vector");
		method.addModifier(Modifier.PUBLIC);
		//method.addModifier(Modifier.STATIC);
		//method.addVariable("String","aMethod");
		method.addVariable("Object","anInstance");
		method.addVariable("RuntimeContext","context");
		method.addVariable("JascoMethod","aJascoMethod");
		method.addVariable("int","startID");
		method.addVariable("boolean","forJutta");
		String code="";
		code+="Vector results = new Vector();"+JavaGenerator.NEWLINE;
		code+="if(!isEnabled()) return results;"+JavaGenerator.NEWLINE;
		code+="if(!adaptOnClasses() && anInstance!=null && !instances.contains(anInstance)) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"return results;"+JavaGenerator.NEWLINE+"}"+JavaGenerator.NEWLINE;
		
		code+="for (int i = 0; i < cutpoints.size() ; i++) { "+JavaGenerator.NEWLINE;
		//code+=JavaGenerator.TAB+"Vector mappedMethods = (Vector)methods.elementAt(i);"+JavaGenerator.NEWLINE;
		//code+=JavaGenerator.TAB+"if (((String)mappedMethods.elementAt(0)).equals(aMethod)) { "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+JavaGenerator.TAB+"CutpointElement element = (CutpointElement) cutpoints.elementAt(i);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+JavaGenerator.TAB+"if(element.isApplicable(context,getMethodHashtable(i),aJascoMethod,forJutta))"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+JavaGenerator.TAB+JavaGenerator.TAB+"results.addElement(element.transformOrder(startID));"+JavaGenerator.NEWLINE; 
		//code+=JavaGenerator.TAB+"}"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		code+="HookList hooks = new HookList(results);";
		code+="HookList returnlist = verifyCombinations(hooks);";
		code+="return returnlist.getCutpointElements();"+JavaGenerator.NEWLINE;
		method.setImplementation(code);
		connectorClass.addMethod(method);
		/*
		try 
		{
			aOut.write("\tpublic static Vector findCutpointForMethod(String aMethod) { Vector results = new Vector(); for (int i = 0; i < methods.size() ; i++) { if (((String)methods.elementAt(i)).equals(aMethod)) { results.add((CutpointElement)cutpoints.elementAt(i)); } } return results; }");
			aOut.newLine();
		}
		catch (Exception e) 
		{
			System.out.println(e.getMessage()); e.printStackTrace();
		}*/
	}
	
	public void adaptMethods(PCutpointDeclaration aCutpointDeclaration, ClassGenerator connectorClass)
	{
		PCutpointDeclarationParameters cutpointdeclarationparameters = aCutpointDeclaration.getCutpointDeclarationParameters();
		
		
		for(int k=0;k<cutpointdeclarationparameters.getCutpointDeclarationParametersSize();++k) {
			
		PCutpointDeclarationParameter cutpointdeclarationparameter = cutpointdeclarationparameters.getCutpointDeclarationParameter(k);
		for (int j = 0 ; j < cutpointdeclarationparameter.getCutpointDeclarationParameterInputSize() ; j++) 
		{ 
			PCutpointDeclarationParameterInput cutpointdeclarationparameterinput = cutpointdeclarationparameter.getCutpointDeclarationParameterInput(j); 
			String classname = cutpointdeclarationparameterinput.getClassName();
			String type = cutpointdeclarationparameterinput.getType();
			String methodname = cutpointdeclarationparameterinput.getMethodName();
			String cutpointname = type + "_" + classname + "_" + methodname + "_";
			for (int i = 0 ; i < cutpointdeclarationparameterinput.getMethodParametersSize() ; i++)
			{
				cutpointname = cutpointname + cutpointdeclarationparameterinput.getMethodParameter(i);
				if (i < (cutpointdeclarationparameterinput.getMethodParametersSize() - 1))
				{
					cutpointname = cutpointname + "_";
				}
			}
			
			/*if (theclass.getFieldPlace(method + "alreadyadapted") == -1 && theclass.getFieldPlace(jasco.tools.BeanTransformer.IS_JASCO_BEAN) == -1) 
			{
				generateNewCalledMethod(type,methodname, cutpointdeclarationparameterinput.getMethodParameters(), cutpointname, connectorClass);
				generateNewIntermediateMethod(methodname, classname, cutpointdeclarationparameterinput.getMethodParameters(), connectorClass);
				System.out.println(methodname);
				generateNewMethodName( methodname);
				generateAdaptedField(cutpointname);
				weaver.finish();
			}*/
			
			if (!Options.isJAsCoBean(classname)&&Options.isClass(classname)&&Options.transformBean())  {
				System.out.println("Transforming bean: "+classname);
				jasco.Jasco.transformBeanInOutputDir(classname,new Vector());//lege vector wordt meegegeven
				
			}
		}
		}
	}
	
	/*
	public void generateNewCalledMethod(String aType, String aMethodName, Vector vars, String aMethodString, ClassGenerator connectorClass) 
	{
		String method = "";
		method = method + "\tpublic " + aType + " " +  aMethodName + "(";
		for (int i = 0 ; i < vars.size() ; i++) 
		{
			method = method + (String)vars.elementAt(i) + " var" + i;
			if (i < (vars.size() - 1))
			{
				method = method + ", ";
			}
		}
		method = method + ") {\n";
		method = method + "\t\tjava.util.Vector args = new java.util.Vector();\n";
		for (int j = 0 ; j < vars.size() ; j++) 
		{
			method = method + "\t\targs.add(";
			if (isPrimitive((String)vars.elementAt(i)))  
			{
				method = method + "new " + makePrimitiveObject((String)vars.elementAt(j)) + "(" + "var" + j + "));\n";
			}
			else 
			{  	 
				method = method + "var" + j + ");\n";
			}
		}
		method = method + "\t\t";
		if (!aType.equals("void")) 
		{
			method = method + "return (" ;
			if (isPrimitive(aType)) 
			{
				method = method + "(";
				method = method + makePrimitiveObject(aType) + ")";
			}
			else 
			{
				method = method + aType + ")";
			}
		}
		method = method + aMethodName + "(" + "ConnectorRegistry" + ".findCutpointForMethod(\"" + aMethodString + "\", this), args)";
		if (isPrimitive(aType)) 
		{
			method = method + ")." + aType + "Value();\n";
		}
		else 
		{
			method = method + ";\n";
		}
		method = method + "\t}\n";
		jasco.weaver.JavaCode themethod = new jasco.weaver.JavaCode(method,jasco.weaver.JavaCode.METHOD);
	 	weaver.addNewMethod(themethod);
	}
	*/
	public String makePrimitiveObject(String aPrimitive) 
	{
		if (aPrimitive.equals("int")) 
		{
			return "Integer";
		}
		if (aPrimitive.equals("char")) 
		{
			return "Character";
		}
		else 
		{
			String firstletter = aPrimitive.substring(0,1);
			String rest = aPrimitive.substring(1,aPrimitive.length());
			String newfirstletter = firstletter.toUpperCase();
			return newfirstletter + rest;
		}
	}
	
	public boolean isPrimitive(String aPrimitive) 
	{
		boolean result = false;
		if (aPrimitive.equals("boolean")) 
		{
			result = true;
		}
		if (aPrimitive.equals("byte")) 
		{
			result = true;
		}
		if (aPrimitive.equals("char")) 
		{
			result = true;
		}
		if (aPrimitive.equals("short")) 
		{
			result = true;
		}
		if (aPrimitive.equals("int")) 
		{
			result = true;
		}
		if (aPrimitive.equals("long")) 
		{
			result = true;
		}
		if (aPrimitive.equals("float")) 
		{
			result = true;
		}
		if (aPrimitive.equals("double")) 
		{
			result = true;
		}
		return result;
	}
	

	
	

		
	public boolean generatedGlobal(String aGlobal) 
	{
		for (int i = 0; i < generatedglobals.size() ; i++) 
		{
			String global = (String)generatedglobals.elementAt(i);
			if (global.equals(aGlobal))
			{
				return true;
			}
		}
		return false;
	}
	
	public int getNextIndex()  {
	
		return index++;
	}
}			
	
