/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.gui;

import jasco.options.Options;
import jasco.util.AbstractDialog;
import jasco.util.MyFileChooser;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;

public class IntrospectorOptionsDialog extends AbstractDialog {

	private    JFrame     mainWindow;

    private    JTextField   classPathTF    = new JTextField();
	private DefaultListModel classPath = new DefaultListModel();
	
	private	JTabbedPane tabs = new JTabbedPane();
	
	private		MyFileChooser fileChooser;
	private JButton okButton,cancelButton;

	public IntrospectorOptionsDialog(JFrame mainWindow) {
		super(mainWindow, "Introspector Options", 100, 100);
		fileChooser= new MyFileChooser(mainWindow);
		this.mainWindow = mainWindow;
		setupDialog();
		Options.loadOptions();
		pack();						   
	}
	

	protected void setupDialog() {
		getContentPane().setLayout(new BorderLayout());

		tabs.addTab("Class Path",setupClassPathTab());
	
		getContentPane().add("Center",tabs);
		getContentPane().add("South",setupButtonPanel());
	}
	
	
	
	protected JPanel setupButtonPanel() {
	
		JPanel buttons = new JPanel();
		okButton       = new JButton("OK");
		cancelButton   = new JButton("CANCEL");
		okButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ev) {
				dialogApproved();
			} 

		});
		cancelButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ev) {
				dialogCanceled();
			} 

		});
		buttons.setLayout(new FlowLayout());
		buttons.add(okButton);
		buttons.add(cancelButton);
		JButton patch = new JButton("PATCH ALL DIRS");
		patch.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ev) {
				
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				fileChooser.showOpenDialog();
				if(fileChooser.isCanceled()) {
					fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
					return;
				}
				File main = fileChooser.getSelectedFile();
				File file = new File(Options.getInstallPath());
				String testParent = file.getParent();
				
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				classPath.removeAllElements();
				Iterator classPathI = Options.getClassPathVector().iterator();
				while(classPathI.hasNext()) {
					String next = classPathI.next().toString();
					System.out.println(next);
					System.out.println(testParent);
					if(next.toLowerCase().startsWith(testParent.toLowerCase())) {
						String name = next.substring(testParent.length());
						classPath.addElement(new File(main,name).getPath());
					}
					else classPath.addElement(next);
				}
				
					
			}
		});
		
		JPanel patchPanel = new JPanel();
		patchPanel.add(patch);
		
		JPanel main = new JPanel();
		main.setLayout(new BorderLayout());
		main.add(buttons);
		main.add("East",patchPanel);
		
		return main;
	}
	
	
	protected GridBagConstraints createConstraints(int x, int y, int width, int height, double weightx, double weighty,int fill,int anchor, Insets insets)  {
		
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.gridx=x;
		constraints.gridy=y;
		constraints.gridheight=height;
		constraints.gridwidth=width;
		constraints.weightx=weightx;
		constraints.weighty=weighty;
		constraints.insets=insets;
		constraints.fill=fill;
		constraints.anchor=anchor;
		
		return constraints;
	}
	
	protected GridBagConstraints createConstraints(int x, int y, int width, int height, double weightx, double weighty,int fill,int anchor)  {
		
		return createConstraints(x,y,width,height,weightx,weighty,fill,anchor,new Insets(0,0,0,0));
	}
	
	protected GridBagConstraints createConstraints(int x, int y, int width, int height, double weightx, double weighty,int fill)  {
		
		return createConstraints(x,y,width,height,weightx,weighty,fill,GridBagConstraints.CENTER,new Insets(0,0,0,0));
	}
	
	protected GridBagConstraints createConstraints(int x, int y, int width, int height, double weightx, double weighty)  {
		
		return createConstraints(x,y,width,height,weightx,weighty,GridBagConstraints.NONE,GridBagConstraints.CENTER,new Insets(0,0,0,0));
	}
	
	
	
	protected JPanel setupClassPathTab() {
		
		JPanel classPathPanel = new JPanel();
		
		GridBagLayout layout = new GridBagLayout();
		
		JButton addPath = new JButton("ADD PATH");
		JButton addJar = new JButton("ADD JAR");
		JButton remove = new JButton("REMOVE SELECTION");
		JButton removeAll = new JButton("REMOVE ALL");
		
		final JList classPathList = new JList(classPath);
		
		addPath.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
			
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				fileChooser.showOpenDialog();
				if(!fileChooser.isCanceled())
					classPath.addElement(fileChooser.getSelectedFile().getPath());
			}
		});
		addJar.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
			
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				fileChooser.showOpenDialog();
				if(!fileChooser.isCanceled())
					classPath.addElement(fileChooser.getSelectedFile().getPath());
			}
		});
		remove.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
			
				int[] indices = classPathList.getSelectedIndices();
				for(int i=0; i<indices.length;++i)
					classPath.remove(indices[i]);
			}
		});
		removeAll.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
			
				classPath.removeAllElements();
			}
		});
		
		JScrollPane classPathScrollPane = new JScrollPane(classPathList);
		classPathScrollPane.setBorder(new TitledBorder("Current ClassPath"));
		
		Iterator classPathI = Options.getClassPathVector().iterator();
		while(classPathI.hasNext()) 
			classPath.addElement(classPathI.next());
		
		classPathPanel.setLayout(layout);
		layout.setConstraints(classPathScrollPane,createConstraints(0,0,1,4,0.80,1.00,GridBagConstraints.BOTH));
		layout.setConstraints(addPath,createConstraints(1,0,1,1,0.20,0.25));
		layout.setConstraints(addJar,createConstraints(1,1,1,1,0.20,0.25));
		layout.setConstraints(remove,createConstraints(1,2,1,1,0.20,0.25));
		layout.setConstraints(removeAll,createConstraints(1,3,1,1,0.20,0.25));
		classPathPanel.add(classPathScrollPane);
		classPathPanel.add(addPath);
		classPathPanel.add(addJar);
		classPathPanel.add(remove);
		classPathPanel.add(removeAll);
		
		return classPathPanel;
	}
	
	protected void dialogApproved() {
    
		
		Enumeration classPathE = classPath.elements();
		Vector classPathVector = new Vector();
		while(classPathE.hasMoreElements()) 
			classPathVector.add(classPathE.nextElement());
		
		Options.setClassPathVectorSafe(classPathVector);
		
        Options.saveOptions();
		mainWindow.repaint();
		setVisible(false);
	} 

	protected void dialogCanceled() {
		setVisible(false);
		mainWindow.repaint();
	}
	
	public void show() {
		
		
		classPath.removeAllElements();
		Iterator classPathI = Options.getClassPathVector().iterator();
		while(classPathI.hasNext()) 
			classPath.addElement(classPathI.next());
		
		
		super.setVisible(true);
	}

}



