/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 20-okt-2004
 */
package jasco.tools.jascoparser;

import jasco.util.generators.ClassGenerator;

import java.util.Iterator;
import java.util.Vector;
/**
 * @author Wim
 */
public class PJavaClass extends PJAsCoParseElement {

	private Vector methods = new Vector();
	private Vector fields = new Vector();
	private Vector staticInits = new Vector();
	private String name="";
	private String superClass="";
	private int modifiers = 0;
	private String annotations = "";
	
	/**
	 * @return Returns the annotations.
	 */
	public String getAnnotations() {
		return annotations;
	}
	/**
	 * @param annotations The annotations to set.
	 */
	public void setAnnotations(String annotations) {
		this.annotations = annotations;
	}
	/**
	 * @return Returns the modifiers.
	 */
	public int getModifiers() {
		return modifiers;
	}
	/**
	 * @param modifiers The modifiers to set.
	 */
	public void setModifiers(int modifiers) {
		this.modifiers = modifiers;
	}
	
	public PJavaClass(int line){
		super(line);
	}
	
	public PJavaClass(){ 
		super(NO_LINE);
	}
	
	public PJavaClass(int line, String name) {
		super(line);
		this.name=name;
	}
	
	public PJavaClass(String name) {
		super(NO_LINE);
		this.name=name;
	}
	
	public void addStaticInitializer(PJavaCode init) {
		staticInits.add(init);
	}
	
	public void setName(String n) {
		this.name=n;
	}
	
	/**
	 * @return Returns the fields.
	 */
	public Iterator getFields() {
		return fields.iterator();
	}
	
	public void addField(PField f) {
		this.fields.add(f);
	}
	
	public Iterator getMethods() {
		return methods.iterator();
	}
	
	public void addMethod(PMethod f) {
		this.methods.add(f);
	}
	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * @return Returns the staticInits.
	 */
	public Iterator getStaticInits() {
		return staticInits.iterator();
	}
	
	public String getSuperClass() {
		return superClass;
	}
	
	public String getSuperClassSimple() {
		
		int i = superClass.lastIndexOf('.');
		if(i<0)
			return superClass;
		else return superClass.substring(i+1);
	}
	
	public boolean hasSuperClass() {
		return !superClass.equals("");
	}
	
	public void setSuperClass(String s) {
		superClass=s;
	}
	
	public void initClassGenerator(ClassGenerator gen) {
		
		gen.addAnnotation(getAnnotations());
		
		Iterator i = getMethods();
		while(i.hasNext()) {
			PMethod method = (PMethod) i.next();
			gen.addMethod(method.getMethodInfo());
		}
		
		i = getStaticInits();
		while(i.hasNext()) {
			PJavaCode java = (PJavaCode) i.next();
			gen.addStaticInit(java.toString());
		}
		
		i = getFields();
		while(i.hasNext()) {
			PField field = (PField) i.next();
			gen.addVariable(field.getFieldInfo());
		}
		
	}
	
}
