/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.traversalparser;

import java.util.Iterator;
import java.util.Vector;

public class HookInstance  {
	
	public static final int NO_EXECUTIONS = -1;
	
	private String hookClass = "";
	private Vector methods = new Vector();
	private String varname = "";
	private Vector exec = new Vector();
	private int beforePri = Integer.MAX_VALUE;
	private int afterPri = Integer.MAX_VALUE;
	private int replacePri = Integer.MAX_VALUE;
		private Vector abstractMethods = new Vector();
	
	public HookInstance(String hookClass,String varname)  {
	
		this.hookClass=hookClass;
		this.varname=varname;
	}
	
	public void addAbstractMethod(AbstractMethodImplementation am)  {
	
		abstractMethods.add(am);
	}
	
	public Iterator getAbstractMethods()  {
	
		return abstractMethods.iterator();
	}
	
	public String getVarName()  {
	
		return varname;
	}
	
	public Iterator getHookInits()  {
	
		return methods.iterator();
	}
	
	public void addHookInit(HookInit hc)  {
	
		methods.add(hc);
	}
	
	public String getDeclaredHookClass()  {
	
		return hookClass;
	}
	
	public String getAspectBean()  {
		
		String hookclass = getDeclaredHookClass();
		int index=hookclass.lastIndexOf(".");
		return hookclass.substring(0,index);
	}
	
	public String getHookClassType()  {
		
		String hookclass = getDeclaredHookClass();
		int index=hookclass.lastIndexOf(".");
		String result = hookclass.substring(0,index)+"_"+hookclass.substring(index+1);
		return result;
	}
	
	protected void setNoexecs()  {
	
		if(beforePri==Integer.MAX_VALUE)
			beforePri=NO_EXECUTIONS;
		if(afterPri==Integer.MAX_VALUE)
			afterPri=NO_EXECUTIONS;
		if(replacePri==Integer.MAX_VALUE)
			replacePri=NO_EXECUTIONS;	
	}	
	
	public void addHookExecution(String init)  {
	
		exec.add(init);
	}
	
	public Iterator getHookExecutions()  {
	
		return exec.iterator();
	}
	
	public void setBeforePriority(int p)  {
		
		setNoexecs();
		beforePri=p;
	}
	
	public void setAfterPriority(int p)  {
		
		setNoexecs();
		afterPri=p;
	}
	
	public void setReplacePriority(int p)  {
		
		setNoexecs();
		replacePri=p;
	}
	
	public int getBeforePriority()  {
	
		return beforePri;
	}
	
	public int getAfterPriority()  {
	
		return afterPri;
	}
	
	public int getReplacePriority()  {
	
		return replacePri;
	}
	
	public void setDefaultPriority(int i)  {
	
		beforePri=i;
		afterPri=i;
		replacePri=i;
	}
	
}
	