/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 20-okt-2004
 */
package jasco.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 * @author Wim
 */
public abstract class StreamGobbler extends Thread {
	InputStream is;
	String type;
	boolean typed;
	
	public static final String ERROR = "ERROR";
	public static final String OUTPUT = "OUTPUT";

	public StreamGobbler(InputStream is, String type, boolean typed) {
		this.is = is;
		this.type = type;
		this.typed = typed;
		setDaemon(true);	
	}

	public StreamGobbler(InputStream is, String type) {
		this(is, type, true);
	}

	public void run() {
		try {
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader br = new BufferedReader(isr);
			String line = null;
			while ((line = br.readLine()) != null)  {
				if (typed)
					line = type + "> " + line;
				
				if(type.equals(ERROR))
					showError(line);
				else showOutput(line);
			}
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}
	
	public abstract void showOutput(String s);
	public abstract void showError(String s);
	public abstract void flush();
}
