/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util.javacompiler;

import jasco.options.Options;

import java.util.Vector;

// Compiles the s returned by the java compiler
public class DefaultCompileErrorParser implements ICompileErrorParser
{
	private Vector errors = new Vector();
	private int cfirst;
	private int clast;
	
	DefaultCompileErrorParser() 
	{
	}
	
	public Vector parseErrors(Vector input)
	{	
		errors.clear();
		for (int i = 0 ; i < (input.size() - 1);) 
		{
			String temp = (String)input.elementAt(i);
			
			String sourcefile = null;
			String linenumber = null;
			String error = null;
			
		
			
			if (temp.startsWith("Note:"))  {
				++i;
				continue;
			}
			
			if (isErrorStart(temp)) 
			{
				sourcefile = temp.substring(0,cfirst);
				
				linenumber = temp.substring(cfirst+1,clast);
				
				error = temp.substring(clast+1) + Options.NEWLINE;
				i++;
				
				while ((i < input.size()) && (!isErrorStart((String)input.elementAt(i))) )
				{	
					String temp2 = (String)input.elementAt(i);
					if(temp2.endsWith("error")||temp2.endsWith("errors"))
						;
					else error = error + temp2 + Options.NEWLINE;
					i++;
					
				}
				//System.out.println(error);
				CompileError cerror = new CompileError(sourcefile,getNumber(linenumber),error);
				errors.add(cerror);
			}
			else  {
				errors.add(new CompileError("",CompileError.GLOBAL_ERROR,temp));
				++i;
			}
			
		
		}
		
		return getErrors();
	}
	
	public Vector parseWarnings(Vector errors) {
		
		return new Vector();
	}
	
	public Vector getErrors()
	{
		return errors;
	}
		
	private boolean isNumber(String aString) 
	{
		boolean isnumber = true;
		try 
		{
			Integer.valueOf(aString);
		}
		catch (NumberFormatException e)
		{
			isnumber = false;
		}
		return isnumber;
	}
	
	private int getNumber(String aString)
	{
		int number = 0;
		try 
		{
			number = (Integer.valueOf(aString)).intValue();
		}
		catch (NumberFormatException e)
		{
			System.out.println(e);
		}
		return number; 
	}
	
	private boolean isErrorStart(String aInput)
	{	
		
		boolean returnvalue = false;
		int last = aInput.lastIndexOf(':');
		if (last != -1) 
		{
			int first = aInput.lastIndexOf(':',last - 1);
			if (first != - 1) 
			{
				String sub = aInput.substring(first+1,last);
				if (isNumber(sub))
				{
					if ((sub.indexOf('"') == -1) || (sub.indexOf('"') > first))
					{
						//set variables first and last for the parseErrors fuction, otherwise, the entire recursive process has to be redone
						returnvalue = true;
						cfirst = first;
						clast = last;
					}
				}
				else 
				{
					returnvalue = isErrorStart(aInput.substring(0,last));
				}
			}
		}
		return returnvalue;
	}	
}
					