/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util.javacompiler;

import jasco.options.Options;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.StringTokenizer;
import java.util.Vector;

public class JavaCompiler  {

	private Vector errors = new Vector();
	private static IJavaCompilerPlugin plugin = new DefaultJavaCompilerPlugin();
	
	public JavaCompiler() 
	{
	}
	
	public static void installNewPlugin(IJavaCompilerPlugin plugin1) {
		
		plugin=plugin1;
	}
	
	public static IJavaCompilerPlugin getPlugin() {
		
		return plugin;
	}
	 
	public void compile(String aSourceFile,String aClassPath) 
	{	
		compile(aSourceFile, aClassPath,false);
	}
	
	public void compile(String aSourceFile,String aClassPath,boolean bootStrap) 
	{	
		if(getPlugin()==null)
			throw new IllegalArgumentException("java compiler plugin not installed");
		
		String command = "";
		
		Vector options = Options.getCompileOptions();
		java.util.Iterator optionsI = options.iterator();
		while(optionsI.hasNext()) {
		    command+=optionsI.next()+" ";	   	
	    }
		
		if(!bootStrap)
			command+=" -classpath \"" + aClassPath + "\" \"" +aSourceFile+"\"";
		else
			command+=" -Xbootclasspath/p:\"" + aClassPath + "\" \"" +aSourceFile+"\"";
	    
		if(jasco.options.Options.showDebugOutput())
			System.out.println("JASCO DEBUG: "+command); 
		
		StringWriter stringEW = new StringWriter();
		StringWriter stringOW = new StringWriter();
		
		PrintWriter epW = new PrintWriter(stringEW);
		PrintWriter opW = new PrintWriter(stringOW);
		
		getPlugin().compile(command,epW,opW);
		
		String s = stringEW.toString(); 
		StringTokenizer tokenizer = new StringTokenizer(s,Options.NEWLINE);
	    while(tokenizer.hasMoreElements()) {
	    	String error = tokenizer.nextToken();
	    	if(jasco.options.Options.showDebugOutput())
				System.out.println(error);
			errors.add(error);
	    }
	 
	    
	    
		/*
		try {
			if(true) {
			compileOriginal(aSourceFile,aClassPath,bootStrap);
			return;
			}
			//System.out.println(System.getProperty("java.class.path"));
			Vector options =  jasco.options.Options.getCompileOptions();
			if(!bootStrap) {
				options.add("-classpath");
				options.add("\"" + aClassPath + "\"");
			}
			else options.add(" -Xbootclasspath/p:\"" + aClassPath + "\"");
			
			options.add(aSourceFile);
			
			StringWriter stringW = new StringWriter();
			
			PrintWriter writer = new PrintWriter(stringW);
			String[] command = new String[options.size()];
			options.copyInto(command);
			
			if(Options.isJavaAtLeastVersion(1,5)) {
			try {
				Class.forName("com.sun.tools.javac.main.Main");
				com.sun.tools.javac.main.Main main = new com.sun.tools.javac.main.Main("",writer);
				main.compile(command);
			} catch(ClassNotFoundException ex) {
				ex.printStackTrace();
				compileOriginal(aSourceFile,aClassPath,bootStrap);
				return;
			}
			}
			
			else if(Options.isJavaAtLeastVersion(1,4)) {
			try {
			Class.forName("com.sun.tools.javac.Main");
			com.sun.tools.javac.Main.compile(command,writer);
			} catch(ClassNotFoundException ex) {ex.printStackTrace();
			compileOriginal(aSourceFile,aClassPath,bootStrap);
			return;
			}
			}
			 
			BufferedReader reader = new BufferedReader(new StringReader(stringW.toString()));
			String buffer = null;
			while ((buffer = reader.readLine()) != null)  
			{ 
				if(jasco.options.Options.showDebugOutput())
					System.out.println(buffer);
				errors.add(buffer);
			}
			reader.close();
		}
		catch(Exception e)
		{
			Logger.getInstance().showFullError(e);
		}
		*/
	}
	
	public Vector getErrors()
	{
		return plugin.getCompileErrorParser().parseErrors(errors);
	}		
	
	public Vector getWarnings()
	{
		return plugin.getCompileErrorParser().parseErrors(errors);
	}
}